(*===========================================================================*)
(* Clean up messages                                                         *)
(*                                                                           *)
(*   Copyright 1990, 1991 by H. Roy Engehausen.  All rights reserved.        *)
(*                                                                           *)
(*===========================================================================*)

{$UNDEF DEBUG_CLEAN_NO}
{$UNDEF DEBUG_NO_KILL}

PROCEDURE do_clean;

  TYPE
    clean_type_type = (clean_unknown,
                       clean_arrl,
                       clean_amsat,
                       clean_races,
                       clean_generate,
                       clean_rttydx,
                       clean_date);

  VAR
    bid          : bid_str;
    clean_first  : BOOLEAN;
    clean_last_n : LONGINT;
    clean_last_p : msg_index_ptr;
    clean_name   : STRING[8];
    clean_num    : LONGINT;
    clean_type   : clean_type_type;
    code         : INTEGER;
    delta_num    : LONGINT;
    delete_msg   : msg_index_ptr;
    january      : BOOLEAN;
    num_end      : BYTE;
    num_end_char : CHAR;
    num_start    : BYTE;
    search_info  : search_block_type;
    this_msg     : msg_index_ptr;

  LABEL
    number_ready,
    delete_now,
    do_next_msg;

  PROCEDURE msg_error(info : STRING);
    BEGIN;
      STR(this_msg^.msg_i_mb.msg_number, s2);
      send_tnc_data_str(info + ' -- ' + s2 + cr);
    END;

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Get the clean type info                                               *)
    (*-----------------------------------------------------------------------*)

    clean_name := s2;
    clean_type := clean_unknown;

    IF clean_name = 'ARRL' THEN
      clean_type := clean_arrl;

    IF clean_name = 'AMSAT' THEN
      clean_type := clean_amsat;

    IF clean_name = 'RACES' THEN
      clean_type := clean_races;

    IF clean_name = 'DATE' THEN
      clean_type := clean_date;

    IF clean_name = 'GENERATE' THEN
      clean_type := clean_generate;

    IF clean_name = 'RTTYDX' THEN
      clean_type := clean_rttydx;

    IF clean_type = clean_unknown THEN
      BEGIN;
        send_tnc_data_str('Invalid clean type -- ' + clean_name + cr);
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* See if the month of January matters                                   *)
    (*-----------------------------------------------------------------------*)

    january := (clean_type <> clean_races)
               AND (clean_type <> clean_date)
                                     AND (COPY(todays_date_time, 3, 2) = '01');

    (*-----------------------------------------------------------------------*)
    (* Build the search arguement                                            *)
    (*-----------------------------------------------------------------------*)

    set_search(s1, @search_info);
    IF active_tcb^.error_sw THEN
      BEGIN;
        send_tnc_data_str('Invalid search on action -- ' + s1 + cr);
        free_task_mem('MSB', TRUE);
      END;

    (*-----------------------------------------------------------------------*)
    (* Prepare to do the search                                              *)
    (*-----------------------------------------------------------------------*)

    search_info.search_last := NIL;
    search_info.search_nok  := TRUE;
    clean_first             := TRUE;

    GOTO do_next_msg;

    (*-----------------------------------------------------------------------*)
    (* Search loop                                                           *)
    (*-----------------------------------------------------------------------*)

    REPEAT

      (*---------------------------------------------------------------------*)
      (* Put msg pointer somewhere handy                                     *)
      (*---------------------------------------------------------------------*)

      this_msg := search_info.search_last;

      (*---------------------------------------------------------------------*)
      (* Skip killed messages                                                *)
      (*---------------------------------------------------------------------*)

      IF (this_msg^.msg_i_mb.msg_flag AND mf_kill) <> 0 THEN
        GOTO do_next_msg;

      (*---------------------------------------------------------------------*)
      (* If the clean type is DATE then use the msg originate date/time      *)
      (* stamp and skip analyzing the bid.                                   *)
      (*---------------------------------------------------------------------*)

      IF clean_type = clean_date THEN
          BEGIN;
            clean_num := this_msg^.msg_i_mb.msg_dt_orig;
            bid       := 'DATE';
            GOTO number_ready;
          END;

      (*---------------------------------------------------------------------*)
      (* Get the bid into somewhere handy                                    *)
      (*---------------------------------------------------------------------*)

      bid := this_msg^.msg_i_mb.msg_bid;

      (*---------------------------------------------------------------------*)
      (* Locate the last position where a number could appear                *)
      (*---------------------------------------------------------------------*)

      CASE clean_type OF

        (*-------------------------------------------------------------------*)
        (* Bid ends in the number.  Either xxxxxdddd or xxxxxx.ddd           *)
        (*-------------------------------------------------------------------*)

        clean_arrl, clean_races:
          num_end := LENGTH(bid);

        (*-------------------------------------------------------------------*)
        (* xxxxdddd.xxx                                                      *)
        (*-------------------------------------------------------------------*)

        clean_amsat, clean_generate:
          BEGIN;

            IF clean_type = clean_amsat THEN
              num_end_char := '.'
            ELSE
              num_end_char := '_';

            num_end := POS(num_end_char, bid);
            IF num_end = 0 THEN
              BEGIN;
                msg_error('No "' + num_end_char + '" in the BID -- ' + bid);
                GOTO do_next_msg;
              END;

            DEC(num_end);

          END;

        (*-------------------------------------------------------------------*)
        (* Number is just inside the end. xxxxxddddx                         *)
        (*-------------------------------------------------------------------*)

        clean_rttydx:
          BEGIN;

            num_end := LENGTH(bid);

            IF num_end > 6 THEN             (* Use num_start to limit search *)
              num_start := num_end - 4
            ELSE
              num_start := 2;

            REPEAT
              DEC(num_end);
            UNTIL (num_end < num_start)
                          OR ((bid[num_end] >= '0') AND (bid[num_end] <= '9'));

          END;

      END; (*----- End CASE statement on clean_type -------------------------*)

      (*---------------------------------------------------------------------*)
      (* Calculate where the number starts                                   *)
      (*---------------------------------------------------------------------*)

      num_start := num_end;
      WHILE (num_start > 0)
                     AND (bid[num_start] >= '0')
                     AND (bid[num_start] <= '9') DO
        DEC(num_start);

      INC(num_start);

      {$IFDEF DEBUG_CLEAN_NO}
        WRITELN('Message no start/end = ', num_start, '/', num_end);
      {$ENDIF}

      (*---------------------------------------------------------------------*)
      (* Validate column numbers                                             *)
      (*---------------------------------------------------------------------*)

      IF (num_start = 1)
                      OR (num_start = num_end)
                      OR ((num_end - num_start) > 3) THEN
        BEGIN;
          msg_error('Invalid number in bid -- ' + bid);
          GOTO do_next_msg;
        END;

      (*---------------------------------------------------------------------*)
      (* Get the number from the bid and make sure it converted ok           *)
      (*---------------------------------------------------------------------*)

      VAL(COPY(bid, num_start, num_end-num_start+1), clean_num, code);

      {$IFDEF DEBUG_CLEAN_NO}
        WRITELN('Message value = ', clean_num, '/', code);
      {$ENDIF}

      IF (code <> 0) OR (clean_num = 0) THEN
        BEGIN;
          msg_error('Invalid number in bid -- ' + bid);
          GOTO do_next_msg;
        END;

      (*---------------------------------------------------------------------*)
      (* This label is used for special code where the number is already     *)
      (* formed                                                              *)
      (*---------------------------------------------------------------------*)

number_ready:

      (*---------------------------------------------------------------------*)
      (* If this is the first message then save the data and go to next file *)
      (*---------------------------------------------------------------------*)

      IF clean_first THEN
        BEGIN;
          clean_first  := FALSE;
          clean_last_n := clean_num;
          clean_last_p := this_msg;
          GOTO do_next_msg;
        END;

      (*---------------------------------------------------------------------*)
      (* Calculate difference between the current message number and         *)
      (* old one                                                             *)
      (*---------------------------------------------------------------------*)

      delta_num := clean_num - clean_last_n;

      {$IFDEF DEBUG_CLEAN_NO}
        WRITELN('Delta = ', delta_num);
      {$ENDIF}

      (*---------------------------------------------------------------------*)
      (* Identical numbers may mean trouble.  If not then just ignore        *)
      (*---------------------------------------------------------------------*)

      IF delta_num = 0 THEN
        BEGIN;
          IF clean_type <> clean_amsat THEN
            msg_error('Two messages have same bid number -- ' + bid);
          GOTO do_next_msg;
        END;

      (*---------------------------------------------------------------------*)
      (* If this is January, we use special code to watch the yearend wrap   *)
      (* of numbers.   If there is a difference of more than 30 between the  *)
      (* old and new numbers then we have a year end wraparound.  The higher *)
      (* numbered message gets deleted since it is thought to be from the    *)
      (* previous year                                                       *)
      (*---------------------------------------------------------------------*)

      IF january THEN
        BEGIN;
          IF (delta_num > 30) THEN
            BEGIN;
              delete_msg   := this_msg;
              GOTO delete_now;
            END;
          IF (delta_num < -30) THEN
            BEGIN;
              delete_msg   := clean_last_p;

              clean_last_n := clean_num;
              clean_last_p := this_msg;
              GOTO delete_now;
            END;
        END;

      {$IFDEF DEBUG_CLEAN_NO}
        WRITELN('Survived January Delete');
      {$ENDIF}

      (*---------------------------------------------------------------------*)
      (* One of the two messages supercedes the other so figure out which one*)
      (*---------------------------------------------------------------------*)

      IF delta_num > 0 THEN
        BEGIN;
          delete_msg   := clean_last_p;

          clean_last_n := clean_num;
          clean_last_p := this_msg;

          {$IFDEF DEBUG_CLEAN_NO}
            WRITELN('Delete by positive delta');
          {$ENDIF}

        END
      ELSE
        delete_msg   := this_msg;

      (*---------------------------------------------------------------------*)
      (* Come here to do the actual delete                                   *)
      (*---------------------------------------------------------------------*)

delete_now:

      {$IFNDEF DEBUG_NO_KILL}
        kill_a_msg(delete_msg, FALSE);
      {$ENDIF}

      (*---------------------------------------------------------------------*)
      (* Tell user we succeeded                                              *)
      (*---------------------------------------------------------------------*)

      STR(delete_msg^.msg_i_mb.msg_number, s2);

      send_tnc_data_str('Message ' + s2 + ' ($' + delete_msg^.msg_i_mb.msg_bid
                                        + ') killed by clean action' + cr);

      {$IFDEF DEBUG_CLEAN_NO}
        DELAY(2000);
      {$ENDIF}

      (*---------------------------------------------------------------------*)
      (* Conduct next search                                                 *)
      (*---------------------------------------------------------------------*)

do_next_msg:

      search_msg(@search_info);

    UNTIL search_info.search_last = NIL; (*----- End of search loop ---------*)

    (*-----------------------------------------------------------------------*)
    (* Free any left over memory                                             *)
    (*-----------------------------------------------------------------------*)

    free_task_mem('MSB', TRUE);

    (*-----------------------------------------------------------------------*)
    (* Leave some time                                                       *)
    (*-----------------------------------------------------------------------*)

    task_switch;

  END;
