(*===========================================================================*)
(* File subsystem -- list a directory                                        *)
(*                                                                           *)
(*   Copyright 1988, 1989, 1992 by H. Roy Engehausen.  All rights reserved.  *)
(*                                                                           *)
(*===========================================================================*)

PROCEDURE list_a_dir;

  (*-------------------------------------------------------------------------*)
  (* Maximum number of directory entires we can handle at one time           *)
  (*-------------------------------------------------------------------------*)

  CONST
    dir_max = 500;

  (*-------------------------------------------------------------------------*)
  (* Data array type                                                         *)
  (*-------------------------------------------------------------------------*)

  TYPE
    fr_type = RECORD
                fr_name : STRING[12];
                fr_time : LONGINT;
                fr_size : LONGINT;
              END;

    fr_array = ARRAY[1..dir_max] OF fr_type;

    fr_display_type = (full, short_size, short_date);

    fr_sort_type    = (name_sort, date_sort, size_sort);

    item_string     = STRING[40];

  (*-------------------------------------------------------------------------*)
  (* Local variables                                                         *)
  (*-------------------------------------------------------------------------*)

  VAR
    column_no     : BYTE;
    dir_cnt       : WORD;
    dir_data      : ^fr_array;
    dir_to_search : fsb_name_str;
    display_fr    : fr_display_type;
    first_pass    : BOOLEAN;
    i             : WORD;
    look          : SEARCHREC;
    look_for      : BYTE;
    more_kill     : BOOLEAN;
    search_arg    : file_name_str;
    sort_fr       : fr_sort_type;
    space_sum     : LONGINT;
    this_fsb      : fsb_ptr;

  (*-------------------------------------------------------------------------*)
  (* Load subroutines                                                        *)
  (*-------------------------------------------------------------------------*)

  {$I BBFSWADX.PAS}           (* Display/Sort programs are in here           *)

  (*-------------------------------------------------------------------------*)
  (* Main program                                                            *)
  (*-------------------------------------------------------------------------*)

  BEGIN;

    more_kill := FALSE;

    (*-----------------------------------------------------------------------*)
    (* Syntax check                                                          *)
    (*-----------------------------------------------------------------------*)

    IF word_count > 3 THEN
      BEGIN;
        send_message(message_err_wrd);
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Parse                                                                 *)
    (*-----------------------------------------------------------------------*)

    dir_to_search := subword(@cmd_string, 2, 1);

    search_arg    := subword(@cmd_string, 3, 1);

    IF search_arg = '' THEN
      search_arg := '*.*';

    (*-----------------------------------------------------------------------*)
    (* Get search type, etc                                                  *)
    (*-----------------------------------------------------------------------*)

    column_no  := 3;
    sort_fr    := name_sort;
    display_fr := short_size;

    CASE cmd_string[2] OF
      'X' : BEGIN;
              column_no  := 2;
              display_fr := full;
            END;
      'D' : BEGIN;
              display_fr := short_date;
            END;
      ' ' : ;
      ELSE
        BEGIN;
          send_message(message_err_2nd);
          active_tcb^.error_sw := TRUE;
          EXIT;
        END;
    END;

    (*-----------------------------------------------------------------------*)
    (* Find the directory                                                    *)
    (*-----------------------------------------------------------------------*)

    this_fsb := find_fsb(dir_to_search);

    IF (this_fsb = NIL) OR
                   (active_tcb^.uid_data.user_class < this_fsb^.fsb_down) THEN
      BEGIN;
        send_message(message_no_files_one);
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Check for subdirectory                                                *)
    (*-----------------------------------------------------------------------*)

    IF (POS('\', search_arg) > 0) AND NOT this_fsb^.fsb_f_subdir_ok THEN
      BEGIN;
        send_message(message_no_slash);
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Prepare output info                                                   *)
    (*-----------------------------------------------------------------------*)

    dir_cnt    := 0;
    first_pass := TRUE;
    space_sum  := 0;

    (*-----------------------------------------------------------------------*)
    (* Get the memory                                                        *)
    (*-----------------------------------------------------------------------*)

    IF MAXAVAIL < (SIZEOF(fr_array) + 128) THEN
      BEGIN;
        send_message(message_overlay_busy);
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    dir_data  := get_task_mem('DIR', SIZEOF(fr_array));

    (*-----------------------------------------------------------------------*)
    (* Start to search the directory                                         *)
    (*-----------------------------------------------------------------------*)


    IF this_fsb^.fsb_f_subdir_ok THEN
      look_for := ARCHIVE + DIRECTORY
    ELSE
      look_for := ARCHIVE;

    FINDFIRST(this_fsb^.fsb_path + search_arg, look_for, look);

    i := DOSERROR;

    (*-----------------------------------------------------------------------*)
    (* Handle nothing found                                                  *)
    (*-----------------------------------------------------------------------*)

    IF i = 2 THEN
      BEGIN;
        send_message(message_file_no_exist);
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Start to search the directory                                         *)
    (*-----------------------------------------------------------------------*)

    WHILE (i = 0) AND NOT more_kill DO
      WITH look DO
        BEGIN;

          (*-----------------------------------------------------------------*)
          (* If we are about to exceed the data array output it to the user  *)
          (* then reset everything                                           *)
          (*-----------------------------------------------------------------*)

          IF dir_cnt >= dir_max THEN
            BEGIN;
              display_dir;
              dir_cnt    := 0;
              first_pass := FALSE;
            END;

          (*-----------------------------------------------------------------*)
          (* Store only valid data                                           *)
          (*-----------------------------------------------------------------*)

          IF (name <> '.') AND (name <> '..') THEN
            BEGIN;

              INC(dir_cnt);

              WITH dir_data^[dir_cnt] DO
                BEGIN;

                  fr_name := name;

                  IF (attr AND directory) = 0 THEN
                    fr_size := size
                  ELSE
                    fr_size := -1;

                  fr_time   := time;

                  space_sum := space_sum + size

                END;

            END;

          (*-----------------------------------------------------------------*)
          (* Find the next item                                              *)
          (*-----------------------------------------------------------------*)

          FINDNEXT(look);
          i := DOSERROR;

        END;

    (*-----------------------------------------------------------------------*)
    (* If any data in the array, display it                                  *)
    (*-----------------------------------------------------------------------*)

    IF (dir_cnt <> 0) AND (i = 18) AND (NOT more_kill) THEN
      display_dir;

    (*-----------------------------------------------------------------------*)
    (* Free the buffer area                                                  *)
    (*-----------------------------------------------------------------------*)

    free_task_mem('DIR', TRUE);

    (*-----------------------------------------------------------------------*)
    (* Handle faulty return code                                             *)
    (*-----------------------------------------------------------------------*)

    IF (i <> 18) AND NOT more_kill THEN
      BEGIN;
        send_tnc_data_str(dos_err_message(i) + cr);
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Handle no data found                                                  *)
    (*-----------------------------------------------------------------------*)

    IF (dir_cnt = 0) AND first_pass THEN
      BEGIN;
        send_message(message_file_no_exist);
        active_tcb^.error_sw := TRUE;
        EXIT;
      END;

    (*-----------------------------------------------------------------------*)
    (* Display space left                                                    *)
    (*-----------------------------------------------------------------------*)

    display_space;

  END;
