(*===========================================================================*)
(* Set up for search                                                         *)
(*                                                                           *)
(*     x_     -- Messages of type ' '                                        *)
(*     xA     -- Messages of type x.  If x is omitted, "A" is used.          *)
(*     xB     -- Messages of type 'B'                                        *)
(*     xC x   -- User class                                                  *)
(*     xD > d -- Messages newer than date d.                                 *)
(*     xD < d -- Messages older than date d.                                 *)
(*     xE u   -- Messages to a server.  Same as > but @ must be blank or     *)
(*               this BBS                                                    *)
(*     xF n   -- Forwarded messages greater than n                           *)
(*     xG u   -- Currently logged on station                                 *)
(*     xH n   -- Held messages greater than n                                *)
(*     xK n   -- Killed messages greater than n                              *)
(*     xL n   -- Last n messages                                             *)
(*     xM u   -- Messages of a specific user.  Default is current user       *)
(*     xN n   -- Messages just sitting here unread greater than n            *)
(*     xO n   -- Messages of with "old" attribute & number greater n         *)
(*     xP     -- Messages of type 'P'                                        *)
(*     xQ     -- Read messages to a specific user                            *)
(*     xS s   -- Messages with string in subject                             *)
(*     xT     -- NTS messages                                                *)
(*     xU u   -- Unread messages to user u.  Current user is default         *)
(*     xV n   -- All messages in review status greater than n                *)
(*     xY n   -- All read messages greater than n                            *)
(*     x< u   -- Messages from a user u.                                     *)
(*     x> u   -- Messages to a user u.                                       *)
(*     x@ b   -- Messages to a bbs b.                                        *)
(*     x% b   -- Messages from a bbs b.                                      *)
(*     x$ s   -- Messages with string in the BID                             *)
(*     x+ n   -- Messages in forwarding progress with number >= n            *)
(*     x- n   -- Messages being forwarded right now with number >= n         *)
(*     x. s   -- Messages addressed to certain H addresses                   *)
(*     x# n   -- Messages with number >= n                                   *)
(*     x? n   -- Messages with number >= n and unknown routing               *)
(*     x0     -- Messages since last "L" command                             *)
(*                                                                           *)
(*   Copyright 1988, 1989, 1990, 1991, 1992 by H. Roy Engehausen.  All       *)
(*   rights reserved.                                                        *)
(*                                                                           *)
(*===========================================================================*)

{$O+}

UNIT BBSEARCH;

INTERFACE

  USES
    bbdummy;

  PROCEDURE set_search (VAR cmd_string  : STRING;
                            search_ptr  : search_block_ptr);

  FUNCTION  search_test(VAR cmd_string  : STRING;
                            search_this : msg_index_ptr) : BOOLEAN;

  FUNCTION  search_test_block(search_ptr : search_block_ptr;
                              search_this : msg_index_ptr) : BOOLEAN;

IMPLEMENTATION

  USES
    bbmem,
    bbmdata,
    bbmess,
    bbmf,
    bbmisc,
    bbmisc2,
    bbsdata,
    bbstr,
    bbtime,
    bbtrace;

(*===========================================================================*)
(* Main routine                                                              *)
(*===========================================================================*)

{$UNDEF  DEBUG1}
{$UNDEF  DEBUG2}
{$UNDEF  DEBUG3}
{$UNDEF  DEBUG4} (* Debug intermittent hits on action stuff *)
{$UNDEF  DEBUG5} (* Debug memory error                      *)

(*===========================================================================*)
(* Set up for single search                                                  *)
(*===========================================================================*)

PROCEDURE set_search_single(cmd_string  : STRING;
                            search_ptr  : search_block_ptr);
  VAR
    b            : BOOLEAN;
    code         : INTEGER;
    search_block : search_block_type;
    type_char    : CHAR;
    word_count   : BYTE;
    work_word    : STRING[14];

  LABEL err_exit_1;
  LABEL err_exit_2;
  LABEL err_exit_c;

  (*=========================================================================*)
  (* Subroutine to get a number                                              *)
  (*=========================================================================*)

  PROCEDURE set_number;

    VAR
      msg_no       : LONGINT;

    BEGIN;

      (*---------------------------------------------------------------------*)
      (* See if operand present                                              *)
      (*---------------------------------------------------------------------*)

      IF word_count > 1 THEN
        BEGIN;

          (*-----------------------------------------------------------------*)
          (* Operand present.  This is a message number                      *)
          (*-----------------------------------------------------------------*)

          VAL(work_word, msg_no, code);

          (*-----------------------------------------------------------------*)
          (* Check for validity                                              *)
          (*-----------------------------------------------------------------*)

          IF (code <> 0) OR (msg_no > 65535) OR (msg_no < 1) THEN
            BEGIN;
              send_message(message_err_ivm);
              active_tcb^.error_sw := TRUE;
              EXIT;
            END;

          search_block.search_msg_no := msg_no;

        END
      ELSE

        (*-------------------------------------------------------------------*)
        (* Operand not present.  Assume zero                                 *)
        (*-------------------------------------------------------------------*)

        search_block.search_msg_no := 0;

    END;

  (*=========================================================================*)
  (* Main routine starts here                                                *)
  (*=========================================================================*)

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Initialize the search block                                           *)
    (*-----------------------------------------------------------------------*)

    FILLCHAR(search_block, SIZEOF(search_block), #0);
    search_block.search_nok  := TRUE;

    (*-----------------------------------------------------------------------*)
    (* Strip things down                                                     *)
    (*-----------------------------------------------------------------------*)

    strip_var(cmd_string, 'B');

    {$IFDEF DEBUG1}
      WRITELN('Search in = ', LENGTH(cmd_string), ' = ', cmd_string);
    {$ENDIF}

    {$IFDEF DEBUG2}
      WRITELN('Search 1 in = ', LENGTH(cmd_string), ' = ', cmd_string);
    {$ENDIF}

    (*-----------------------------------------------------------------------*)
    (* See if the '^' is present                                             *)
    (*-----------------------------------------------------------------------*)

    IF cmd_string[1] = '^' THEN
      BEGIN;
        search_block.search_invert := TRUE;
        cmd_string := COPY(cmd_string, 2, 255);
        strip_var(cmd_string, 'B');
      END;

    {$IFDEF DEBUG1}
      WRITELN('Search not = ', LENGTH(cmd_string), ' = ', cmd_string);
    {$ENDIF}

    (*-----------------------------------------------------------------------*)
    (* Parse command                                                         *)
    (*-----------------------------------------------------------------------*)

    upcase_str_var(cmd_string);

    word_count := WORDS(cmd_string);

    type_char := cmd_string[1];

    work_word := subword(@cmd_string, 2, 1);

    (*-----------------------------------------------------------------------*)
    (* Make sure the type only had one letter                                *)
    (*-----------------------------------------------------------------------*)

    IF (cmd_string[2] <> ' ') AND (LENGTH(cmd_string) > 1) THEN
      GOTO err_exit_1;

    (*-----------------------------------------------------------------------*)
    (* Use the search character                                              *)
    (*-----------------------------------------------------------------------*)

    search_block.search_type := type_char;

    (*-----------------------------------------------------------------------*)
    (* Execute right routine                                                 *)
    (*-----------------------------------------------------------------------*)

    CASE type_char OF

      (*---------------------------------------------------------------------*)
      (* Handle 'D' -- By date                                               *)
      (*---------------------------------------------------------------------*)

      'D': BEGIN;

             (*--------------------------------------------------------------*)
             (* Check word counts -- Two mandatory operands                  *)
             (*--------------------------------------------------------------*)

             IF word_count > 3 THEN
               BEGIN;
                 send_message(message_err_wrd);
                 GOTO err_exit_1;
               END;

             IF word_count < 3 THEN
               BEGIN;
                 send_message(message_not_en);
                 GOTO err_exit_1;
               END;

             (*--------------------------------------------------------------*)
             (* First operand is test                                        *)
             (*--------------------------------------------------------------*)

             IF work_word = 'AFTER' THEN
               work_word := '>';

             IF work_word = 'BEFORE' THEN
               work_word := '<';

             IF LENGTH(work_word) <> 1 THEN
               GOTO err_exit_2;

             (*--------------------------------------------------------------*)
             (* Set the parameters                                           *)
             (*--------------------------------------------------------------*)

             CASE work_word[1] OF
               '<' : search_block.search_ascend := TRUE;
               '>' : search_block.search_above  := TRUE;
               ELSE
                 GOTO err_exit_2;
             END;

             (*--------------------------------------------------------------*)
             (* Set the direction                                            *)
             (*--------------------------------------------------------------*)

             search_block.search_direction := TRUE;

             (*--------------------------------------------------------------*)
             (* Get the date                                                 *)
             (*--------------------------------------------------------------*)

             search_block.search_str := subword(@cmd_string, 3, 1);

             IF search_block.search_str[1] <> '#' THEN
               search_block.search_dt := time_unstr(search_block.search_str, b)
             ELSE
               BEGIN;
                 work_word := copy(search_block.search_str, 2, 255);
                 set_number;
                 b := search_block.search_msg_no > 1000;
                 IF NOT b THEN
                   search_block.search_dt := last_midnight -
                           LONGINT(search_block.search_msg_no) * ticks_per_day;
               END;

             IF b THEN
               BEGIN;
                 send_message(message_bad_date);
                 GOTO err_exit_1;
               END;

           END; (*----- End of 'D' command handler --------------------------*)

      (*---------------------------------------------------------------------*)
      (* Handle 'F', 'H', 'K', 'O', 'V', 'Y', '+', '-', '?'                  *)
      (*---------------------------------------------------------------------*)

      'F',
      'H',
      'K',
      'O',
      'V',
      'Y',
      '+',
      '-',
      '?': BEGIN;

             (*--------------------------------------------------------------*)
             (* Check privilege                                              *)
             (*--------------------------------------------------------------*)

             IF NOT check_priv(user_c_rsu) THEN
               BEGIN;
                 work_word := type_char;
                 GOTO err_exit_2;
                 EXIT;
               END;

             (*--------------------------------------------------------------*)
             (* Too many words                                               *)
             (*--------------------------------------------------------------*)

             IF word_count > 2 THEN
               BEGIN;
                 send_message(message_err_wrd);
                 GOTO err_exit_1;
               END;

             (*--------------------------------------------------------------*)
             (* Set a number if present                                      *)
             (*--------------------------------------------------------------*)

             set_number;
             IF active_tcb^.error_sw THEN
               GOTO err_exit_2;

             (*--------------------------------------------------------------*)
             (* Build rest of search block                                   *)
             (*--------------------------------------------------------------*)

             search_block.search_above := TRUE;

             IF type_char = 'K' THEN
               search_block.search_nok := FALSE;

             IF type_char = 'V' THEN
               search_block.search_ascend := TRUE;

             search_block.search_type   := 'K';

             CASE type_char OF
               'F' : search_block.search_mf := mf_fwd;
               'H' : search_block.search_mf := mf_hold;
               'K' : search_block.search_mf := mf_kill;
               'O' : search_block.search_mf := mf_old;
               'V' : search_block.search_mf := mf_review;
               'Y' : search_block.search_mf := mf_read;
               '+' : search_block.search_mf := mf_fwd_select OR mf_fwd_process;
               '-' : search_block.search_mf := mf_fwd_process;
               '?' : search_block.search_mf := mf_unknown;
             END;

           END; (*- End of 'F', 'H', 'K', 'Y', '+', and '-' command handler -*)

      (*---------------------------------------------------------------------*)
      (* Handle 'L' -- Last "n" messages                                     *)
      (*---------------------------------------------------------------------*)

      'L': BEGIN;

             (*--------------------------------------------------------------*)
             (* Check word counts -- One operand only                        *)
             (*--------------------------------------------------------------*)

             IF word_count > 2 THEN
               BEGIN;
                 send_message(message_err_wrd);
                 GOTO err_exit_1;
               END;

             (*--------------------------------------------------------------*)
             (* Set a number if present.  Make 0 into 1                      *)
             (*--------------------------------------------------------------*)

             IF word_count = 2 THEN
               BEGIN;
                 set_number;

                 IF active_tcb^.error_sw THEN
                 GOTO err_exit_2;
               END;

             IF search_block.search_msg_no = 0 THEN
               search_block.search_msg_no := 1;

             (*--------------------------------------------------------------*)
             (* Build rest of search block                                   *)
             (*--------------------------------------------------------------*)

             search_block.search_direction := TRUE;

           END; (*----- End of 'L' command handler --------------------------*)

      (*---------------------------------------------------------------------*)
      (* Handle 'M' -- Messages for this user                                *)
      (* Handle 'U' -- Unread messages for this user                         *)
      (*--------------------------------------------------------------------*)

      'M',
      'U': BEGIN;

             (*--------------------------------------------------------------*)
             (* Check word counts -- Either none or one operands             *)
             (*--------------------------------------------------------------*)

             IF word_count > 2 THEN
               BEGIN;
                 send_message(message_err_wrd);
                 GOTO err_exit_1;
               END;

             (*--------------------------------------------------------------*)
             (* Build search block                                           *)
             (*--------------------------------------------------------------*)

             IF type_char <> 'M' THEN
               search_block.search_type := 'R';

             search_block.search_str    := active_tcb^.uid_data.user_id;

             (*--------------------------------------------------------------*)
             (* If second word, use it as search string                      *)
             (*--------------------------------------------------------------*)

             IF word_count = 2 THEN
               search_block.search_str := work_word;

           END; (*----- End of 'M' command handler --------------------------*)

      (*---------------------------------------------------------------------*)
      (* Handle 'N' -- Unselected messages                                   *)
      (*---------------------------------------------------------------------*)

      'N': BEGIN;

             (*--------------------------------------------------------------*)
             (* Too many words                                               *)
             (*--------------------------------------------------------------*)

             IF word_count > 2 THEN
               BEGIN;
                 send_message(message_err_wrd);
                 GOTO err_exit_1;
               END;

             (*--------------------------------------------------------------*)
             (* Set a number if present                                      *)
             (*--------------------------------------------------------------*)

             set_number;
             IF active_tcb^.error_sw THEN
               GOTO err_exit_2;

             (*--------------------------------------------------------------*)
             (* Build rest of search block                                   *)
             (*--------------------------------------------------------------*)

             search_block.search_above  := TRUE;

             search_block.search_mf := mf_hold OR mf_read OR mf_fwd
                                      OR mf_review OR mf_fwd_select OR mf_kill;

           END; (*----- End of 'N' command handler --------------------------*)

      (*---------------------------------------------------------------------*)
      (* Handle '_' -- Type " " bulletins                                    *)
      (* Handle 'B' -- Type "B" bulletins                                    *)
      (* Handle 'P' -- Private messages                                      *)
      (* Handle 'T' -- NTS messages                                          *)
      (*---------------------------------------------------------------------*)

      '_',
      'B',
      'P',
      'T': BEGIN;

             (*--------------------------------------------------------------*)
             (* Check word count -- No operands allowed                      *)
             (*--------------------------------------------------------------*)

             IF word_count > 1 THEN
               BEGIN;
                 send_message(message_err_wrd);
                 GOTO err_exit_1;
               END;

             (*--------------------------------------------------------------*)
             (* Handle the underscore                                        *)
             (*--------------------------------------------------------------*)

             IF type_char = '_' THEN
               type_char := ' ';

             (*--------------------------------------------------------------*)
             (* Build search block                                           *)
             (*--------------------------------------------------------------*)

             search_block.search_type := 'T';

             search_block.search_str    := type_char;

           END; (*----- End of '_BFPT' command handler ----------------------*)

      (*---------------------------------------------------------------------*)
      (* Handle 'A' -- Type x messages                                       *)
      (*---------------------------------------------------------------------*)

      'A': BEGIN;

             (*--------------------------------------------------------------*)
             (* Check word count                                             *)
             (*--------------------------------------------------------------*)

             IF word_count > 2 THEN
               BEGIN;
                 send_message(message_err_wrd);
                 GOTO err_exit_1;
               END;

             (*--------------------------------------------------------------*)
             (* Get the type if givem else we use the default                *)
             (*--------------------------------------------------------------*)

             IF word_count <> 1 THEN
               BEGIN;
                 IF LENGTH(work_word) > 1 THEN
                   GOTO err_exit_2;
                 type_char := work_word[1];
               END;

             (*--------------------------------------------------------------*)
             (* Build search block                                           *)
             (*--------------------------------------------------------------*)

             search_block.search_type := 'T';

             search_block.search_str    := type_char;

           END; (*----- End of '_BFPT' command handler ----------------------*)

      (*---------------------------------------------------------------------*)
      (* Handle 'E' -- Messages to this user (@ field blank or this BBS)     *)
      (* Handle 'Q' -- Read messages to a user                               *)
      (* Handle 'S' -- Messages with string match                            *)
      (* Handle '>' -- Messages to this user                                 *)
      (* Handle '<' -- Messages from this user                               *)
      (* Handle '@' -- Messages to a bbs                                     *)
      (* Handle '%' -- Messages from a bbs                                   *)
      (* Handle '$' -- Messages with bid match                               *)
      (* Handle '.' -- Messages with hierarchial address match               *)
      (*---------------------------------------------------------------------*)

      'E',
      'G',
      'Q',
      'S',
      '>',
      '<',
      '@',
      '%',
      '$',
      '.': BEGIN;

             (*--------------------------------------------------------------*)
             (* Check word counts -- One mandatory operand                   *)
             (*--------------------------------------------------------------*)

             IF (word_count > 2) AND (type_char <> 'S') THEN
               BEGIN;
                 send_message(message_err_wrd);
                 GOTO err_exit_1;
               END;

             IF word_count < 2 THEN
               BEGIN;
                 send_message(message_not_en);
                 GOTO err_exit_1;
               END;

             (*--------------------------------------------------------------*)
             (* Build search block                                           *)
             (*--------------------------------------------------------------*)

             IF type_char <> 'S' THEN
               search_block.search_str  := work_word
             ELSE
               search_block.search_str := subword(@cmd_string, 2, 0);

           END; (*----- End of 'EQS><@%$' command handler -------------------*)

      (*---------------------------------------------------------------------*)
      (* Handle '#' -- Messages with a higher number                         *)
      (*---------------------------------------------------------------------*)

      '#': BEGIN;

             (*--------------------------------------------------------------*)
             (* Build search block                                           *)
             (*--------------------------------------------------------------*)

             search_block.search_direction := TRUE;

             (*--------------------------------------------------------------*)
             (* Word count check                                             *)
             (*--------------------------------------------------------------*)

             IF word_count > 3 THEN
               BEGIN;
                 send_message(message_err_wrd);
                 GOTO err_exit_1;
               END;

             IF word_count < 3 THEN
               BEGIN;
                 send_message(message_not_en);
                 GOTO err_exit_1;
               END;

             (*--------------------------------------------------------------*)
             (* First operand is test                                        *)
             (*--------------------------------------------------------------*)

             IF work_word = 'AFTER' THEN
               work_word := '>';

             IF work_word = 'BEFORE' THEN
               work_word := '<';

             IF LENGTH(work_word) <> 1 THEN
               GOTO err_exit_2;

             (*--------------------------------------------------------------*)
             (* Set the parameters                                           *)
             (*--------------------------------------------------------------*)

             CASE work_word[1] OF
               '<' : search_block.search_ascend := TRUE;
               '>' : search_block.search_above  := TRUE;
               ELSE
                 GOTO err_exit_2;
             END;

             (*--------------------------------------------------------------*)
             (* Set the number                                               *)
             (*--------------------------------------------------------------*)

             work_word := subword(@cmd_string, 3, 1);
             set_number;
             IF active_tcb^.error_sw THEN
               GOTO err_exit_2;

           END; (*----- End of '#' command handler --------------------------*)

      (*---------------------------------------------------------------------*)
      (* Handle '0' -- Messages since last "L" command                       *)
      (*---------------------------------------------------------------------*)

      '0': BEGIN;

             (*--------------------------------------------------------------*)
             (* Word count check                                             *)
             (*--------------------------------------------------------------*)

             IF word_count > 1 THEN
               BEGIN;
                 send_message(message_err_wrd);
                 GOTO err_exit_1;
               END;

             (*--------------------------------------------------------------*)
             (* Build search block                                           *)
             (*--------------------------------------------------------------*)

             search_block.search_dt    := active_tcb^.uid_data.user_l_time;
             search_block.search_type  := 'D';
             search_block.search_nok   := TRUE;
             search_block.search_above := TRUE;

           END; (*----- End of '0' command handler --------------------------*)

      (*---------------------------------------------------------------------*)
      (* Handle 'C' -- Test user class for same class or below               *)
      (*---------------------------------------------------------------------*)

      'C': BEGIN;

             (*--------------------------------------------------------------*)
             (* Word count check                                             *)
             (*--------------------------------------------------------------*)

             IF word_count > 2 THEN
               BEGIN;
                 send_message(message_err_wrd);
                 GOTO err_exit_1;
               END;

             IF word_count < 2 THEN
               BEGIN;
                 send_message(message_not_en);
                 GOTO err_exit_1;
               END;

             (*--------------------------------------------------------------*)
             (* Check the operand                                            *)
             (*--------------------------------------------------------------*)

             code := POS(work_word[1], user_class_string);
             IF (LENGTH(work_word) <> 1) OR (code = 0) THEN
               GOTO err_exit_2;

             (*--------------------------------------------------------------*)
             (* Set the parameters                                           *)
             (*--------------------------------------------------------------*)

             search_block.search_user_class := user_class_type(code - 1);

           END; (*----- End of 'C' command handler --------------------------*)

      (*---------------------------------------------------------------------*)
      (* Handle illegal commands                                             *)
      (*---------------------------------------------------------------------*)

      ELSE
        BEGIN;
          work_word := type_char;
          GOTO err_exit_2;
        END;

    END; (*----- End main case statement ------------------------------------*)

    (*-----------------------------------------------------------------------*)
    (* Finish search block prep                                              *)
    (*-----------------------------------------------------------------------*)

    search_block.search_last := NIL;

    search_ptr^ := search_block;

    (*-----------------------------------------------------------------------*)
    (* Done                                                                  *)
    (*-----------------------------------------------------------------------*)

    EXIT;

    (*-----------------------------------------------------------------------*)
    (* Error exits                                                           *)
    (*-----------------------------------------------------------------------*)

err_exit_1:
    set_dollar1_parm(@cmd_string);
    GOTO err_exit_c;

err_exit_2:
    set_dollar1_parm(@work_word);

err_exit_c:
    send_message(message_bad_search);
    active_tcb^.error_sw := TRUE;
    EXIT;

  END;

(*===========================================================================*)
(* Set up for multiple searches.  Note: first character of cmd_string is     *)
(* skipped                                                                   *)
(*===========================================================================*)

PROCEDURE set_search(VAR cmd_string : STRING;
                         search_ptr : search_block_ptr);

  VAR
    end_pos    : INTEGER;
    last_pos   : BYTE;
    len        : BYTE;
    start_pos  : BYTE;
    new_block  : search_block_ptr;
    this_block : search_block_ptr;

  BEGIN;

    {$IFDEF DEBUG2}
      WRITELN('Search M in = ', LENGTH(cmd_string), ' = ', cmd_string);
    {$ENDIF}

    {$IFDEF DEBUG5}
      WRITELN('Search trace -- ', LENGTH(cmd_string), ' = ', cmd_string);
    {$ENDIF}

    (*-----------------------------------------------------------------------*)
    (* Free any old searches                                                 *)
    (*-----------------------------------------------------------------------*)

    IF NOT active_tcb^.tcb_no_search_rel THEN
      free_task_mem(search_memory_block_id, TRUE);

    (*-----------------------------------------------------------------------*)
    (* Prepare loop variables                                                *)
    (*-----------------------------------------------------------------------*)

    end_pos    := -1;
    last_pos   := LENGTH(cmd_string);

    this_block := search_ptr;

    (*-----------------------------------------------------------------------*)
    (* Loop thru the string breaking at the proper places                    *)
    (*-----------------------------------------------------------------------*)

    REPEAT

      (*---------------------------------------------------------------------*)
      (* Bump the end pointer past the previous stopping point               *)
      (*---------------------------------------------------------------------*)

      INC(end_pos, 2);

      (*---------------------------------------------------------------------*)
      (* The current end pointer is now the starting point                   *)
      (*---------------------------------------------------------------------*)

      start_pos := end_pos;

      (*---------------------------------------------------------------------*)
      (* Go forward until we find the next break point                       *)
      (*---------------------------------------------------------------------*)

      WHILE (end_pos < last_pos)
             AND ((cmd_string[end_pos] <> ' ')
                                         OR (cmd_string[end_pos+1] <> '&')) DO
        INC(end_pos);

      (*---------------------------------------------------------------------*)
      (* Calculate what is remaining                                         *)
      (*---------------------------------------------------------------------*)

      len := end_pos - start_pos + 1;

      (*---------------------------------------------------------------------*)
      (* Build a search block for it                                         *)
      (*---------------------------------------------------------------------*)

      set_search_single(COPY(cmd_string, start_pos, len), this_block);

      (*---------------------------------------------------------------------*)
      (* If we failed then leave                                             *)
      (*---------------------------------------------------------------------*)

      IF active_tcb^.error_sw THEN
        BEGIN;
          IF NOT active_tcb^.tcb_no_search_rel THEN
            free_task_mem(search_memory_block_id, TRUE);
          EXIT;
        END;

      (*---------------------------------------------------------------------*)
      (* If we have more to go then build a new search block                 *)
      (*---------------------------------------------------------------------*)

      IF end_pos < last_pos THEN
        BEGIN;

          new_block := get_task_mem(search_memory_block_id,
                                    SIZEOF(search_block_type));

          this_block^.search_next := new_block;
          this_block := new_block;

        END;

    UNTIL (end_pos >= last_pos) AND (end_pos <> 255); (*---- End loop -------*)

  END;

(*===========================================================================*)
(* Set up a search and test one message                                      *)
(*===========================================================================*)

FUNCTION search_test(VAR cmd_string  : STRING;
                         search_this : msg_index_ptr) : BOOLEAN;

  VAR
    this_block : search_block_type;

  BEGIN;

    {$IFDEF DEBUG3}
      WRITELN('Search M in = ', LENGTH(cmd_string), ' = ', cmd_string);
    {$ENDIF}

    (*-----------------------------------------------------------------------*)
    (* Set up the search                                                     *)
    (*-----------------------------------------------------------------------*)

    set_search(cmd_string, @this_block);

    {$IFDEF DEBUG4}
      trace_data('Search test', ORD(active_tcb^.error_sw), NIL, cmd_string);
    {$ENDIF}

    (*-----------------------------------------------------------------------*)
    (* If OK then do the search                                              *)
    (*-----------------------------------------------------------------------*)

    IF NOT active_tcb^.error_sw THEN
      BEGIN;
        this_block.search_last     := search_this;
        this_block.search_one_only := TRUE;
        search_msg(@this_block);
      END;

    (*-----------------------------------------------------------------------*)
    (* Free any memory blocks we don't need                                  *)
    (*-----------------------------------------------------------------------*)

    IF NOT active_tcb^.tcb_no_search_rel THEN
      free_task_mem(search_memory_block_id, TRUE);

    (*-----------------------------------------------------------------------*)
    (* Set result                                                            *)
    (*-----------------------------------------------------------------------*)

    search_test := this_block.search_last <> NIL;

    {$IFDEF DEBUG3}
      WRITELN('Search answer is ', ORD(this_block.search_last <> NIL));
    {$ENDIF}

    {$IFDEF DEBUG4}
      trace_data('Search test', ORD(active_tcb^.error_sw),
                                           this_block.search_last, cmd_string);
    {$ENDIF}

  END;

(*===========================================================================*)
(* Test one message based on a previoursly built search                      *)
(*===========================================================================*)

FUNCTION search_test_block(search_ptr : search_block_ptr;
                           search_this : msg_index_ptr) : BOOLEAN;

  VAR
    this_block : search_block_type;

  BEGIN;

    (*-----------------------------------------------------------------------*)
    (* Set up the search                                                     *)
    (*-----------------------------------------------------------------------*)

    this_block := search_ptr^;

    (*-----------------------------------------------------------------------*)
    (* If OK then do the search                                              *)
    (*-----------------------------------------------------------------------*)

    this_block.search_last     := search_this;
    this_block.search_one_only := TRUE;
    search_msg(@this_block);

    (*-----------------------------------------------------------------------*)
    (* Set result                                                            *)
    (*-----------------------------------------------------------------------*)

    search_test_block := this_block.search_last <> NIL;

    {$IFDEF DEBUG3}
      WRITELN('Search answer is ', ORD(this_block.search_last <> NIL));
    {$ENDIF}

  END;

END.
