/*
   buf.c: Copyright 1993, Paul Healy.

   The idea of using file handles rather than pointers is to avoid any
   buffering that file pointers (might - doesn't happen with borland
   compilers as far as I know) entail.

   930503 : Created
*/
#include <stdio.h>
#include <stdlib.h>
#include "config.h"
#include "buf.h"
#include "rc.h"
#include "pc.h"

static uchar *buf_Cache = NULL;
static size_t buf_Size = 0;
                                                                                                                                                                                                               
void
buf_init(buf *bp)
{
   bp->buffer = NULL;
   bp->size = 0;
   bp->eof = 0;
   bp->curpos = -1L;
   bp->nextpos = 0L;
   bp->filesize = -1L;
}

static uchar *
buf_malloc(buf *bp, size_t size, int usecache)
{
   uchar *p = NULL;
#ifdef DEBUG
   fprintf(stderr, "bufmalloc: size=%u\n", size);
#endif
   if ( usecache && (buf_Cache != NULL) && (buf_Size == size) ) {
      p = buf_Cache;
      buf_Cache = NULL;
      buf_Size = 0;
      }
   else
      p = (uchar *) malloc(size);

   bp->buffer = p;
   bp->size = size;

   bp->curpos = -1L;
   bp->eof = 0;
      
   return p;
}

void
buf_free(buf *bp, int cacheit)
{
   if (bp->size == 0)
      return ;
   else if (cacheit) {
      /*
       * dealing with the main mailfile buffer
       */
      if (buf_Cache != NULL) {
         fprintf(stderr, "buffree: warning freeing cached buffer? old=%u, new=%u\n", buf_Size, bp->size);
         free(buf_Cache);
         }

      buf_Cache = bp->buffer;
      buf_Size = bp->size;
      bp->buffer = NULL;
      bp->size = 0;
      }
   else {
      /*
       * some smaller buffer
       */
      free(bp->buffer);
      bp->buffer = NULL;
      bp->size = 0;
      }
}

/*
 * buf_read has to deal with four cases:
 *
 *   read new file from position 0         : in cache     not in cache
 *   read a block from some other position : in cache     not in cache
 *
 *               Holds
 *   buffer[0] : '\n' for the first buffer or else last character from prev buf
 *   buffer[1] : first real char of this buffer
 *   ...
 *   &buffer[n-1] : endp-1 holds last real char
 *   &buffer[n]   : endp holds 'F' always
 *
 *   n is the number of bytes plus one that was last read in
 */
static int
buf_read(int fd, buf *bp, int prevchar)
{
   int count;
   uchar *start = buf_bufp(bp)+1;

#ifdef DEBUG
   fprintf(stderr, "bufread: curpos=%ld nextpos=%ld ", bp->curpos, bp->nextpos);
#endif

   if (fd == -1) {
      fprintf(stderr, "bmh: bad fd\n");
      return -1;
      }
   else if (bp->curpos == -1L)         /* empty buffer? */
      bp->filesize = filelength(fd);
   else {
      if (bp->filesize != filelength(fd)) {
         fprintf(stderr, "bmh: mailfile has changed\n");
         return -1;
         }
      else if ((bp->curpos==0L)&&(bp->curpos == bp->nextpos)) {  /* cache here */
         bp->nextpos = ((long) bp->endp) - ((long) bp->buffer) - 1L;
#ifdef DEBUG
         fprintf(stderr, "bmh: cache folder return, next=%ld\n", bp->nextpos);
         fprintf(stderr, "bmh: endp=%p, buffer=%p\n", bp->endp, bp->buffer);
#endif
         return 0;
         }
      else if (bp->eof) {
#ifdef DEBUG
         fprintf(stderr, "bufread: eof set\n");
#endif
         return -1;
         }
      else if (bp->nextpos != tell(fd))
         if (lseek(fd, bp->nextpos, SEEK_SET) == -1L) {
            fprintf(stderr, "bmh: seek failure to %ld\n", bp->nextpos);
            return -1;
            }
      }

   bp->curpos = tell(fd);

#ifdef DEBUG
   fprintf(stderr, "tell=%ld nextpos=%ld ", tell(fd), bp->nextpos);
#endif

   /*
    * read returns an integer, but since we may be asking for more
    * bytes that a 16 bit signed integer can hold, the cast to size_t
    * further on is required.
    */
   if ( (count = read(fd, start, (bp->size) - 2)) == -1 ) {
      perror("mailfile read");
      return -1;
      }

   buf_setptr(bp, start);
   buf_setend(bp, start + (size_t) count);

   /*
    * Ensure that a 'From' at the start of the file is 
    * recognised by starting the buffer with a new line (on the 1st read)
    */
   *(buf_bufp(bp)) = prevchar; /* '\n' for the first read */

   *(buf_endp(bp)) = 'F';      /* Optimization for searching */

   bp->nextpos = tell(fd);
   bp->eof = eof(fd);

#ifdef DEBUG
   fprintf(stderr, "eof=%d count=%u p=%p e=%p\n", bp->eof, count, buf_curp(bp), buf_endp(bp));
#endif

   return 0;
}

int
buf_fill(int fd, buf *bp)
{
/*
   fprintf(stderr, "bmh: buf fill called\n");
*/
   if (buf_bufp(bp) == NULL)
      if (buf_malloc(bp, (size_t) atol(getrc(bufsize)), 1) == NULL) {
         fprintf(stderr, "bmh: no room for mail file buffer\n");
         return -1;
         }

   if (buf_eof(bp)) {
      fprintf(stderr, "buffill: eof set\n");
      return -1;
      }
   else if (bp->curpos == -1L)
      return buf_read(fd, bp, '\n');
   else 
      return buf_read(fd, bp, *(buf_endp(bp)-1));
}

void
buf_reset(buf *bp)
{
   if ((bp == NULL) || (bp->buffer == NULL)) {
#ifdef DEBUG
      fprintf(stderr, "bmh: buf reset called with buffer=%p\n", bp->buffer);
#endif
      return ;
      }
   else if (bp->curpos == 0L) {
      bp->curp = bp->buffer+1;
      bp->nextpos= 0L;
      }
   else {
      bp->curpos = -1L;
      bp->nextpos= 0L;
      bp->eof = 0;

      *(bp->buffer) = '\n';
      }
}

/*
 * Note that there will always be at least one 'F' at the end of
 * a buffer.
 */
char *
buf_findF(buf *bp)
{
   register uchar *p = buf_curp(bp);

   while (1) {
      /*
       * 60 to 70% of this programs life is spent executing the
       * next 2 lines of code according to the profiler.
       */
      while ( *p != 'F' )
         p++;

      if ( (p == buf_endp(bp)) || (*(p-1) == '\n'))
         break;
      else
         p++;
      }

   buf_setptr(bp, p);

   return (char *)p;
}

long
buf_offset(buf *bp, uchar *p)
{
   return bp->curpos + 1L + (long) ((unsigned)p) - ((unsigned) bp->buffer);
}
