/*************************************************************************
** ^FILE: getopts.c - implement the functions defined in <getopts.h>
**
** ^HISTORY:
**    01/16/92	Brad Appleton	<brad@ssd.csd.harris.com>	Created
***^^********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>

#include "getopts.h"

int parse_longopt(options_t *opts, int argc,
                  const char *argv[], const char **optarg);

   /* define what an options_t structure really looks like */
struct options_s {
   unsigned      optctrls : 8;  /* control settings (set of optctrl_t masks) */
   int           optind;        /* current index in argv */
   const char ** optvec;        /* vector of option-sepcifications */
   const char  * nextchar;      /* next option-character to process */
   const char  * listopt;       /* last list-option we matched */
   const char  * cmdname;       /* name of the command */
} ;

   /* return values for a keyword matching function */
typedef  enum { NO_MATCH, PARTIAL_MATCH, EXACT_MATCH } kwdmatch_t ;

   /* Get the option-char of an option-spec. */
#define OPTCHAR(optspec)  *(optspec)

   /* Get the long-option of an option-spec. */
#define LONGOPT(optspec)  \
   ((optspec)[1] && (optspec)[2] && \
    (! isspace((optspec)[2]))) ? ((optspec) + 2) : NULL

   /* Is the option-char null? */
#define isNULLOPT(optchar)  \
   ((! optchar) || isspace(optchar) || (! isprint(optchar)))

   /* Does this option require an argument? */
#define isREQUIRED(optspec)  (((optspec)[1] == ':') || ((optspec)[1] == '+'))

   /* Does this option take an optional argument? */
#define isOPTIONAL(optspec)  (((optspec)[1] == '?') || ((optspec)[1] == '*'))

   /* Does this option take no arguments? */
#define isNOARG(optspec)     (((optspec)[1] == '|') || (! (optspec)[1]))

   /* Can this option take more than one argument? */
#define isLIST(optspec)      (((optspec)[1] == '+') || ((optspec)[1] == '*'))

   /* Does this option take any arguments? */
#define isVALTAKEN(optspec)  (isREQUIRED(optspec) || isOPTIONAL(optspec))

   /* Check for explicit "end-of-options" */
#define isENDOPTS(token)  ((! token) || (! strcmp(token, "--")))

   /* See if an argument is an option */
#define isOPTION(flags,arg) \
   (((*arg != '\0') || (arg[1] != '\0')) && \
    ((*arg == '-')  || ((flags & opt_PLUS) && (*arg == '+'))))

   /* These next three macros help manage the `optind' field of an options_t
   ** structure as the current element of an iterator for an argc/argv pair.
   **
   **   isENDARGS() determines if we are at the end-of-arguments
   **   CURR() returns the current iterm (without advancing the iterator)
   **   NEXT() advances to the next item in the iterator.
   */
#define isENDARGS(opts,ac,av)  ((opts->optind == ac) || (! av[opts->optind]))
#define CURR(opts,ac,av)  isENDARGS(opts,ac,av) ? NULL : av[opts->optind]
#define NEXT(opts,ac,av)  if (! isENDARGS(opts,ac,av)) ++(opts->optind)

/********
** ^FUNCTION: verify - verify the syntax of each element of the option vector
**
** ^SYNOPSIS:
**    static void verify(optv[])
**
** ^PARAMETERS:
**    char * optv[] - the vector of option-specs to inspect.
**
** ^DESCRIPTION:
**    All we have to do is iterate through the option vector and make sure
**    That each option-spec is of the proper format.
**
** ^REQUIREMENTS:
**    - optv[] should be non-NULL and terminated by a NULL pointer.
**
** ^SIDE-EFFECTS:
**    If an invalid option-spec is found, prints a message on stderr and
**    exits with a status of 127.
**
** ^RETURN-VALUE:
**    None.
**
** ^ALGORITHM:
**    For each option-spec
**       - ensure (length > 0)
**       - verify the the second character is one of "|?:*+"
**    end-for
***^^***/
static void
verify(const char * optv[])
{
   int errors = 0;
   if ((optv == NULL) || (*optv == NULL))  return;

   for (; *optv != NULL; optv++) {
      if (! **optv) {
         fprintf(stderr, "getopts(3): invalid option spec \"%s\".\n", *optv);
         fprintf(stderr, "\tmust be at least 1 character long.\n");
         ++errors;
      }
      if ((**optv) && ((*optv)[1]) &&
          (strchr("|?:*+", (*optv)[1]) == NULL)) {
         fprintf(stderr, "getopts(3): invalid option spec \"%s\".\n", *optv);
         fprintf(stderr, "\t2nd character must be in the set \"|?:*+\".\n");
         ++errors;
      }
   }/*for*/

   if (errors)  exit(127);
}

/********
** ^FUNCTION: kwdmatch - match a keyword
**
** ^SYNOPSIS:
**    static kwdmatch_t kwdmatch(src, attempt, len)
**
** ^PARAMETERS:
**    char * src -- the actual keyword to match
**    char * attempt -- the possible keyword to compare against "src"
**    int len -- number of character of "attempt" to consider
**               (if negative then we should use all of "attempt")
**
** ^DESCRIPTION:
**    See if "attempt" matches some prefix of "src" (case insensitive).
**
** ^REQUIREMENTS:
**    - tolower() should NOT modify a non-uppercase character.
**
** ^SIDE-EFFECTS:
**    None.
**
** ^RETURN-VALUE:
**    An enumeration value of type kwdmatch_t corresponding to whether
**    We had an exact match, a partial match, or no match.
**
** ^ALGORITHM:
**    Trivial
***^^***/
static kwdmatch_t
kwdmatch(const char * src, const char * attempt, int len)
{
   unsigned  i;

   if (src == attempt)  return  EXACT_MATCH ;
   if ((! src) || (! attempt))  return  NO_MATCH ;
   if ((! *src) && (! *attempt))  return  EXACT_MATCH ;
   if ((! *src) || (! *attempt))  return  NO_MATCH ;

   for (i = 0 ; ((i < len) || (len < 0)) && (attempt[i]) ; i++) {
      if (tolower(src[i]) != tolower(attempt[i]))  return  NO_MATCH ;
   }

   return  (src[i]) ? PARTIAL_MATCH : EXACT_MATCH ;
}

/********
** ^FUNCTION: match_opt - match an option
**
** ^SYNOPSIS:
**    static const char * match_opt(optv[], opt)
**
** ^PARAMETERS:
**    char * optv[] -- vector of option-specifications
**    char opt -- the option-character to match
**
** ^DESCRIPTION:
**    See if "opt" is found in "optv"
**
** ^REQUIREMENTS:
**    - optv should be non-NULL and terminated by a NULL pointer.
**    - tolower() should NOT modify a non-uppercase character.
**
** ^SIDE-EFFECTS:
**    None.
**
** ^RETURN-VALUE:
**    NULL if no match is found,
**    otherwise a pointer to the matching option-spec.
**
** ^ALGORITHM:
**    foreach option-spec
**       - see if "opt" is a match, if so return option-spec
**    end-for
***^^***/
static const char *
match_opt(const char * optv[], char opt)
{
   if ((! optv) || (! *optv))  return  NULL;

   for (; *optv ; ++optv) {
     char optchar = OPTCHAR(*optv);
     if (isNULLOPT(optchar))  continue;
     if (opt == optchar) {
        return *optv;
     } 
   }

   return  NULL;  /* not found */
}

/********
** ^FUNCTION: match_longopt - match a long-option
**
** ^SYNOPSIS:
**   static const char * match_longopt(optv, opt, len, ambiguous)
**
** ^PARAMETERS:
**    char * optv[] -- the vector of option-specs
**    char * opt -- the long-option to match
**    int len -- the number of character of "opt" to match
**    int * ambiguous -- set by this routine before returning.
**
** ^DESCRIPTION:
**    Try to match "opt" against some unique prefix of a long-option
**    (case insensitive).
**
** ^REQUIREMENTS:
**    - opts->optvec should be non-NULL and terminated by a NULL pointer.
**
** ^SIDE-EFFECTS:
**    - *ambiguous is set to '1' if "opt" matches >1 long-option
**      (otherwise it is set to 0).
**
** ^RETURN-VALUE:
**    NULL if no match is found,
**    otherwise a pointer to the matching option-spec.
**
** ^ALGORITHM:
**    ambiguous is FALSE
**    foreach option-spec
**       if we have an EXACT-MATCH, return the option-spec
**       if we have a partial-match then
**          if we already had a previous partial match then 
**             set ambiguous = TRUE and retrun NULL
**          else
**             remember this options spec and continue matching
**          end-if
**       end-if
**    end-for
**    if we had exactly 1 partial match return it, else return NULL
***^^***/
static const char *
match_longopt(const char * optv[], const char * opt, int len, int * ambiguous)
{
   kwdmatch_t  result;
   const char * matched = NULL ;

   *ambiguous = 0;
   if ((! optv) || (! *optv))  return  NULL;

   for (; *optv ; ++optv) {
      const char * longopt = LONGOPT(*optv) ;
      if (! longopt)  continue;
      result = kwdmatch(longopt, opt, len);
      if (result == EXACT_MATCH) {
         return  *optv;
      } else if (result == PARTIAL_MATCH) {
         if (matched) {
            ++(*ambiguous);
            return  NULL;
         } else {
            matched = *optv;
         }
      }
   }/*for*/

   return  matched;
}

/********
** ^FUNCTION: parse_opt - parse an option
**
** ^SYNOPSIS:
**    int parse_opt(opts, argc, argv, optarg)
**
** ^PARAMETERS:
**    options_t * opts -- the options structure
**    int  argc -- argc from main
**    const char * argv[] -- argv[] from main
**    const char ** optarg -- where to store any option-argument
**
** ^DESCRIPTION:
**    Parse the next option in argv (advancing optind as necessary).
**    Make sure we update the opts->nextchar pointer along the way. Any option
**    we find should be returned and optarg should point to its argument.
**
** ^REQUIREMENTS:
**    - opts->nextchar must point to the prospective option character
**
** ^SIDE-EFFECTS:
**    - opts->optind is advanced when an argument completely parsed
**    - optarg is modified to point to any option argument
**    - if opt_QUIET is not set, error messages are printed on stderr
**
** ^RETURN-VALUE:
**    'c' if the -c option was matched (optarg points to its argument)
**    -1 if the option is invalid (optarg points to the bad option-character).
**
** ^ALGORITHM:
**    It gets complicated -- follow the comments in the source.
***^^***/
int
parse_opt(options_t    * opts,
          int            argc,
          const char   * argv[],
          const char  ** optarg)
{
   const char * optspec = NULL, * nextarg = NULL;

   opts->listopt = NULL;  /* reset the list pointer */

   if ((! opts->optvec) || (! *(opts->optvec)))  return  NULL;

      /* Try to match a known option */
   optspec = match_opt(opts->optvec, *(opts->nextchar++));

      /* Check for an unknown option */
   if (! optspec) {
      /* See if this was a long-option in disguise */
      if (! (opts->optctrls & opt_NOGUESSING)) {
         int  optchar;
         unsigned  save_ctrls = opts->optctrls;
         const char * save_nextchar = opts->nextchar;
         opts->nextchar -= 1;
         opts->optctrls |= (opt_QUIET | opt_NOGUESSING);
         optchar = parse_longopt(opts, argc, argv, optarg);
         opts->optctrls = save_ctrls;
         if (optchar > 0) {
            return  optchar;
         } else {
            opts->nextchar = save_nextchar;
         }
      }
      if (! (opts->optctrls & opt_QUIET)) {
         fprintf(stderr, "%s: unknown option -%c.\n",
                         opts->cmdname, *(opts->nextchar - 1));
      }
      *optarg = (opts->nextchar - 1);  /* record the bad option in optarg */
      return  opt_BADCHAR;
   }

      /* If no argument is taken, then leave now */
   if (isNOARG(optspec)) {
      optarg = NULL;
      return  OPTCHAR(optspec);
   }

      /* Check for argument in this arg */
   if (*(opts->nextchar)) {
      *optarg = opts->nextchar;  /* the argument is right here */
      opts->nextchar = NULL;     /* we've exhausted this argument */
      if (isLIST(optspec))  opts->listopt = optspec ;  /* save the list-spec */
      return  OPTCHAR(optspec);
   }

      /* Check for argument in next arg */
   nextarg = CURR(opts, argc, argv);
   if ((nextarg != NULL) && (! isOPTION(opts->optctrls, nextarg))) {
      *optarg = nextarg;       /* the argument is here */
      NEXT(opts, argc, argv);  /* end of arg - advance */
      if (isLIST(optspec))  opts->listopt = optspec ;  /* save the list-spec */
      return  OPTCHAR(optspec);
   }

     /* No argument given - if its required, thats an error */
   *optarg = NULL;
   if (isREQUIRED(optspec) &&  !(opts->optctrls & opt_QUIET)) {
      fprintf(stderr, "%s: argument required for -%c option.\n",
                      opts->cmdname, OPTCHAR(optspec));
   }
   return  OPTCHAR(optspec);
}

/********
** ^FUNCTION: parse_longopt - parse a long-option
**
** ^SYNOPSIS:
**    int parse_longopt(opts, argc, argv, optarg)
**
** ^PARAMETERS:
**    options_t * opts -- the options structure
**    int  argc -- argc from main
**    const char * argv[] -- argv[] from main
**    const char ** optarg -- where to store any option-argument
**
** ^DESCRIPTION:
**    Parse the next long-option in argv (advancing optind as necessary).
**    Make sure we update the opts->nextchar pointer along the way. Any option
**    we find should be returned and optarg should point to its argument.
**
** ^REQUIREMENTS:
**    - opts->nextchar must point to the prospective option character
**
** ^SIDE-EFFECTS:
**    - optind is advanced when an argument completely parsed
**    - optarg is modified to point to any option argument
**    - if opt_QUIET is not set, error messages are printed on stderr
**
** ^RETURN-VALUE:
**    'c' if the the long-option corresponding to the -c option was matched
**         (optarg points to its argument)
**    -2 if the option is invalid (optarg points to the bad long-option name).
**
** ^ALGORITHM:
**    It gets complicated -- follow the comments in the source.
***^^***/
int
parse_longopt(options_t    * opts,
              int            argc,
              const char   * argv[],
              const char  ** optarg)
{
   int  len = -1, ambiguous = 0;
   const char * optspec = NULL, * val = NULL, * nextarg = NULL;

   opts->listopt = NULL ;  /* reset the list-spec */

   if ((! opts->optvec) || (! *(opts->optvec)))  return  NULL;

      /* if a value is supplied in this argv element, get it now */
   val = strpbrk(opts->nextchar, ":=") ;
   if (val) {
      len = val - opts->nextchar ;
      ++val ;
   }

      /* Try to match a known long-option */
   optspec = match_longopt(opts->optvec, opts->nextchar, len, &ambiguous);

      /* Check for an unknown long-option */
   if (! optspec) {
      /* See if this was a short-option in disguise */
      if ((! ambiguous) && (! (opts->optctrls & opt_NOGUESSING))) {
         int  optchar;
         unsigned  save_ctrls = opts->optctrls;
         const char * save_nextchar = opts->nextchar;
         opts->optctrls |= (opt_QUIET | opt_NOGUESSING);
         optchar = parse_opt(opts, argc, argv, optarg);
         opts->optctrls = save_ctrls;
         if (optchar > 0) {
            return  optchar;
         } else {
            opts->nextchar = save_nextchar;
         }
      }
      if (! (opts->optctrls & opt_QUIET)) {
         fprintf(stderr, "%s: %s option %s%s.\n",
                         opts->cmdname,
                         ((ambiguous) ? "ambiguous" : "unknown"),
                         ((opts->optctrls & opt_LONG_ONLY) ? "-" : "--"),
                         opts->nextchar);
      }
      *optarg = opts->nextchar;  /* record the bad option in optarg */
      opts->nextchar = NULL;     /* we've exhausted this argument */
      return  (ambiguous) ? opt_AMBIGUOUS : opt_BADKWD;
   }

      /* If no argument is taken, then leave now */
   if (isNOARG(optspec)) {
      if ((val) && ! (opts->optctrls & opt_QUIET)) {
         fprintf(stderr, "%s: option %s%s does NOT take an argument.\n",
                         opts->cmdname,
                         ((opts->optctrls & opt_LONG_ONLY) ? "-" : "--"),
                         LONGOPT(optspec));
      }
      *optarg = val;          /* record the unexpected argument */
      opts->nextchar = NULL;  /* we've exhausted this argument */
      return  OPTCHAR(optspec);
   }

      /* Check for argument in this arg */
   if (val) {
      *optarg = val;      /* the argument is right here */
      opts->nextchar = NULL;           /* we exhausted the rest of this arg */
      if (isLIST(optspec))  opts->listopt = optspec ;  /* save the list-spec */
      return  OPTCHAR(optspec);
   }

      /* Check for argument in next arg */
   nextarg = CURR(opts, argc, argv);  /* find next argument to parse */
   if ((nextarg != NULL) && (! isOPTION(opts->optctrls, nextarg))) {
      *optarg = nextarg;        /* the argument is right here */
      NEXT(opts, argc, argv);   /* end of arg advance optind */
      opts->nextchar = NULL;    /* we exhausted the rest of this arg */
      if (isLIST(optspec))  opts->listopt = optspec ;  /* save the list-spec */
      return  OPTCHAR(optspec);
   }

     /* No argument given - if its required, thats an error */
   optarg = NULL;
   if (isREQUIRED(optspec) &&  !(opts->optctrls & opt_QUIET)) {
      fprintf(stderr, "%s: argument required for %s%s option.\n",
              opts->cmdname,
              ((opts->optctrls & opt_LONG_ONLY) ? "-" : "--"),
              LONGOPT(optspec));
   }
   opts->nextchar = NULL;           /* we exhausted the rest of this arg */
   return  OPTCHAR(optspec);
}

/********
** ^FUNCTION: fmt_opt - format an option-spec for a usage message
**
** ^SYNOPSIS:
**    unsigned fmt_opt(ctrls, optspec, buf)
**
** ^PARAMETERS:
**    unsigned ctrls -- a combination of optctrl_t flags.
**    char * optspec -- the option-specification
**    char * buf -- where to print the formatted option
**
** ^DESCRIPTION:
**    Self-explanatory.
**
** ^REQUIREMENTS:
**    - optspec must be a valid option-spec.
**    - buf must be large enough to hold the result
**
** ^SIDE-EFFECTS:
**    - writes to buf.
**
** ^RETURN-VALUE:
**    Number of characters written to buf.
**
** ^ALGORITHM:
**    Trivial.
***^^***/
unsigned
fmt_opt(unsigned ctrls, const char * optspec, char * buf)
{
   char optchar = OPTCHAR(optspec);
   const char * longopt = LONGOPT(optspec);
   char * p = buf ;

   if ((ctrls & opt_SHORT_ONLY) &&
       ((! isNULLOPT(optchar)) || (ctrls & opt_NOGUESSING))) {
      longopt = NULL;
   }
   if ((ctrls & opt_LONG_ONLY) && (longopt || (ctrls & opt_NOGUESSING))) {
      optchar = '\0';
   }
   if (isNULLOPT(optchar) && (! longopt)) {
      *buf = '\0';
      return  0;
   }

   *(p++) = '[';

   /* print the single character option */
   if (! isNULLOPT(optchar)) {
      *(p++) = '-';
      *(p++) = optchar;
   }

   if ((! isNULLOPT(optchar)) && (longopt))  *(p++) = '|';

   /* print the long option */
   if (longopt) {
      *(p++) = '-';
      if (! (ctrls & (opt_LONG_ONLY | opt_SHORT_ONLY))) {
         *(p++) = '-';
      }
      strcpy(p, longopt);
      p += strlen(longopt);
   }

   /* print any argument the option takes */
   if (isVALTAKEN(optspec)) {
      *(p++) = ' ' ;
      if (isOPTIONAL(optspec))  *(p++) = '[' ;
      strcpy(p, "<value>");
      p += 7;
      if (isLIST(optspec)) {
         strcpy(p, " ...");
         p += 4;
      }
      if (isOPTIONAL(optspec))  *(p++) = ']' ;
   }

   *(p++) = ']';
   *p = '\0';

   return  (unsigned) strlen(buf);
}

/********
** ^FUNCTION: optusage - print usage
**
** ^SYNOPSIS:
**    void optusage(opts, positionals, fp)
**
** ^PARAMETERS:
**    const options_t * opts -- the options structure
**    char * positionals -- command-line syntax for any positional args
**    FILE * fp -- where to print the usage
**
** ^DESCRIPTION:
**    Print command-usage (using either option or long-option syntax) on fp.
**
** ^REQUIREMENTS:
**    fp should correspond to an open output file.
**
** ^SIDE-EFFECTS:
**    Prints on fp
**
** ^RETURN-VALUE:
**    None.
**
** ^ALGORITHM:
**    Print usage on fp, wrapping long lines where necessary.
***^^***/
void
optusage(const options_t * opts, const char * positionals, FILE * fp)
{
   const char ** optv = opts->optvec;
   unsigned  ll, margin, cols = 79;
   int  first, nloop;
   char  buf[256] ;

   if ((! optv) || (! *optv))  return;

      /* print first portion "usage: progname" */
   fprintf(fp, "usage: %s", opts->cmdname) ;
   ll = strlen(opts->cmdname) + 7;

      /* Save the current length so we know how much space to skip for
      ** subsequent lines.
      */
   margin = ll + 1;

      /* print the options and the positional arguments */
   for (nloop = 0, first = 1 ; !nloop ; optv++, first = 0) {
      unsigned  len;

         /* figure out how wide this parameter is (for printing) */
      if (! *optv) {
         len = strlen(positionals);
         ++nloop;  /* terminate this loop */
      } else {
         len = fmt_opt(opts->optctrls, *optv, buf);
      }

      /*  Will this fit? */
      if ((ll + len + 1) > (cols - first)) {
         fprintf(fp, "\n%*s", margin, "");   /* No - start a new line; */
         ll = margin;
      } else {
         fputc(' ', fp);  /* Yes - just throw in a space */
         ++ll;
      }
      ll += len;
      fputs(((nloop) ? positionals : buf), fp) ;
   }/* for each ad */

   fputc('\n', fp);
   fflush(fp);
}


/********
** ^FUNCTION: getopts - get options from the command-line
**
** ^SYNOPSIS:
**   int getopts(opts, argc, argv, optarg)
**
** ^PARAMETERS:
**    options_t * opts -- the options structure
**    int  argc -- argc from main
**    const char * argv[] -- argv[] from main
**    const char ** optarg -- where to store any option-argument
**
** ^DESCRIPTION:
**    Parse the next option in argv (advancing optind as necessary).
**    Make sure we update the opts->nextchar pointer along the way. Any option
**    we find should be returned and optarg should point to its argument.
**
** ^REQUIREMENTS:
**    opts should have been created by optinit().
**
** ^SIDE-EFFECTS:
**    - optind is advanced when an argument is completely parsed
**    - optarg is modified to point to any option argument
**    - if opt_QUIET is not set, error messages are printed on stderr
**
** ^RETURN-VALUE:
**     0 if all options have been parsed.
**    'c' if the the option or long-option corresponding to the -c option was
**         matched (optarg points to its argument).
**    -1 if the option is invalid (optarg points to the bad option character).
**    -2 if the option is invalid (optarg points to the bad long-option name).
**
** ^ALGORITHM:
**    It gets complicated -- follow the comments in the source.
***^^***/
int
getopts(options_t    * opts,
        int            argc,
        const char   * argv[],
        const char  ** optarg)
{
   const char * arg;

      /* See if we have an option left over from before ... */
   if ((opts->nextchar) && *(opts->nextchar)) {
      return  parse_opt(opts, argc, argv, optarg);
   }

      /* Check for end-of-options ... */
   arg = CURR(opts, argc, argv);
   if (! arg) {
      opts->listopt = NULL;
      return  opt_ENDOPTS;
   } else if (isENDOPTS(arg)) {
      NEXT(opts, argc, argv);  /* advance past end-of-options arg */
      opts->listopt = NULL;
      return  opt_ENDOPTS;
   }

      /* Do we have a positional arg? */
   if (! opts->listopt) {
      if ((! *arg) || (! arg[1])) {
         return  opt_ENDOPTS;
      } else if ((*arg != '-') &&
                 ((! (opts->optctrls & opt_PLUS)) ||
                  (*arg != '+'))) {
         return  opt_ENDOPTS;
      }
   }

   /* pass the argument that opts->nextarg already points to */
   NEXT(opts, argc, argv);

      /* See if we have a long option ... */
   if (! (opts->optctrls & opt_SHORT_ONLY)) {
      if ((*arg == '-') && (arg[1] == '-')) {
         opts->nextchar = arg + 2;
         return  parse_longopt(opts, argc, argv, optarg);
      } else if ((opts->optctrls & opt_PLUS) && (*arg == '+')) {
         opts->nextchar = arg + 1;
         return  parse_longopt(opts, argc, argv, optarg);
      }
   }
   if (*arg == '-') {
      opts->nextchar = arg + 1;
      if (opts->optctrls & opt_LONG_ONLY) {
         return  parse_longopt(opts, argc, argv, optarg);
      } else {
         return  parse_opt(opts, argc, argv, optarg);
      }
   }

      /* If we get here - it is because we have a list value */
   *optarg = arg ;          /* record the list value */
   return  OPTCHAR(opts->listopt) ;
}

/********
** ^FUNCTION: optctrl - get/set parse-controls
**
** ^SYNOPSIS:
**    extern  unsigned optctrl(opts, newflags)
**
** ^PARAMETERS:
**    options_t * opts -- the options structure
**    unsigned newflags -- combination of bitmasks defined by optctrl_t
**
** ^DESCRIPTION:
**    If newflags contains opt_QUERY, then just get the current flags.
**    Otherwise, set "opt_ctrls" to the given flags and retrun what they were.
**
** ^REQUIREMENTS:
**    None.
**
** ^SIDE-EFFECTS:
**    Modifies opts->optctrls
**
** ^RETURN-VALUE:
**    The current opt-ctrls for "opts" at the time of the call.
**
** ^ALGORITHM:
**    set the new-flags (if !opt_QUERY), return the old-ones.
***^^***/
extern  unsigned
optctrl(options_t * opts, unsigned newflags)
{
   unsigned  oldctrls = opts->optctrls;  /* save old flags */

      /* Set new flags if not a QUERY ... */
   if (! (newflags & opt_QUERY)) {
      opts->optctrls = newflags;
   }
   return  oldctrls;  /* return old flags */
}

/********
** ^FUNCTION: getoptind - get option index.
**
** ^SYNOPSIS:
**    int getoptind(opts);
**
** ^PARAMETERS:
**    const options_t * opts; -- options structure
**
** ^DESCRIPTION:
**    After getopts() has indicated that the end-of-options has been
**    encountered, the programmer may wish to know the index (in argv)
**    of the first no-option element in the array. This function provides
**    precisely this service.
**
** ^REQUIREMENTS:
**    Getopts() should already have returne '0' (meaning end-of-options)
**    for the corresponing argc/argv pair.
**
** ^SIDE-EFFECTS:
**    None.
**
** ^RETURN-VALUE:
**    Index of the first non-option element in the argv vector that was
**    repeatedly passed to getopts.
**
** ^ALGORITHM:
**    Trivial.
***^^***/
int
getoptind(const options_t * opts)  { return  opts->optind; }

/********
** ^FUNCTION: optreset - reset for parsing more options.
**
** ^SYNOPSIS:
**    optreset(opts);
**
** ^PARAMETERS:
**    options_t * opts; -- the options structure
**
** ^DESCRIPTION:
**    After having parsed all the options in an argv[] array, the programmer
**    may wish to make a second pass over argv to reparse the options (or
**    perhaps the programmer wants to parse yet-another argv vector).
**    This function performs the necessary re-initialization for such
**    "re-parsing" of program options.
**
** ^REQUIREMENTS:
**    opts should have been created using optinit!
**
** ^SIDE-EFFECTS:
**    opts is modified.
**
** ^RETURN-VALUE:
**    None.
**
** ^ALGORITHM:
**    Trivial.
***^^***/
void
optreset(options_t * opts) {
   opts->optind = 0;
   opts->nextchar = opts->listopt = NULL;
}

/********
** ^FUNCTION: optinit, optfree - create and destroy an options structure
**
** ^SYNOPSIS:
**    opts = optinit(name, optv);
**    optfree(opts);
**
** ^PARAMETERS:
**    options_t * opts;  -- the options structure
**    const char * name; -- the program name
**    const char * argv[]; -- the NULL terminated vector of option-specs
**
** ^DESCRIPTION:
**    Optinit allocates space for an options_t structure and initializes it.
**    It returns a pointer to the newly created options structure. If space
**    cannot be allocated of if one of the option-specifications is invalid
**    then a message is printed on stderr and exit(127) is called.
**
**    Optfree deallocates the storage associated with an options structure.
**
**    Optinit and optfree manage a very small buffer of statically allocated
**    options_t structures. If less than two options_t structures are in use
**    at any given time than the normal overhead associated with malloc and
**    is avoided. 
**
** ^REQUIREMENTS:
**    The options structure passed to optfree MUST have been created with
**    optinit.
**
** ^SIDE-EFFECTS:
**    for optinit() - space is allocated and initialized. Any error messages
**    are printed on stderr.
**
**    for optfree() - space is deallocated (and the caller should probably
**    set the corresponding options_t pointer to NULL upon return).
**
** ^RETURN-VALUE:
**    Optinit returns a pointer to the newly allocated options structure.
**
** ^ALGORITHM:
**    Fairly trivial.
***^^***/

/*
** We keep a small pool of options-structures here for "allocation" via
** optinit(). We assume that the user usually wont need more than two
** option-structures at a time. In such cases we try to avoid the overhead
** of malloc and free (and the need for error-checking).
*/

#define  POOL_SIZE 2

static  options_t opt_pool[POOL_SIZE] ;
   /* Our local pool of readily available option-structures.
   */
static  char opt_used[POOL_SIZE] = { '\0', '\0' } ;
   /* Indicates which "members" of our local pool are currently in use.
   ** opts_used[i] == 0 implies that opt_pool[i] is available for use.
   */

options_t *
optinit(const char * name, const char * optv[]) {
   options_t * opts;

   verify(optv);  /* make sure all the option-specs are valid */

      /* get space */
   if (! *opt_used) {
      opts = opt_pool;
      ++(*opt_used);
   } else if (! opt_used[1]) {
      opts = &(opt_pool[1]);
      ++(opt_used[1]);
   } else {
      opts = (options_t *) malloc(sizeof(options_t));
      if (! opts) {
         perror("Unable to allocate space for options-structure.");
         exit(127);
      }
   }

      /* initialize */
   opts->optctrls = opt_DEFAULT;
   opts->optind = 0;
   opts->nextchar = opts->listopt = NULL;
   opts->optvec = optv;
   opts->cmdname = strrchr(name, '/');
   opts->cmdname = (opts->cmdname == NULL) ? name : (opts->cmdname + 1);

   return  opts;
}

void
optfree(options_t * opts) {
   if (opts == NULL)  return;

   if (opts == opt_pool) {
      *opt_used = '\0';
   } else if (opts == &(opt_pool[1])) {
      opt_used[1] = '\0';
   } else {
      free(opts);
   }
}
