#ifndef _GETOPTS_H
#define _GETOPTS_H

/*
** bitmasks that may be used to control parsing behavior
*/
typedef enum {
   opt_DEFAULT    = 0x00,  /* Default setting */
   opt_QUIET      = 0x01,  /* Dont print error messages */
   opt_PLUS       = 0x02,  /* Allow "+" as a long-option prefix */
   opt_SHORT_ONLY = 0x04,  /* Dont accept long-options */
   opt_LONG_ONLY  = 0x08,  /* Dont accept short-options (also lets "-" work
                           ** as a long-option prefix).
                           */
   opt_NOGUESSING = 0x10,  /* Normally, when we see a short (long) option
                           ** on the command line that doesnt match any
                           ** known short (long) options, then we try to
                           ** "guess" by seeing if it will match any known
                           ** long (short) option. Setting this mask prevents
                           ** this "guessing" from occurring.
                           */
   opt_QUERY      = 0x20,  /* Dont set flags, just find out what they are */
} optctrl_t ;

struct options_s ;
typedef struct options_s options_t ;

#ifdef __cplusplus
extern "C" {
#endif

/*
** error return values for getopts(3C)
*/
typedef enum {
   opt_ENDOPTS    =  0,
   opt_BADCHAR    = -1,
   opt_BADKWD     = -2,
   opt_AMBIGUOUS  = -3,
} opt_rc_t ;

/*
** NAME
** ====
** getopts(3C)  -- parse command-line options
**
** SYNOPSIS
** ========
**   #include <stdio.h>
**   #include <getopts.h>
**
**   options_t * opts = optinit(*argv[0], optv);
**   int optchar = getopts(opts, argc, argv, &optarg);
**
**   int  argc;
**   char *optv[], *argv[], *optarg;
**
** DESCRIPTION
** ===========
** Get program options and present them to the programmer one at a time.
**
** Returns:
**   0 : When all options have been parsed.
**
**   'c' : If "-c" (or it's corresponding long-option) was found.
**         Any argument for the option is in optarg. If an argument
**         is not required or was not given, then optarg will be NULL.
**         If an argument was required but not given, then an error
**         message has been printed on stderr (unless opt_QUIET is set).
**
**  opt_BADCHAR :
**       If an unknown option was found. An error message has been printed
**       on stderr (unless opt_QUIET is set). The offending option character
**       is denoted by *optarg.
**
**  opt_BADKWD :
**       If an unknown long option was found. An error message has been
**       printed on stderr (unless opt_QUIET is set). The offending
**       long-option name is pointed to by optarg.
**
**  opt_AMBIGUOUS :
**       If an ambiguous long option was found. An error message has been
**       printed on stderr (unless opt_QUIET is set). The ambiguous
**       long-option name is pointed to by optarg.
**
**
** "optv[]" is a pointer to an array of strings.  The last element in this
**  array MUST be NULL. Each non-NULL string in the array must have the
**  following format:
**
**   The 1st character must be the option-name ('c' for a -c option).
**
**   The 2nd character must be one of '|', '?', ':', '*', or '+'.
**      '|' -- indicates that the option takes NO argument;
**      '?' -- indicates that the option takes an OPTIONAL argument;
**      ':' -- indicates that the option takes a REQUIRED argument;
**      '*' -- indicates that the option takes 0 or more arguments;
**      '+' -- indicates that the option takes 1 or more arguments;
**
**   The remainder of the string must be the long-option name.
**
**   One may use a space to indicate that a particular option does not
**   have a corresponding long-option.  For example, "c: " (or "c:")
**   means the -c option takes a value & has NO corresponding long-option.
**
**   To specify a long-option that has no corresponding single-character
**   option is a bit trickier: getopts() still needs an "option-character"
**   to return when that option is matched. One may use a whitespace
**   character or a non-printable character as the single-character option
**   in such a case. (hence " |hello" would only match "--hello").
**
**   EXCEPTIONS TO THE ABOVE:
**   ------------------------ 
**   If the 2nd character of the string is '\0' then it is assumed
**   that there is no corresponding long-option and that the option
**   takes no argument (hence "f", and "f| " are equivalent).
**
**   Examples:
**      const char * optv[] = {
**          "c:count",
**          "s?string",
**          "x",
**          " |hello",
**          "g+groups",
**          NULL
**      } ;
**
**      optv[] now corresponds to the following:
**
**            usage: progname [-c|--count value] [-s|--string [value]]
**                            [-x] [--hello] [-g|--groups value ...]
**
** Long-option names are matched case-insensitive and only a unique prefix
** of the name needs to be specified.
**
** Option-name characters are case-sensitive!
**
** CAVEAT
** ======
** Because of the way in which multi-valued options and options with optional
** values are handled, it is NOT possible to supply a value to an option in
** a separate argument (different argv[] element) if the value begins with
** a '-'. What this means is that if an option "-s" requires a value and
** you wish to supply a value of "-foo" then you must specify this on the
** command-line as "-s-foo" instead of "-s -foo" because "-s -foo" will be
** considered to be two separate sets of options.
**
** A multi-valued option is terminated bu another option or by the end-of
** options. The following are all equivalent (if "-l" is a multi-valued
** option and "-x" is an option that takes no value):
**
**    progname -x -l item1 item2 item3 -- arg1 arg2 arg3
**    progname -x -litem1 -litem2 -litem3 -- arg1 arg2 arg3
**    progname -l item1 item2 item3 -x arg1 arg2 arg3
**
**
** EXAMPLE
** =======
**    #include <stdio.h>
**    #include <getopts.h>
**
**    static const char * optv[] = {
**       "c:count", "s?string", "x", " |hello", "g+groups", NULL
**    } ;
**
**    main(int argc, char * argv[]) {
**       char * options_t = optinit(*argv, optv);
**       char * optarg;
**       int i, optchar, errors = 0;
**       int cflag = 1, xflag = 0, hello = 0, ngroups = 0;
**       char * sflag = "default" ;
**
**       --argc, ++argv;
**       while( optchar = getopts(opts, argc, argv, &optarg) ) {
**          switch(optchar) {
**          case 'c' :
**             if (optarg == NULL) ++errors;
**             else  cflag = (int) atol(optarg);
**             break;
**
**          case 's' : sflag = optarg; break;
**
**          case 'x' : ++xflag; break;
**
**          case ' ' : ++hello; break;
**
**          case 'g' : ++ngroups; break;    <---- the group-name is in "optarg"
**
**          default : ++errors; break;
**       }
**
**       if (errors) {
**          optusage(opts, "any positional arguments", stderr);
**          exit(1);
**       }
**
**       for (i = getoptind(opts) ; i < argc ; i++ ) {
**          ... do whatever with argv[i] ...
**       }
**
**       optfree(opts);
**    }
*/

extern  options_t *
optinit(const char * name, const char * optv[]);

extern  int
getopts(options_t * opts, int argc, const char * argv[], const char ** optarg);

extern  int
getoptind(const options_t * opts);

extern  void
optfree(options_t * opts);

/*
** Reset optind for another pass.
*/
extern  void
optreset(options_t * opts);

/*
** Print "usage: progname" + "[-options]" on fp (dont print trailing newline)
*/
extern void
optusage(const options_t * opts, const char * positonals, FILE * fp); 

/*
** Return what the controls were, then set them to the given flags. See the
** definition of optctrl_t for a list of the control masks and what they do.
*/
extern  unsigned
optctrl(options_t * opts, unsigned ctrlflags);

#ifdef __cplusplus
}
#endif

#endif /* _GETOPTS_H */
