/*  Driver for HAPN-1 8273 card
 *  Jon Bloom, KE3Z; adapted from KA9Q's PC-100 driver
 */

#include "machdep.h"
#include "mbuf.h"
#include "iface.h"
#include "hapn.h"
#include "ax25.h"

#ifdef TRACE
#include <stdio.h>
#include "trace.h"
int ax25_dump();
#endif

struct hapn hapn[NHAPN];
int ha0vec();
int (*h_handle[])() = { ha0vec };
int16 nhapn;
unsigned char inportb();

/*  send command to the 8273
 *  "base" = base port of 8273
 *  "cmd"  = command byte
 *  "np"   = number of parameter bytes
 *  "p1"   = first parameter (parameters are int)
 */
/*VARARGS3*/
static
cmd_8273(base, cmd, np, p1)
int16 base;
int cmd, np, p1;
{
        int *p;

        while (inport(base+STA) & CBSY)
                ;
        outportb(base+CMD, cmd);
        p = &p1;
        while (np--) {
                while (inport(base+STA) & CPBF)
                        ;
                outportb(base+PAR, *p++);
        }
}

/*  Start receiver of 8273 */
static
hrxgo(hp)
register struct hapn *hp;
{
        cmd_8273(hp->base, GENERAL_RX, 2, hp->bufsiz & 0xff, hp->bufsiz >> 8);
}

/*  interrupt service function.  Entered with hapn index */
haint(dev)
int dev;
{
        register struct hapn *hp;
        register int16 base;
        char isav;

        isav = disable();

        hp = &hapn[dev];
        base = hp->base;

        /*  Check for TX interrupt  */

        if (inport(base+STA) & TXINT)
                htxint(hp);

        /*  Check for RX interrupt  */

        if (inport(base+STA) & RXINT)
                hrxint(hp);

        restore(isav);
}

/*  RX interrupt service
 *  if status register bit "RXIRA" is set, interrupt is final,
 *  otherwise, interrupt is data request
 */
hrxint(hp)
register struct hapn *hp;
{
        register struct mbuf *bp;
        register int16 base;
        char isav;
        char results[10];
        char c;
        int i;

        isav = disable();
        hp->rxints++;
        base = hp->base;

        if (inport(base+STA) & RXIRA) {                /* RX result interrupt */
                i = 0;

                /*  Read results  */
                while ((c = inportb(base+STA)) & RXINT) {
                        if (c & RXIRA)
                                results[i++] = inportb(base+RXI);
                }
                if ((c = results[0] & 0x1f) == A1_MATCH || c == A2_MATCH) {
                        /*  Frame received.  Enqueue it if no residue  */
                        if (hp->rcvbuf != NULLBUF && (results[0] & 0xe0) == 0xe0) {
                                enqueue(&hp->rcvq, hp->rcvbuf);
                                hp->rcvbuf = NULLBUF;
                                hp->rcvcnt++;
                        }
                } else {                        /*  Error termination  */
                        switch (c & 0x1f) {
                        case CRCERR:
                                hp->crcerr++;
                                break;
                        case ABORT_DET:
                                hp->aborts++;
                                break;
                        case MEM_OVFL:
                                hp->toobig++;
                                break;
                        }
                        /*  If terminating condition disables rx, restart it */
                        if (c != CRCERR && c != ABORT_DET && c != SHORT_FRM)
                                hrxgo(hp);
                }
                /*  Throw rx buffer contents away to start over  */
                if (hp->rcvbuf != NULLBUF) {
                        hp->rcp = hp->rcvbuf->data;
                        hp->rcvbuf->cnt = 0;
                }
        } else {        /* RX data interrupt */
                /*  Allocate new rx buffer if none present  */
                if ((bp = hp->rcvbuf) == NULLBUF) {
                        bp = hp->rcvbuf = alloc_mbuf(hp->bufsiz);
                        if (bp == NULLBUF) {        /* No memory available */
                                hp->nomem++;
                                cmd_8273(base, RX_DISABLE, 0);
                                hrxgo(hp);
                                goto ret;
                        }
                        /*  Init buffer pointer  */
                        hp->rcp = hp->rcvbuf->data;
                }
                /*  Barf if rx data is more than buffer can hold (should never
                 *  happen since 8273 is also counting bytes).
                 */ 
                if (bp->cnt++ >= hp->bufsiz) {
                        hp->toobig++;
                        cmd_8273(base, RX_DISABLE, 0);
                        hrxgo(hp);
                        free_p(bp);
                        hp->rcvbuf = NULLBUF;
                        goto ret;
                }
                /*  Store the received byte  */
                *hp->rcp++ = inportb(base+RXD);
        }
ret:        restore(isav);
}

/*  test for busy channel (CD active)
 *  returns TRUE if channel busy
 */
static
hcdchk(base)
int16 base;
{
        cmd_8273(base, READ_A, 0);
        while (!inportb(base+STA) & CRBF)
                ;
        return ((inportb(base+RES) & CD) != 0);
}

/*  TX interrupt service
 *  if status register bit "TXIRA" is set, interrupt is final,
 *  otherwise, interrupt is data request
 */
htxint(hp)
register struct hapn *hp;
{
        char isav;
        register int16 base;
        int16 len;
        char c;

        isav = disable();
        hp->txints++;
        base = hp->base;

        c = 0;
        if (inport(base+STA) & TXIRA) {                /* TX result interupt */
                hp->tstate = IDLE;
                free_p(hp->sndbuf);
                hp->sndbuf = NULLBUF;

                /*  Read result  */
                while ((inport(base+STA) & (TXINT | TXIRA)) != (TXINT | TXIRA))
                        ;
                c = inportb(base+TXI);

                /*  Test for tx abort  */
                switch (c & 0x1f) {
                case DMA_URUN:
                        hp->t_urun++;
                        break;
                case CTS_LOSS:
                        hp->ctsloss++;
                        break;
                case ABORT_CMPLT:
                        hp->taborts++;
                        break;
                }
        }
        switch (hp->tstate) {
        case IDLE:        /*  See if a buffer is ready to be sent  */
                if ((hp->sndbuf = dequeue(&hp->sndq)) == NULLBUF)
                        break;

        case DEFER:        /*  Busy-channel check  */
                if (hp->mode == CSMA && (c & 0x1f) != EARLY_TXI){
                        if (hcdchk(base)) {
                                hp->tstate = DEFER;
                                break;
                        }
                }
                /*  Start transmitter  */
                len = len_mbuf(hp->sndbuf);
                cmd_8273(base, TX_FRAME, 2, len & 0xff, len >> 8);
                hp->tstate = ACTIVE;
                hp->tframes++;
                break;
        case ACTIVE:        /*  Get next byte to send  */
                if (pullup(&hp->sndbuf, &c, 1) != 1) {
                        cmd_8273(base, ABORT_TXF, 0);
                        hp->tstate = IDLE;
                } else
                        outportb(base+TXD, c);
                break;
        }
        restore(isav);
}

/*  Attach a HAPN adaptor to the system
 *  argv[0]:  hardware type, must be "hapn"
 *  argv[1]:  I/O address, e.g. "0x310"
 *  argv[2]:  vector, e.g. "2"
 *  argv[3]:  mode, must be "ax25"
 *  argv[4]:  interface name, e.g. "ha0"
 *  argv[5]:  rx packet buffer size in bytes
 *  argv[6]:  maximum transmission unit in bytes
 *  argv[7]:  channel-access mechanism, "csma" or "full"
 */
hapn_attach(argc, argv)
int argc;
char *argv[];
{
        register struct interface *if_h;
        extern struct interface *ifaces;
        struct hapn *hp;
        int dev, i;
        char *malloc(),*calloc();
        char isav;
        int hapn_init(), dohapn(), hapn_stop(), ax_send(), ax_output(),
                hapn_raw();
        long getvec();
        unsigned short getcs();
        static struct {
                char *str;
                char type;
        } ch_access [] = { "csma", 0, "full", 1 };

        if (nhapn >= NHAPN) {
                printf("Too many HAPN adaptors\r\l");
                return -1;
        }
        dev = nhapn++;

        /*  Initialize hardware constants */
        hapn[dev].base = htoi(argv[1]);
        hapn[dev].vec = htoi(argv[2]);

        /*  Save original interrupt vector  */
        hapn[dev].oldvec = getvec(8 + hapn[dev].vec);

        /*  Set new interrupt vector  */
        setvec(8 + hapn[dev].vec, getcs(), (unsigned) h_handle[dev]);

        /*  Create new interface structure  */
        if_h = (struct interface *) calloc(1,sizeof (struct interface));

        /*  Fill interface structure  */
        if_h->name = malloc((unsigned) strlen(argv[4]) +1);
        strcpy(if_h->name, argv[4]);
        if_h->mtu = atoi(argv[6]);
        if_h->dev = dev;
        if_h->recv = dohapn;
        if_h->stop = hapn_stop;
        if_h->output = ax_output;
        if_h->raw = hapn_raw;

        if (strcmp(argv[3], "ax25")) {
                printf("Mode %s unknown for interface %s\r\l", argv[3], argv[4]);
                free(if_h->name);
                free((char *) if_h);
                return -1;
        }
        if(mycall.call[0] == '\0'){
                printf("set mycall first\r\l");
                free(if_h->name);
                free((char *) if_h);
                return -1;
        }                
        if_h->send = ax_send;
        if_h->flags = IF_BROADCAST;
        if(if_h->hwaddr == NULLCHAR)
                if_h->hwaddr = malloc(AXALEN);
        bcopy((char *)&mycall,if_h->hwaddr,AXALEN);

        /*  Link the interface into the interface list  */
        if_h->next = ifaces;
        ifaces = if_h;

        /*  Fill the local data structure  */
        hp = &hapn[dev];
        hp->bufsiz = atoi(argv[5]);
        for (i = 0; i < (sizeof ch_access / sizeof ch_access[0]); i++)
                if (!strcmp(argv[7], ch_access[i].str))
                        hp->mode = ch_access[i].type;

        /*  Initialize the hardware  */
        isav = disable();
        hapn_init(hp);

        /*  Enable the interrupt  */
        clrbit(INTMASK, (char) (1 << hapn[dev].vec));

        restore(isav);
        return 0;
}

/*  initialize the HAPN adaptor */
hapn_init(hp)
register struct hapn *hp;
{
        register int16 base;
        char isav;

        isav = disable();
        base = hp->base;

        /*  Reset the 8273 */
        outportb(base+RST, 1);
        outportb(base+RST, 0);
        inportb(base+TXI);                /* Clear any old IR contents */
        inportb(base+RXI);

        /*  Select the operating modes  */
        cmd_8273(base, SET_XFER, 1, 1);
        cmd_8273(base, SET_MODE, 1, HDLC | EARLY | PREFRM | FLG_STM);
        cmd_8273(base, SET_SERIAL, 1, NRZI);
        cmd_8273(base, SET_B, 1, IRQ_ENB | RTS);
        cmd_8273(base, RST_B, 1, 0xff ^ RTS);
        hrxgo(hp);
        restore(isav);
}

/*  shut down the HAPN adaptor */
hapn_stop(iface)
struct interface *iface;
{
        int16 dev;
        int16 base;
        struct hapn *hp;

        dev = iface->dev;
        hp = &hapn[dev];
        base = hp->base;

        /*  Mask off interrupt input  */
        setbit(INTMASK, (char) (1 << hp->vec));

        /*  Restore original interrupt vector  */
        setvec(8 + hp->vec, (unsigned) (hp->oldvec >> 16),
                (unsigned) (hp->oldvec & 0xffff));

        /*  Reset the 8273  */
        outportb(base+RST, 1);
        outportb(base+RST, 0);
}

/* Display adaptor statistics */
dohapnstat()
{
        struct hapn *hp;
        int i;

        if (nhapn == 0) {
                printf("No HAPN adaptor attached\r\l");
                return;
        }
        for (i = 0; i < nhapn; i++) {
                hp = &hapn[i];
                printf("HAPN %d:   rxints: %ld   txints: %ld\r\l", i,
                        hp->rxints,hp->txints);
                printf(" receive  - crcerrs: %-5d  aborts: %-5d  nomem: %-5d  toobig: %-5d\r\l",
                        hp->crcerr, hp->aborts, hp->nomem, hp->toobig);
                printf(" transmit - frames: %-5d   aborts: %-5d  uruns: %-5d  ctsloss: %-5d\r\l",
                        hp->tframes,hp->taborts, hp->t_urun, hp->ctsloss);
        }
}

/*  periodically kicked by mainline routine
 *  process any queued received frames
 *  kick tx if waiting on busy channel
 */
dohapn(iface)
struct interface *iface;
{
        struct hapn *hp;
        struct mbuf *bp;

        hp = &hapn[iface->dev];

        /*  Process any received frames  */
        while ((bp = dequeue(&hp->rcvq)) != NULLBUF) {
                hp->rcvcnt--;
                ax_recv(iface, bp);
        }

        /*  Test for deferred transmit (CSMA)  */
        if (hp->tstate == DEFER)
                htxint(hp);
}

/* Send raw packet on HAPN interface */
hapn_raw(interface,bp)
struct interface *interface;
struct mbuf *bp;
{
        struct hapn *hp;

        hp = &hapn[interface->dev];

#ifdef TRACE
        if (trace & TRACE_SDLC) {                /* Debuggy stuff */
                printf("STA: %02x\n\r", inportb(hp->base+STA));
        }
#endif
        /*  See if anything being transmitted  */
        enqueue(&hp->sndq, bp);
        if (hp->tstate == IDLE)
                htxint(hp);
}

