/* OS- and machine-dependent stuff for the 8250 asynch chip on a IBM-PC
 * Copyright 1991 Phil Karn, KA9Q
 *
 * 16550A support plus some statistics added mah@hpuviea.at 15/7/89
 *
 * CTS hardware flow control from dkstevens@ucdavis,
 * additional stats from delaroca@oac.ucla.edu added by karn 4/17/90
 *
 * Sep '91      Bill Simpson@um.cc.umich.edu
 *		all control signals re-organized
 * Jan '92      Bill Simpson
 *		re-organized Iface device and encapsulation scheme
 */
#include <stdio.h>
#include <dos.h>
#include "global.h"
#include "mbuf.h"
#include "proc.h"
#include "iface.h"
#include "i8250.h"
#include "asy.h"
#include "nrs.h"
#include "ppp.h"
#include "slip.h"
#include "devparam.h"
#include "pc.h"
#include "commands.h"

static void asy_monitor __ARGS((int dev,void *p1,void *p2));

static void asy_output __ARGS((int dev,char *buf,unsigned short cnt));
static int asyrxint __ARGS((struct asy *asyp));
static void asytxint __ARGS((int dev));
static void asymsint __ARGS((int dev));

static void asy_tx __ARGS((int dev,void *p1,void *p2));

struct asy Asy[ASY_MAX];

/* ASY interrupt handlers */
static INTERRUPT (*Handle[])() = {asy0vec,asy1vec,asy2vec,asy3vec,asy4vec};


/* Attach a serial interface to the system
 * argv[0]: hardware type, must be "asy"
 * argv[1]: I/O address, e.g., "0x3f8"
 * argv[2]: vector, e.g., "4"
 * argv[3]: mode, may be:
 *		"ax25" (AX.25 frame format in SLIP for raw TNC)
 *		"nrs" (NET/ROM format serial protocol)
 *		"ppp" (Point-to-Point Protocol)
 *		"slip" (point-to-point SLIP)
 * argv[4]: interface label, e.g., "sl0"
 * argv[5]: receiver ring buffer size in bytes
 * argv[6]: maximum transmission unit, bytes
 * argv[7]: interface speed, e.g, "9600"
 * argv[8]: optional flags,
 *		'v' Van Jacobson TCP header compression (SLIP only,
 *		    use ppp command for VJ compression with PPP)
 */
int
asy_attach(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	register struct iface *ifp;
	int dev;
	int trigchar = -1;
	char blind = FALSE;
	char monitor = FALSE;
	register unsigned base;
	register struct fifo *fp;
	register struct asy *ap;
	char *ifn;
	char i_state;

	/* Find unused asy control block */
	for(dev=0;dev < ASY_MAX;dev++){
		ap = &Asy[dev];
		if(ap->iface == NULLIF)
			break;
	}
	if(dev >= ASY_MAX){
		tprintf("Too many asynch controllers\n");
		return -1;
	}

	/* Create interface structure and fill in details */
	if ( (ifp = if_attach( argv[4] )) == NULLIF ) {
		tprintf("Interface %s already exists\n",argv[4]);
		return -1;
	}
	ifp->ddv = ap;
	ifp->d_index = dev;
	ifp->d_type = IFD_COM_PORT;
	ifp->ioget = get_asy;
	ifp->iosend = asy_send;
	ifp->ioctl = asy_ioctl;
	ifp->iostop = asy_stop;

	ifp->mtu = atoi(argv[6]);

	if(stricmp(argv[3],"AX25") == 0) {
		/* Set up a SLIP link to use AX.25 */
		setencap(ifp,"KISS");
		trigchar = FR_END;
		blind = TRUE;

	} else if(stricmp(argv[3],"NRS") == 0) {
		/* Set up a net/rom serial iface */
		setencap(ifp,"NRS");
		trigchar = ETX;
		blind = TRUE;

	} else if(stricmp(argv[3],"PPP") == 0) {
		/* Point-to-Point Protocol */
		setencap(ifp,"PPP");
		trigchar = HDLC_FLAG;
		monitor = TRUE;

	} else if(stricmp(argv[3],"SLIP") == 0) {
		/* Serial Line IP encapsulation */
                if((argc > 8) && (strchr(argv[8],'v') != NULLCHAR)) {
			ifp->flags = SLIP_VJWANT | SLIP_VJWILL;
                }

		setencap(ifp,"SLIP");
		trigchar = FR_END;
	} else {
		tprintf("Mode %s unknown for interface %s\n",
			argv[3],argv[4]);
		if_detach(ifp);
		return -1;
	}

	if((argc > 8) && (strchr(argv[8],'b') != NULLCHAR)) {
		blind = TRUE;
	}

	/* Initialize asynch port "dev" */
	ap->iface = ifp;
	ap->addr = base = htoi(argv[1]);
	ap->vec = htoi(argv[2]);
	ap->trigchar = trigchar;

	/* Set up receiver FIFO */
	fp = &ap->fifo;
	fp->bufsize = (unsigned)atol(argv[5]);
	free(fp->buf);			/* just in case */
	fp->buf = mallocw(fp->bufsize);
	fp->wp = fp->rp = fp->buf;
	fp->cnt = 0;
	fp->hiwat = 0;
	fp->overrun = 0;

	/* Purge the receive data buffer */
	(void)inportb(base+RBR);

	i_state = dirps();

	/* Save original interrupt vector, mask state, control bits */
	ap->save.vec = getirq(ap->vec);
	ap->save.mask = getmask(ap->vec);
	ap->save.lcr = inportb(base+LCR);
	ap->save.ier = inportb(base+IER);
	ap->save.mcr = inportb(base+MCR);
	ap->save.msr = inportb(base+MSR);

	/* save speed bytes */
	setbit(base+LCR,LCR_DLAB);
	ap->save.divl = inportb(base+DLL);
	ap->save.divh = inportb(base+DLM);
	clrbit(base+LCR,LCR_DLAB);

	/* save some information on the starting state */
	ap->dtrState = (ap->save.mcr & MCR_DTR) ? rs232ON : rs232OFF;
	ap->rtsState = (ap->save.mcr & MCR_RTS) ? rs232ON : rs232OFF;
	ap->dsrState = (ap->save.msr & MSR_DSR) ? rs232ON : rs232OFF;
	ap->ctsState = (ap->save.msr & MSR_CTS) ? rs232ON : rs232OFF;
	ap->dcdState = (ap->save.msr & MSR_RLSD) ? rs232ON : rs232OFF;

	ap->dtrChanges = 0;
	ap->rtsChanges = 0;
	ap->dsrChanges = 0;
	ap->ctsChanges = 0;
	ap->dcdChanges = 0;

	/* Set interrupt vector to SIO handler */
	setirq(ap->vec,Handle[dev]);

	/* Set line control register: 8 bits, no parity */
	outportb(base+LCR,(char)LCR_8BITS);

	/* determine if 16550A, turn on FIFO mode and clear RX and TX FIFOs */
	outportb(base+FCR,(char) FIFO_ENABLE);

	/* According to National ap note AN-493, the FIFO in the 16550 chip
	 * is broken and must not be used. To determine if this is a 16550A
	 * (which has a good FIFO implementation) check that both bits 7
	 * and 6 of the IIR are 1 after setting the fifo enable bit. If
	 * not, don't try to use the FIFO.
	 */
	if ((inportb(base+IIR) & IIR_FIFO_ENABLED) == IIR_FIFO_ENABLED) {
		ap->is_16550a = TRUE;
		outportb(base+FCR,(char) FIFO_SETUP);
	} else {
		/* Chip is not a 16550A. In case it's a 16550 (which has a
		 * broken FIFO), turn off the FIFO bit.
		 */
		outportb(base+FCR,(char)0);
		ap->is_16550a = FALSE;
	}

	/* Turn on modem status and receive interrupts,
	 * leave transmit interrupts off for now.
	 */
	outportb(base+IER, (char)(IER_MS + IER_DAV));

	/* Turn on 8250 master interrupt enable (connected to OUT2) */
	setbit(base+MCR,(MCR_OUT2));

	/* Enable interrupt */
	maskon(ap->vec);
	restore(i_state);

	asy_speed(dev,atol(argv[7]));

	if ( monitor ) {
		/* set up to monitor line signals */
		ap->monitor = newproc( ifn = if_name( ifp, " monitor" ),
				2048, asy_monitor, ifp->d_index, ifp, NULL, 0);
		free(ifn);
		pwait(NULL);	/* let hooks get set up */
	}

	ifp->txproc = newproc( ifn = if_name(ifp," tx"),
			256, asy_tx, dev, ifp, NULL, 0);
	free(ifn);

        /* Exercise the line signals to see what effect they have.
         * By default, tell the line to come up.
         */
	asy_ioctl(ifp, PARAM_UP, TRUE, 0L);

	if ( ifp->iostatus != NULL ) {
		/* if dtr is up, the line could be up from a previous program.
		 * if dcd is also up, the other end is probably there.
		 * (dsr is a pretty unreliable indicator.)
		 */
		if ( (ap->dtrState & ~rs232CHANGED) == rs232ON
		  && (ap->dcdState & ~rs232CHANGED) == rs232ON
		 || ap->monitor == NULL ) {
			(*ifp->iostatus)(ifp, PARAM_UP, 0L);
		}
	}

	if ( blind ) {
		asy_ioctl(ifp, PARAM_BLIND, TRUE, 0L);
	}
	return 0;
}


int
asy_stop(ifp)
struct iface *ifp;
{
	register struct asy *ap = ifp->ddv;
	register unsigned base;
	char i_state;

	if ( ap->monitor != NULL ) {
		ap->dcdState = rs232NONE;
		killproc( ap->monitor );
		ap->monitor = NULL;
	}
	ap->iface = NULLIF;

	base = ap->addr;

	/* Purge the receive data buffer */
	(void)inportb(base+RBR);

	/* and hardware fifos if available */
	if (ap->is_16550a)
		outportb(base+FCR,(char) FIFO_SETUP);

	/* Restore original interrupt vector and 8259 mask state */
	i_state = dirps();
	setirq(ap->vec,ap->save.vec);
	if(ap->save.mask)
		maskon(ap->vec);
	else
		maskoff(ap->vec);

	/* Restore speed regs */
	setbit(base+LCR,LCR_DLAB);
	outportb(base+DLL,ap->save.divl);	/* Low byte */
	outportb(base+DLM,ap->save.divh);	/* Hi byte */
	clrbit(base+LCR,LCR_DLAB);

	/* Restore control regs */
	outportb(base+LCR,ap->save.lcr);
	outportb(base+IER,ap->save.ier);

#ifdef i8250_SAVE_DTR
	/* except, leave DTR in current state {bsimpson} */
	outportb(base+MCR,(ap->save.mcr &~ MCR_DTR)
			| (inportb(base+MCR) & MCR_DTR) );
#else
	outportb(base+MCR,ap->save.mcr);
#endif

	restore(i_state);
	free(ap->fifo.buf);
	return 0;
}


/* Set asynch line speed */
int
asy_speed(dev,bps)
int dev;
long bps;
{
	register unsigned base;
	register long divisor;
	struct asy *asyp;
	char i_state;

	if(bps <= 0 || dev >= ASY_MAX)
		return -1;
	asyp = &Asy[dev];
	if(asyp->iface == NULLIF)
		return -1;

	switch(bps) {
	case 300:
	case 1200:
	case 2400:
	case 4800:
	case 9600:
	case 19200:
	case 38400L:
		/* supported speed */
		asyp->InSpeed = asyp->OutSpeed = bps;
		break;
	default:
		/* unsupported speed */
		return -1;
	};

	base = asyp->addr;
	divisor = BAUDCLK / bps;

	i_state = dirps();

	/* Purge the receive data buffer */
	(void)inportb(base+RBR);
	if (asyp->is_16550a)		/* clear tx+rx fifos */
		outportb(base+FCR,(char) FIFO_SETUP);

	/* Turn on divisor latch access bit */
	setbit(base+LCR,LCR_DLAB);

	/* Load the two bytes of the register */
	outportb(base+DLL,(char)(divisor & 0xff));		/* Low byte */
	outportb(base+DLM,(char)((divisor >> 8) & 0xff));	/* Hi byte */

	/* Turn off divisor latch access bit */
	clrbit(base+LCR,LCR_DLAB);

	restore(i_state);
	return 0;
}


/* Asynchronous line I/O control */
int32
asy_ioctl(ifp,cmd,set,val)
struct iface *ifp;
int cmd;
int set;
int32 val;
{
	struct asy *ap = ifp->ddv;
	int16 base = ap->addr;

	switch(cmd){
	case PARAM_SPEED:
		if(set)
			asy_speed(ifp->d_index,val);
		return ap->OutSpeed;
	case PARAM_DTR:
		if ( set ) {
			int have = inportb(base+MCR) & MCR_DTR;
			if ( val && !have ) {
				setbit(base+MCR,MCR_DTR);
				ap->dtrState = rs232UP;
				ap->dtrChanges++;
			} else if ( !val && have ) {
				clrbit(base+MCR,MCR_DTR);
				ap->dtrState = rs232DOWN;
				ap->dtrChanges++;
			}
		}
		return (inportb(base+MCR) & MCR_DTR) ? TRUE : FALSE;
	case PARAM_RTS:
		if ( set ) {
			int have = inportb(base+MCR) & MCR_RTS;
			if ( val && !have ) {
				setbit(base+MCR,MCR_RTS);
				ap->rtsState = rs232UP;
				ap->rtsChanges++;
			} else if ( !val && have ) {
				clrbit(base+MCR,MCR_RTS);
				ap->rtsState = rs232DOWN;
				ap->rtsChanges++;
			}
		}
		return (inportb(base+MCR) & MCR_RTS) ? TRUE : FALSE;
	case PARAM_DOWN:
		clrbit(base+IER,(char)IER_DAV);
		asy_ioctl( ifp, PARAM_RTS, TRUE, FALSE );
		asy_ioctl( ifp, PARAM_DTR, TRUE, FALSE );
		return FALSE;
	case PARAM_UP:
		setbit(base+IER,(char)IER_DAV);
		asy_ioctl( ifp, PARAM_DTR, TRUE, TRUE );
		asy_ioctl( ifp, PARAM_RTS, TRUE, TRUE );
		return TRUE;
	case PARAM_BLIND:
		setbit(base+IER,(char)IER_DAV);
#ifdef notdef
		ap->rtsState = rs232NONE;
		ap->dtrState = rs232NONE;
#endif
		ap->ctsState = rs232NONE;
		ap->dcdState = rs232NONE;

		/* can't see what we are doing, so pretend we're up */
		if ( ifp->iostatus != NULL ) {
			(*ifp->iostatus)(ifp, PARAM_UP, 0L);
		}
		return TRUE;
	};
	return -1;
}


/* Start transmission of a buffer on the serial transmitter */
static void
asy_output(dev,buf,cnt)
int dev;
char *buf;
unsigned short cnt;
{
	register struct dma *dp;
	unsigned base;
	char i_state;
	char ier;
	struct asy *asyp;

	if(dev < 0 || dev >= ASY_MAX)
		return;
	asyp = &Asy[dev];
	if(asyp->iface == NULLIF)
		return;

	base = asyp->addr;
	dp = &asyp->dma;
	i_state = dirps();

	if(dp->flags){
		restore(i_state);
		return;	/* Already busy */
	}
	dp->data = buf;
	dp->cnt = cnt;
	dp->flags = 1;

	switch (asyp->ctsState) {
	case rs232UP:
	case rs232DOWN:
		/* CTS flow control is enabled; let the modem control
		 * interrupt enable transmit interrupts if CTS is off
		 */
		ier = IER_MS;
		if(inportb(base+MSR) & MSR_CTS)
			ier |= IER_TxE;
	default:
		/* Enable transmit interrupts; this will cause an immediate
		 * interrupt that will start things going
		 */
		ier = IER_TxE;
	};
	setbit(base+IER,ier);

	/* "Kick start" the transmitter interrupt routine, in case just
	 * setting the interrupt enable bit doesn't cause an interrupt
	 */
	if(ier & IER_TxE)
		asytxint(dev);
	restore(i_state);
}


/* Blocking read from asynch line
 * Returns character or -1 if aborting
 */
int
get_asy(dev)
int dev;
{
	char i_state;
	register struct fifo *fp;
	char c;

	fp = &Asy[dev].fifo;

	i_state = dirps();
	while(fp->cnt == 0){
		if(pwait(fp) != 0){
			restore(i_state);
			return -1;
		}
	}
	fp->cnt--;
	restore(i_state);

	c = *fp->rp++;
	if(fp->rp >= &fp->buf[fp->bufsize])
		fp->rp = fp->buf;

	return uchar(c);
}


/* Interrupt handler for 8250 asynch chip (called from asyvec.asm) */
void
asyint(dev)
int dev;
{
	struct asy *asyp = &Asy[dev];
	unsigned base = asyp->addr;
	char iir;

	while(((iir = inportb(base+IIR)) & IIR_IP) == 0){
		switch(iir & IIR_ID_MASK){
		case IIR_RDA:	/* Receiver interrupt */
			asyrxint(asyp);
			break;
		case IIR_THRE:	/* Transmit interrupt */
			asytxint(dev);
			break;
		case IIR_MSTAT:	/* Modem status change */
			asymsint(dev);
			asyp->msint_count++;
			break;
		}
		/* should happen at end of a single packet */
		if(iir & IIR_FIFO_TIMEOUT)
			asyp->hwtimeouts++;
	}
}


/* Process 8250 receiver interrupts */
static int
asyrxint(asyp)
struct asy *asyp;
{
	register struct fifo *fp;
	unsigned base;
	char c,lsr;
	int cnt = 0;
	int trigseen = FALSE;

	asyp->rxints++;
	base = asyp->addr;
	fp = &asyp->fifo;
	for(;;){
		lsr = inportb(base+LSR);
		if(lsr & LSR_OE)
			asyp->OverrunErrors++;

		if(lsr & LSR_DR){
			asyp->rxchar++;
			c = inportb(base+RBR);
			if(asyp->trigchar == uchar(c))
				trigseen = TRUE;
			/* If buffer is full, we have no choice but
			 * to drop the character
			 */
			if(fp->cnt != fp->bufsize){
				*fp->wp++ = c;
				if(fp->wp >= &fp->buf[fp->bufsize])
					/* Wrap around */
					fp->wp = fp->buf;
				fp->cnt++;
				if(fp->cnt > fp->hiwat)
					fp->hiwat = fp->cnt;
				cnt++;
			} else
				fp->overrun++;
		} else
			break;
	}
	if(cnt > asyp->hwhiwat)
		asyp->hwhiwat = cnt;
	if(trigseen)
		psignal(fp,1);
	return cnt;
}


/* Handle 8250 transmitter interrupts */
static void
asytxint(dev)
int dev;
{
	struct asy *asyp = &Asy[dev];
	register unsigned base = asyp->addr;
	register struct dma *dp = &asyp->dma;
	register int count;

	asyp->txints++;
	if(!dp->flags){
		/* "Shouldn't happen", but disable transmit
		 * interrupts anyway
		 */
		clrbit(base+IER,IER_TxE);
		return;	/* Nothing to send */
	}
	if(!(inportb(base+LSR) & LSR_THRE))
		return;	/* Not really ready */

	/* If it's a 16550A, load up to 16 chars into the tx hw fifo
	 * at once. With an 8250, it can be one char at most.
	 */
	if(asyp->is_16550a){
		count = min(dp->cnt,OUTPUT_FIFO_SIZE);

		/* 16550A: LSR_THRE will drop after the first char loaded
		 * so we can't look at this bit to determine if the hw fifo is
		 * full. There seems to be no way to determine if the tx fifo
		 * is full (any clues?). So we should never get here while the
		 * fifo isn't empty yet.
		 */
		asyp->txchar += count;
		dp->cnt -= count;
#ifdef	notdef	/* This is apparently too fast for some chips */
		dp->data = outbuf(base+THR,dp->data,count);
#else
		while(count-- != 0)
			outportb(base+THR,*dp->data++);
#endif
		if(dp->cnt == 0){
			dp->flags = 0;
			/* Disable transmit interrupts */
			clrbit(base+IER,IER_TxE);
			psignal(asyp,1);
		}
	} else {	/* 8250 */
		do {
			asyp->txchar++;
			outportb(base+THR,*dp->data++);

			if(--dp->cnt == 0){
				dp->flags = 0;
				/* Disable transmit interrupts */
				clrbit(base+IER,IER_TxE);
				psignal(asyp,1);
				break;
			}
		} while(inportb(base+LSR) & LSR_THRE);
	}
}


/* Handle 8250 modem status change
 *	If the signals are unchanging, we ignore them.
 *	If they change, we use them to condition the line.
 */
static void
asymsint(dev)
int dev;
{
	struct asy *ap = &Asy[dev];
	unsigned base = ap->addr;

	ap->msr = inportb(base+MSR);

	if ( ap->msr & MSR_CTS ) {
		switch ( ap->ctsState ) {
		case rs232OFF:
		case rs232DOWN:
			ap->ctsState = rs232UP;
			ap->ctsChanges++;

			/* CTS now asserted, enable Transmit interrupts */
			if(ap->dma.flags)
				setbit(base+IER,IER_TxE);
			break;
		};
	} else {
		switch ( ap->ctsState ) {
		case rs232ON:
		case rs232UP:
			ap->ctsState = rs232DOWN;
			ap->ctsChanges++;

			/* CTS now dropped, disable Transmit interrupts */
			clrbit(base+IER,IER_TxE);
			break;
		};
	}

	if ( ap->msr & MSR_RLSD ) {
		switch ( ap->dcdState ) {
		case rs232OFF:
		case rs232DOWN:
			ap->dcdState = rs232UP;
			ap->dcdChanges++;

			/* RLSD just went up */
			psignal( &(ap->dcdState), 1 );
			break;
		};
	} else {
		switch ( ap->dcdState ) {
		case rs232ON:
		case rs232UP:
			ap->dcdState = rs232DOWN;
			ap->dcdChanges++;

			/* RLSD just went down */
			psignal( &(ap->dcdState), 1 );
			break;
		};
	}

	if ( ap->msr & (MSR_TERI | MSR_RI) ) {
		asy_ioctl( ap->iface, PARAM_UP, TRUE, 0L );
	}
}


/* Wait for a signal that the RLSD modem status has changed */
int
get_rlsd_asy(dev, want)
int dev;
int want;
{
	struct asy *ap = &Asy[dev];
	struct iface *ifp = ap->iface;
	int result;

	if ( ap->dcdState == rs232NONE ) {
		return rs232NONE;
	}

	switch ( want ) {
	case rs232NONE:
		/* Just return the current value */
		return(ap->dcdState);

	case rs232OFF:
	case rs232DOWN:
		switch (ap->dcdState) {
		case rs232OFF:
		case rs232DOWN:
			/* Already at requested value */
			return(want);
		}
		break;

	case rs232ON:
	case rs232UP:
		switch (ap->dcdState) {
		case rs232ON:
		case rs232UP:
			/* Already at requested value */
			return(want);
		}
		break;
	};

	/* Wait for state change to requested value */
	want |= rs232CHANGED;
	while (ap->dcdState != want
	    && ap->dcdState != rs232NONE) {
		pause(2L);
		pwait( &(ap->dcdState) );
	}

	if ( ap->dcdState == rs232UP )
		result = PARAM_UP;
	else /* DOWN or NONE */
		result = PARAM_DOWN;

	/* set our control signals to follow RLSD */
	if ( ifp->ioctl != NULL )
		(*ifp->ioctl)( ifp, result, TRUE, 0L );
	if ( ifp->iostatus != NULL )
		(*ifp->iostatus)( ifp, result, 0L );
	return(ap->dcdState);
}


/* Monitor RLSD signal, and report status changes */
static void
asy_monitor( dev, p1, p2 )
int dev;
void *p1;
void *p2;
{
	int saved = Asy[dev].dcdState;

	while ( saved != rs232NONE ) {
		saved = get_rlsd_asy( dev, saved ^ 0x01 );
	}
	Asy[dev].monitor = NULL;
}


/* Poll the asynch input queues; called on every clock tick.
 * This helps limit the interrupt ring buffer occupancy when long
 * packets are being received.
 */
void
asytimer()
{
	register struct asy *asyp;
	register struct fifo *fp;
	register int i;

	for(i=0;i<ASY_MAX;i++){
		asyp = &Asy[i];
		fp = &asyp->fifo;
		if(fp->cnt != 0)
			psignal(fp,1);
		if(asyp->dma.flags != 0 && (inportb(asyp->addr+LSR) & LSR_THRE) ) {
			asyp->txto++;
			asytxint(asyp->iface->d_index);
		}
	}
}
int
doasystat(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	register struct asy *asyp;

	for(asyp = Asy;asyp < &Asy[ASY_MAX];asyp++){
		if(asyp->iface == NULLIF)
			continue;

		tprintf("%s:",asyp->iface->name);
		if(asyp->is_16550a)
			tprintf(" [NS16550A]");
		if(asyp->trigchar != -1)
			tprintf(" [trigger 0x%02x]",asyp->trigchar);

		switch (asyp->ctsState) {
		case rs232OFF:
		case rs232DOWN:
			tprintf(" [cts down %u]", asyp->ctsChanges);
			break;
		case rs232ON:
		case rs232UP:
			tprintf(" [cts up %u]", asyp->ctsChanges);
			break;
		}
		switch (asyp->dcdState) {
		case rs232NONE:
			tprintf(" [blind]");
			break;
		case rs232OFF:
		case rs232DOWN:
			tprintf(" [dcd down %u]", asyp->dcdChanges);
			break;
		case rs232ON:
		case rs232UP:
			tprintf(" [dcd up %u]", asyp->dcdChanges);
			break;
		}

		tprintf(" %lu bps\n",asyp->OutSpeed);

		tprintf(" RX: %lu int, %lu chr, %u hw over, %u hw hi,",
			asyp->rxints,
			asyp->rxchar,
			asyp->OverrunErrors,
			asyp->hwhiwat);
		asyp->hwhiwat = 0;
		if(asyp->is_16550a)
			tprintf(" %lu fifo TO,",asyp->hwtimeouts);
		tprintf(" %u sw over, %u sw hi\n",
			asyp->fifo.overrun,
			asyp->fifo.hiwat);
		asyp->fifo.hiwat = 0;

		if(tprintf(" TX: %lu int, %lu chr, %u q, %lu MS int, %lu THRE TO\n",
			asyp->txints,
			asyp->txchar,
			len_q(asyp->sndq),
			asyp->msint_count,
			asyp->txto) == EOF)
			break;
	}
	return 0;
}
/* Send a message on the specified serial line */
int
asy_send(dev,bp)
int dev;
struct mbuf *bp;
{
	if(dev < 0 || dev >= ASY_MAX)
		return -1;
	enqueue(&Asy[dev].sndq,bp);
	return 0;
}

/* Serial transmit process, common to all protocols */
static void
asy_tx(dev,p1,p2)
int dev;
void *p1;
void *p2;
{
	register struct mbuf *bp;
	struct asy *asyp;
	struct dma *dp;
	int i_state;

	asyp = &Asy[dev];
	dp = &asyp->dma;

	for(;;){
		/* Fetch a buffer for transmission */
		while(asyp->sndq == NULLBUF)
			pwait(&asyp->sndq);

		bp = dequeue(&asyp->sndq);

		while(bp != NULLBUF){
			/* Start the transmitter */
			asy_output(dev,bp->data,bp->cnt);

			/* Wait for completion */
			i_state = dirps();
			while(dp->flags == 1)
				pwait(asyp);
			restore(i_state);

			/* Now do next buffer on chain */
			bp = free_mbuf(bp);
		}
	}
}

