/*
   lock.c : Copyright Paul Healy, EI9GL, 1992.

   Note: Turbo C has lock and unlock functions which call MSDOS locking
         facilities.

   Derived from bm.

   Copyright 1986 Bdale Garbee, All Rights Reserved.
   Permission granted for non-commercial copying and use, provided
   this notice is retained.
   Copyright 1987 1988 Dave Trulli NN2Z, All Rights Reserved.
   Permission granted for non-commercial copying and use, provided
   this notice is retained.

   911218 : Added this header.
   920601 : Added test programs - define LOCK_PROG, UNLOCK_PROG to use them.
   921105 : Added some diagnostics to fopenlocked when a lock op fails.
*/
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <io.h>
#include <conio.h>
#include <ctype.h>
#include <errno.h>
#include <time.h>
#include "config.h"
#include "lock.h"
#include "misc.h"

#undef LOCKDEBUG

int
islocked(char *name)
{
   char s[MAXNAMLEN];

   sprintf(s, "%s.lck", name);

   return access(s, 0) == 0;
}

/*
 *  Make lock file for file specified in name (no extension) by
 *  opening <name>.lck in a 'atomic' fashion.
 *
 *  returns 0 for success, -1 otherwise
 */
int
mklock(char *name)
{
   /* Try to create the lock file in an atomic operation */

   int fd = open(name, O_WRONLY|O_EXCL|O_CREAT, 0600);

   return fd == -1 ? -1 : close(fd);
}

int
bm_lock(char *name)
{
   char s[MAXNAMLEN];

   sprintf(s, "%s.lck", name);
#ifdef LOCKDEBUG
   printf("bm_lock: locking %s%s\n", name, EXT);
#endif

   return mklock(s);
}

/*
 *  returns 0 for success, -1 otherwise
 */
int
rmlock(char *name)
{
   return unlink(name);
}

int
bm_unlock(char *name)
{
   char s[MAXNAMLEN];

   sprintf(s, "%s.lck", name);
#ifdef LOCKDEBUG
   printf("bm_unlock: unlocking %s%s\n", name, EXT);
#endif
   return rmlock(s);
}

static int
fileinfo(char *name)
{
   struct stat sbuf;
   char s[MAXNAMLEN];

   sprintf(s, "%s.lck", name);

   if (stat(s, &sbuf) != 0)
      return -1;     

   fprintf(stderr, "%s was locked at %s", name, ctime(&sbuf.st_atime));

   return 0;
}

/*
 *  returns 0 for success, -1 otherwise
 */
int
lockit(char *name)
{
   fileinfo(name);

   if (bm_lock(name) == 0)
      return 0;

   while (bm_lock(name)) {
      time_t t = time(NULL);

      fprintf(stderr, "\rTime is %s: Abort or Retry ? [R]:", rip(ctime(&t)));

      switch ( tolower(getch()) ) {
         case 'a':
         case 'q':
            putchar('\n');
            return -1;
         case 'u':      /* hidden option: unlock/ignore old lock */
            putchar('\n');
            return 0;
         case 'r':
         default:
            continue;
         }
      }
   putchar('\n');

   return 0;
}

FILE *
fopenlocked(char *dir, char *base, char *ext, char *mode)
{
   char lock[MAXNAMLEN], file[MAXNAMLEN];
   FILE *fp;

   sprintf(lock, "%s/%s", dir, base);

   if (bm_lock(lock) == -1) {
      fprintf(stderr, "bmh: can't lock %s (%s)\n", lock, sys_errlist[errno]);
      if (islocked(lock))
         fileinfo(lock);
      return NULL;
      }

   sprintf(file, "%s%s", lock, ext);

   if ((fp = fopen(file, mode)) == NULL) {
      if (mode[0] != 'r')
         fprintf(stderr, "lock: can't open %s\n", file);
      bm_unlock(lock);
      return NULL;
      }

   return fp;
}

int
fcloselocked(FILE *fp, char *dir, char *base)
{
   char lock[MAXNAMLEN];
   int ret = fclose(fp);

   sprintf(lock, "%s/%s", dir, base);
   if (bm_unlock(lock) == -1) {
      fprintf(stderr, "bmh: warning can't unlock %s\n", lock);
      return -1;
      }
   else 
      return ret;
}

#ifdef LOCK_PROG
/*
   Syntax:
            lock <file>

            eg lock /usr/spool/mail/ei9gl
               will try to create a lock file /usr/spool/mail/ei9gl.lck
*/
int
main(int argc, char *argv[])
{
   if (argc != 2) {
      fprintf(stderr, "Usage: lock <mailbox>\n");
      return -1;
      }
   return bm_lock(argv[1]);
}
#endif

#ifdef UNLOCK_PROG
/*
   Syntax:
            unlock <file>

            eg unlock /usr/spool/mail/ei9gl
               will try to remove the lock file /usr/spool/mail/ei9gl.lck
*/
int
main(int argc, char *argv[])
{
   if (argc != 2) {
      fprintf(stderr, "Usage: unlock <mailbox>\n");
      return -1;
      }
   return bm_unlock(argv[1]);
}
#endif
