/* Internet LPQ, LPR, & LPRM clients
 *   written by David Johnson (dave@cs.olemiss.edu)
 *
 * This code is in the public domain.
 *
 * Revision History:
 *
 * Revision 1.3  91/09/28  dave (part from Hans-Juergen Knobloch)
 * If LPR file does not exist, skip it and continue
 * LPR did not accept more than one file to print
 * If errors occur during an LPR session, force user to press a key at the end
 * If the job request find no files to send, cancel request to LPD
 *
 * Revision 1.2  91/09/26  dave (from Hans-Juergen Knobloch)
 * Changed definition of priority to char
 *
 * Revision 1.1  91/09/19  dave
 * Added option to lpq and lprm to allow calling from lpc and not
 *     allocate a session in that case
 *
 * Revision 1.0  91/09/04  dave
 * Initial Release
 *
 */
#include <stdio.h>
#include <string.h>
#include <time.h>
#include "global.h"
#include "mbuf.h"
#include "socket.h"
#include "session.h"
#include "proc.h"
#include "netuser.h"
#include "commands.h"
#include "tty.h"
#include "lp.h"


/*****************************************************************************/
/*
 * LPR Client
 */
static void addoption __ARGS((char option, char *arg));
static int  send_data_file __ARGS((char *filename, char *remotefile, int binary, int s));
static int  send_control_file __ARGS((char *remotefile, int s));

static char *cf_buffer;
static char *cf_ptr = NULL;


int
dolpr(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	struct session *sp;
	struct sockaddr_in sock;
	char *cp, remotefile[32], buf[128];
	int c, s, i;
	int connected = 0;
	int errors = 0;
	int files_sent = 0;
	struct mbuf *bp;

	int binary = 0, burst = 1, copies = 1;
	char *printer = "lp", priority = 'Z';
	char *server = "127.0.0.1";
	char format = 'f';

	int jobname = 0, jobclass = 0, jobtitle = 0;

	time_t current_time;
	struct tm *timeptr;


	if (argc < 2) {
		tprintf( "Invalid options\n" );
		return -1;
	}

	/* Allocate a session control block */
	if((sp = newsession(argv[1],LPUTILS)) == NULLSESSION){
		tprintf("Too many sessions\n");
		keywait(NULLCHAR,1);
		return 1;
	}

	cf_ptr = cf_buffer = callocw( 1, 512 );

	if( Hostname == NULL )
		addoption( 'H', "unknown" );
	else
		addoption( 'H', Hostname );
	if( (cp = getenv("USER")) != NULL ) {
		while( *cp == ' ' )	/* skip leading spaces */
			++cp;
		addoption( 'P', cp );
	} else
		addoption( 'P', "NOS/PC" );

	optind = 1;		/* reinit getopt() */
loop:	while((c = getopt(argc,argv,"S:P:pltndgvcfhi::C:J:T:#::w::")) != EOF){
		switch( c ) {
			case 'S':	/* server */
				server = optarg;
				break;
			case 'P':       /* printer */
				printer = optarg;
				break;
			case 'C':       /* Class */
				if( jobclass++ < 1 ) {	/* only once */
					addoption( 'C', optarg );
					if( strlen( optarg ) == 1 )
						priority = *optarg;
				}
				break;
			case 'J':       /* Job */
				if( jobname++ < 1 )	/* only once */
					addoption( 'J', optarg );
				break;
			case 'T':       /* Title */
				if( jobtitle++ < 1 )	/* only once */
					addoption( 'T', optarg );
				break;
			case '#':       /* Number of copies */
				if( optarg )
					copies = atoi( optarg );
				break;
			case 'w':       /* Page Width */
				if( optarg )
					addoption( 'W', optarg );
				break;
			case 'f':       /* ordinary file type */
			case 'p':       /* use pr formatter */
				format = c;
				binary = 0;
				break;
			case 'c':       /* cifplot file type */
			case 'd':       /* tex file type */
			case 'g':       /* plot file type */
			case 'l':       /* allow control chars */
			case 'n':       /* ditroff file type */
			case 't':       /* troff file type */
			case 'v':       /* raster image file type */
				format = c;
				binary = 1;
				break;
			case 'h':       /* no burst page */
				burst = 0;
				break;
			case 'i':       /* indent # of spaces */
				if( optarg )
					addoption( 'I', optarg );
				else
					addoption( 'I', "8" );
				break;
		}
	}
	if( !connected ) {
		sock.sin_family = AF_INET;
		sock.sin_port = IPPORT_LPD;
		tprintf("Resolving %s... ",server);
		if((sock.sin_addr.s_addr = resolve(server)) == 0 ) {
			tprintf( "Host %s unknown\n",server );
			goto finish;
		}
		tprintf("trying %s\n",psocket((struct sockaddr *)&sock));
		if((s = socket(AF_INET,SOCK_STREAM,0)) == -1){
			tprintf("Can't create socket\n");
			goto finish;
		}
		sockmode(s,SOCK_BINARY);
		if(connect(s,(char *)&sock,sizeof(sock)) == -1){
			cp = sockerr(s);
			tprintf(" -- Connect failed: %s\n",cp != NULLCHAR ? cp : "");
			close_s(s);
			goto finish;
		}
		tprintf( "\n" );
		connected = 1;
	}
	while( optind < argc ) {
		if( *argv[optind] == '-' )	/* more options */
			goto loop;

		/* file to be printed in argv[optind] */

		if( printer ) {			/* haven't sent printer */
			usprintf( s, "%c%s\n", RECEIVE_CMD, printer );
			printer = NULL;
			c = GET_RESPONSE( s );
			if( c == STATUS_ERR ) {
				tprintf( "printer not supported\n" );
				close_s( s );
				goto finish;
			}
			time( &current_time );
			timeptr = gmtime( &current_time );
			sprintf( remotefile, "dfA%02d%02dI%s", timeptr->tm_min, timeptr->tm_sec, inet_ntoa( Ip_addr ) );
		} else {			/* increment file letter */
                        remotefile[2] = remotefile[2] + 1;
		}
		if( jobname < 1 )
			addoption( 'J', argv[optind] );
		if( burst-- )			/* only once */
			addoption( 'L', "user" );

		tprintf( "sending %s...", argv[optind] );
		if( send_data_file( argv[optind], remotefile, binary, s ) < 0 ) {
			errors++;
			optind++;
			continue;
		}
		tprintf( "done\n" );

		files_sent++;
/*		if( format == 'p' )
			addoption( 'T', "title" ); */
		addoption( 'N', argv[optind] );

		for( i = copies; i; --i )
			addoption( format, remotefile );
		addoption( 'U', remotefile );
		optind++;
	}

	if( files_sent > 0 ) {
		/*
		* Send control file
		*/
		remotefile[0] = 'c';
		remotefile[2] = priority;
		if( send_control_file( remotefile, s ) < 0 ) {
			errors++;
			tprintf( "control file not accepted\n" );
		}
	} else {
		send(s,"\0",1,0);	/* cancel current requests */
	}
	close_s(s);
finish:
	free( cf_buffer );
	if( errors )
		keywait(NULLCHAR,1);
	freesession( sp );
	return 0;
}

static void
addoption( option, arg )
char option, *arg;
{
	sprintf( cf_ptr, "%c%s\n", option, arg );
	cf_ptr += strlen( arg ) + 2;
}

static int
send_data_file( filename, remotefile, binary, s )
char *filename, *remotefile;
int binary, s;
{
	int bytes, c;
	unsigned long filesize, total = 0;
	char *buffer;
	FILE *file;

#ifdef LPD_DEBUG
	tprintf( "send_data_file( %s, %s, %d )\n", filename, remotefile, binary );
	tflush();
#endif
	if( access( filename, 0 ) < 0 ) {
		/* file does not exist */

		tprintf( "file not found\n" );
		return -1;
	}

	buffer = (char *)malloc( BLOCKSIZE );

	/*
	 * Open file and obtain correct file size (Damn DOS!!)
	 */
	if( binary ) {
		file = fopen( filename, READ_BINARY );

		fseek( file, 0L, SEEK_END );
		filesize = ftell( file );
		c = getc( file );
		while( c == 26 || c == EOF ) {
			filesize--;
			fseek( file, filesize, 0 );
			c = getc( file );
		}
		filesize++;         /* last character was good */
	} else {
		file = fopen( filename, READ_TEXT );

		filesize = 0L;
		while( (c = fread( buffer, 1, BLOCKSIZE, file )) != 0 )
			filesize += c;
	}
	rewind( file );

	/*
	 * Send LPD command
	 */
	usprintf( s, "%c%ld %s\n", RECEIVE_DF_SCMD, filesize, remotefile );
	c = GET_RESPONSE( s );
	if( c == STATUS_ERR ) {
		tprintf( "file rejected\n" );
		return -1;
	} else if( c == 2 ) {
		tprintf( "server out of storage space\n" );
		return -1;
	}

	while( (bytes = fread( buffer, 1, 512, file ) ) == 512 ) {
		if( total + bytes > filesize )          /* don't send ^Z */
			bytes -= (total + bytes) - filesize;
		total += bytes;
		send( s, buffer, bytes, 0 );
	}
	if( bytes ) {
		if( total + bytes > filesize )          /* don't send ^Z */
			bytes -= (total + bytes) - filesize;
		total += bytes;
		send( s, buffer, bytes, 0 );
	}
	fclose( file );
	free( buffer );
	RESPOND_OK( s );
	c = GET_RESPONSE( s );
	if( c == STATUS_ERR ) {
		tprintf( "file rejected\n" );
		return -1;
	} else if( c == 2 ) {
		tprintf( "server out of storage space\n" );
		return -1;
	}
	return 0;
}

static int
send_control_file( remotefile, s )
char *remotefile;
int s;
{
	int c;
	unsigned int filesize;

#ifdef LPD_DEBUG
	tprintf( "send_control_file( %s )\n", remotefile );
	tflush();
#endif

	filesize = strlen( cf_buffer );

	/*
	 * Send LPD command
	 */
	usprintf( s, "%c%d %s\n", RECEIVE_CF_SCMD, filesize, remotefile );
	c = GET_RESPONSE( s );
	if( c == STATUS_ERR ) {
		tprintf( "control file rejected\n" );
		return -1;
	} else if( c == 2 ) {
		tprintf( "server out of storage space\n" );
		return -1;
	}
	send( s, cf_buffer, filesize, 0 );
	RESPOND_OK( s );

	c = GET_RESPONSE( s );
	if( c == STATUS_ERR ) {
		tprintf( "control file rejected\n" );
		return -1;
	} else if( c == 2 ) {
		tprintf( "server out of storage space\n" );
		return -1;
	}
	return 0;
}

/*****************************************************************************/
/*
 * LPQ Client
 */
int
dolpq(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	struct session *sp;
	struct sockaddr_in sock;
	char buf[128];
	int c, s, cnt, long_list = 0, from_lpc = 0;
	int connected = 0;
	char *cp;

	char *printer = "lp";
	char *server = "127.0.0.1";

	optind = 1;		/* reinit getopt() */
loop:	while((c = getopt(argc,argv,"lcP:S:")) != EOF){
		switch( c ) {
			case 'S':	/* server */
				server = optarg;
				break;
			case 'P':       /* printer */
				printer = optarg;
				break;
			case 'l':       /* long listing */
				long_list = 1;
				break;
			case 'c':       /* called from lpc */
				from_lpc = 1;
				break;
		}
	}
	if( !from_lpc ) {
		/* Allocate a session control block */
		if((sp = newsession(argv[1],LPUTILS)) == NULLSESSION){
			tprintf("Too many sessions\n");
			keywait(NULLCHAR,1);
			return 1;
		}
	}
	if( !connected ) {
		sock.sin_family = AF_INET;
		sock.sin_port = IPPORT_LPD;
		tprintf("Resolving %s... ",server);
		if((sock.sin_addr.s_addr = resolve(server)) == 0 ) {
			tprintf( "Host %s unknown\n",server );
			goto finish;
		}
		tprintf("trying %s\n",psocket((struct sockaddr *)&sock));
		if((s = socket(AF_INET,SOCK_STREAM,0)) == -1){
			tprintf("Can't create socket\n");
			goto finish;
		}
		sockmode(s,SOCK_BINARY);
		if(connect(s,(char *)&sock,sizeof(sock)) == -1){
			cp = sockerr(s);
			tprintf(" -- Connect failed: %s\n",cp != NULLCHAR ? cp : "");
			close_s(s);
			goto finish;
		}
		tprintf( "\n" );
		connected = 1;
	}

	if( long_list )
		usprintf( s, "%c%s\n", QUEUE_CMD, printer );
	else
		usprintf( s, "%c%s\n", SQUEUE_CMD, printer );

rloop:	if( (cnt = recvline( s, buf, sizeof(buf) )) == -1){
		/* He closed on us */
		close_s(s);
		goto finish;
	}
	if( cnt ){
		tputs( buf );
		goto rloop;
	}
	close_s(s);
finish:
	if( !from_lpc ) {
		keywait(NULLCHAR,1);
		freesession( sp );
	}
	return 0;
}


/*****************************************************************************/
/*
 * LPRM Client
 */

/* in lpd.c */
extern int  trunc_strcat __ARGS((char *dest, char *source, int max_length));

int
dolprm(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	struct session *sp;
	struct sockaddr_in sock;
	char *user, *buffer, *strdup();
	int c, s, cnt, all = 0, from_lpc = 0;
	int connected = 0;
	char *cp;

	char *printer = "lp";
	char *server = "127.0.0.1";

	if( (cp = getenv("USER")) != NULL ) {
		while( *cp == ' ' )	/* skip leading spaces */
			++cp;
		user = strdup( cp );
	} else
		user = strdup( "NOS/PC" );

	buffer = (char *)mallocw( BLOCKSIZE );
	*buffer = NULL;

	optind = 1;		/* reinit getopt() */
loop:	while((c = getopt(argc,argv,"S:P:ac")) != EOF){
		switch( c ) {
			case 'S':	/* server */
				server = optarg;
				break;
			case 'P':       /* printer */
				printer = optarg;
				break;
			case 'a':       /* long listing */
				all = 1;
				break;
			case 'c':       /* called from lpc */
				from_lpc = 1;
				break;
		}
	}
	if( !from_lpc ) {
		/* Allocate a session control block */
		if((sp = newsession(argv[1],LPUTILS)) == NULLSESSION){
			tprintf("Too many sessions\n");
			keywait(NULLCHAR,1);
			return 1;
		}
	}
	if( !connected ) {
		sock.sin_family = AF_INET;
		sock.sin_port = IPPORT_LPD;
		tprintf("Resolving %s... ",server);
		if((sock.sin_addr.s_addr = resolve(server)) == 0 ) {
			tprintf( "Host %s unknown\n",server );
			goto finish;
		}
		tprintf("trying %s\n",psocket((struct sockaddr *)&sock));
		if((s = socket(AF_INET,SOCK_STREAM,0)) == -1){
			tprintf("Can't create socket\n");
			goto finish;
		}
		sockmode(s,SOCK_BINARY);
		if(connect(s,(char *)&sock,sizeof(sock)) == -1){
			cp = sockerr(s);
			tprintf(" -- Connect failed: %s\n",cp != NULLCHAR ? cp : "");
			close_s(s);
			goto finish;
		}
		tprintf( "\n" );
		connected = 1;
	}

	while( optind < argc ) {
		trunc_strcat( buffer, argv[ optind ], BLOCKSIZE );
		optind++;
	}
	if( printer == NULL )
		printer = strdup( "lp" );

	usprintf( s, "%c%s %s %s\n", REMOVE_CMD, printer, user, buffer );

rloop:	if( (cnt = recvline( s, buffer, BLOCKSIZE )) == -1){
		/* He closed on us */
		close_s(s);
		goto finish;
	}
	if( cnt ){
		tputs( buffer );
		goto rloop;
	}
	close_s(s);
finish:
	if( !from_lpc ) {
		keywait(NULLCHAR,1);
		freesession( sp );
	}
	free( buffer );
	free( user );
	return 0;
}
