/* Internet LPD Server unspooler subroutines
 *   written by David Johnson (dave@cs.olemiss.edu)
 *
 * This code is in the public domain.
 *
 * Revision History:
 *
 * Revision 1.1  91/09/14  dave
 * Assign value to new device field in filter_stream for access to printer
 *	status
 *
 * Revision 1.0  91/09/04  dave
 * Initial Release
 *
 */
#include <stdio.h>
#include <fcntl.h>
#include <ctype.h>
#include <time.h>
#include <sys/stat.h>
#ifdef	__TURBOC__
#include <io.h>
#include <dir.h>
#endif
#include "global.h"
#include "mbuf.h"
#include "proc.h"
#include "iface.h"
#include "socket.h"
#include "dirutil.h"
#include "commands.h"
#include "files.h"
#include "lp.h"
#include "lpd.h"
#include "lpdfilt.h"


/* in lpdbann.c */
extern void print_banner __ARGS((struct LPDtrans *LPD, struct filter_stream *io_stream, char *username));

/* in lpdsubr.c */
extern int  cf_parse __ARGS((struct LPDtrans *LPD, FILE *cfp));
extern int  close_device __ARGS((struct LPDtrans *LPD));
extern void free_job __ARGS((struct LPDjob *job));
extern struct LPDdevice *get_device_entry __ARGS((char *device_name));
extern int  get_job_list __ARGS((char *directory, struct job_entry **job_list));
extern char **get_queue_list __ARGS((void));
extern struct LPDstatus *get_status_entry __ARGS((char *queue));
extern void lpd_log __ARGS((struct LPDtrans *LPD, char *message, ...));
extern int  open_device __ARGS((struct LPDtrans *LPD));
extern char *read_printcap __ARGS((struct LPDtrans *LPD, char *queue));
extern int  save_status __ARGS((char *queue, unsigned int flags, unsigned int next_seq));
extern void set_status_message __ARGS((char **dest, char *message, ...));
extern char *time_string __ARGS((void));

/* in lpdfilt.c */
extern void output_char __ARGS((struct filter_stream *io_stream, int c));
extern void output_string __ARGS((struct filter_stream *io_stream, char *s));
extern void stream_flush __ARGS((struct filter_stream *io_stream));

/* local functions - public */
       void start_all_unspoolers __ARGS((int s, void *unused, void *p));
       void start_unspooler __ARGS((char *queue));
       void start_unspooler_task __ARGS((int s, void *queue, void *p));

/* local functions - private */
static void pause_queue __ARGS((struct LPDtrans *LPD));
static int  print_file __ARGS((struct LPDtrans *LPD, char *df_name, char *filename, char format, int banner));
static int  print_job __ARGS((struct LPDtrans *LPD, char *cf_name));

/* Status messages */
static char msg_checking[] = "Checking queue at %s";
static char msg_done[] = "Work done at %s";
static char msg_processing[] = "Processing, active job started %s";
static char msg_busy_device[] = "Waiting for %s to become free since %s";
static char msg_offline_device[] = "Waiting for %s to become ready since %s (offline?)";
static char msg_device_opened[] = "Printer %s output %s opened at %s";
static char msg_paused[] = "Waiting for signal to print next page at %s";



/*
 * Called upon server startup to start an unspooler for each queue
 */
void
start_all_unspoolers( s, unused, p )
int s;
void *unused;
void *p;
{
	char **qu, **oqu, *queue, *strdup();

#ifdef LPD_DEBUG
	tprintf( "start_all_printers()\n" );
	tflush();
#endif

	/*
	 * Restart the server.
	 * Start scanning the printcap for entries
	 */
	for( oqu = qu = get_queue_list(); qu && *qu; ++qu ){
		/* Spawn a printer server */
		queue = strdup( *qu );		/* child will free */
		newproc("LPDprint",2048,start_unspooler_task,0,(void *)queue,NULL,0);
		free( *qu );
	}
	if( oqu ) {
		free( *qu );	/* null entry */
		free( oqu );
	}
}

/*
 * Task-based entry point to start_unspooler
 */
void
start_unspooler_task( s, queue, p )
int s;
void *queue;
void *p;
{
	start_unspooler( (char *)queue );
}


/*
 * Procedure to start an unspooler for a given queue
 */
void
start_unspooler( queue )
char *queue;
{
	char *cp, *pcptr, *strdup(), *path, *cf_path;
	struct LPDtrans LPD;
	struct ffblk file_info;
	struct stat statbuf;
	struct job_entry *job_list;
	unsigned int job_count, job_list_size, job;
	FILE *fp;

#ifdef LPD_DEBUG
	tprintf( "start_unspooler( %s )\n", queue );
	tflush();
#endif

	if( read_printcap( &LPD, queue ) == NULL ) {
		free( queue );
		free( LPD.name );
		return;
	}
	free( queue );

	/*
	 * Setup log file
	 */
	LPD.logfp = fopen( LPD.pc->LF, "a" );

	/*
	 * locate device entry for this printer
	 */
	if( (LPD.device = get_device_entry( LPD.pc->LP )) == NULL ) {
		lpd_log( &LPD, "start_printer: device %s not supported ", LPD.pc->LP );
		goto finish;
	}

	/*
	 * locate and update status information
	 */
	if( (LPD.status = get_status_entry( LPD.name )) == NULL ) {
		lpd_log( &LPD, "start_printer: status for %s not found", LPD.name );
		goto finish;
	}

	/*
	 * If another daemon is already running, exit.
	 */
	if( LPD.status->busy ) {
		lpd_log( &LPD, "start_printer: unspooler for %s already running", LPD.name );
		goto finish;
	}

	/*
	 * If printing is disabled, exit
	 */
	if( (LPD.status->flags & PRINT_ENABLED) == 0 ) {
		lpd_log( &LPD, "start_printer: printing for %s is disabled", LPD.name );
		goto finish;
	}

	LPD.status->busy = 1;
	LPD.status->unspooler_proc = Curproc;
	set_status_message( &LPD.status->message, msg_checking, time_string() );
	LPD.status->current_job = NULL;
	LPD.status->jobno = 0;

	YIELD;

	for( ;; ) {
		job_count = get_job_list( LPD.pc->SD, &job_list );

		if( job_count == 0 )			/* no more jobs */
			break;

		/*
		 * Print each job
		 */
		for( job = 0; job < job_count; job++ ) {
			if( !(LPD.status->flags & FORCE_REQUE)
			    && (LPD.status->flags & PRINT_ENABLED) )
				print_job( &LPD, job_list[ job ].j_name );
			free( job_list[ job ].j_name );
		}
		free( job_list );
		if( (LPD.status->flags & PRINT_ENABLED) == 0 )
			break;
		if( LPD.status->flags & FORCE_REQUE ) {
			LPD.status->flags &= ~FORCE_REQUE;
			save_status( LPD.name, LPD.status->flags, LPD.status->next_seq );
		}

		YIELD;
	}

	if( LPD.status->jobno > 0 ) {
		struct filter_stream output;

		if( LPD.status->type == INTERNAL ) {
			output.type = IO_LOCAL;
			output.stream.ifp = LPD.status->device.ifp;
		} else {
			output.type = IO_FILE;
			output.stream.fp = LPD.status->device.fp;
		}
		output.device = LPD.device;
		if( LPD.pc->FQ && LPD.pc->FF )
			output_string( &output, LPD.pc->FF );
		if( LPD.pc->TR )
			output_string( &output, LPD.pc->TR );
		LPD.status->jobno = 0;

		YIELD;
	}
	close_device( &LPD );

	LPD.status->unspooler_proc = NULLPROC;
	LPD.status->busy = 0;
	set_status_message( &LPD.status->message, msg_done, time_string() );
finish:
	free( LPD.pc );
	free( LPD.name );
	if( LPD.logfp )
		fclose( LPD.logfp );
}

/*
 * Pause queue, <LPD>, and wait for signal to continue
 */
static void
pause_queue( LPD )
struct LPDtrans *LPD;
{
	set_status_message( &LPD->status->message, msg_paused, time_string() );
	LPD->status->flags |= PAUSED;
	pwait( &LPD->status->flags );	/* wait for continue signal */
	LPD->status->flags &= ~PAUSED;
	set_status_message( &LPD->status->message, msg_processing, time_string() );
}

/*
 * Print the job described in <cf_name>
 */
static int
print_job( LPD, cf_name )
struct LPDtrans *LPD;
char *cf_name;
{
	struct filter_stream output, banner_output;
	int copies = 0, banner = 0, need_wait;
	char line[81], *path, *filename = NULL, *previous_file = NULL;
	FILE *cfp, *dfp;
	FILE *test;

#ifdef LPD_DEBUG
	tprintf( "print_job( %s )\n", cf_name );
	tflush();
#endif

	if( open_device( LPD ) < 0 ) {
		lpd_log( LPD, "print_job: device (%s) could not be opened", LPD->pc->LP );
		return -1;
	}

	path = pathname( LPD->pc->SD, cf_name );

	if( (cfp = fopen( path, "r" )) == NULL ) {
		lpd_log( LPD, "print_job: control file %s not found", cf_name );
		free( path );
		return -1;
	}
	free( path );

	LPD->status->current_job = cf_name;
	set_status_message( &LPD->status->message, msg_processing, time_string() );

	YIELD;

	LPD->job = calloc( 1, sizeof( struct LPDjob ) );

	/*
	 * Get job options
	 */
	if( cf_parse( LPD, cfp ) < 0 ) {
		fclose( cfp );
		free_job( LPD->job );
		close_device( LPD );
		return -1;
	}
	YIELD;

	/*
	 * Begin job
	 */
	if( LPD->status->type == INTERNAL ) {
		output.type = IO_LOCAL;
		output.stream.ifp = LPD->status->device.ifp;
	} else {
		output.type = IO_FILE;
		output.stream.fp = LPD->status->device.fp;
	}
	output.device = LPD->device;

	if( LPD->pc->SS ) {			/* single-sheet printer */
		pause_queue( LPD );
		need_wait = 0;
	}

	if( LPD->status->jobno == 0 ) {		/* first job */
		if( LPD->pc->LD )
			output_string( &output, LPD->pc->LD );
		if( LPD->pc->FO && LPD->pc->FF )
			output_string( &output, LPD->pc->FF );
	} else if( !LPD->pc->SF && LPD->pc->FF ) {
		output_string( &output, LPD->pc->FF );
	}

	YIELD;

	if( LPD->pc->JL )			/* job leader */
		output_string( &output, LPD->pc->JL );

	/*
	 * Print each file in job
	 */
	rewind( cfp );
	while( fgets( line, 80, cfp ) != NULL ) {
		rip( line );
		if( *line == 'L' ) {
		    if( !(LPD->pc->SH || (LPD->job->noheader && LPD->pc->AB))
			&& !LPD->pc->HL ) {
				if( need_wait && LPD->pc->SS )
					pause_queue( LPD );
				else
					need_wait = 1;
				print_banner( LPD, &output, line+1 );
				banner = 1;
			}
		} else if( *line == 'N' ) {
			if( filename )
				free( filename );
			filename = strdup( line+1 );
		} else if( islower( *line ) ) {

			if( strchr( LPD->pc->FX, *line ) == NULL ) {
				lpd_log( LPD, "print_job: %c format not supported", *line );
				continue;
			}
			if( strcmp( previous_file, line+1 ) == 0 ) { /* copy? */
				if( LPD->pc->SC )	/* suppress copies? */
					break;
				else
					copies++;
			} else {			/* no copy */
				copies = 0;
				if( previous_file )
					free( previous_file );
				previous_file = strdup( line+1 );
			}

			if( LPD->pc->MC > 0 && copies > LPD->pc->MC ) {
				/* too many copies */
				lpd_log( LPD, "print_job: too many copies of %s requested", line+1 );
				break;
			}
			YIELD;

			/*
			 * Print the file
			 */
			if( need_wait && LPD->pc->SS )
				pause_queue( LPD );
			else
				need_wait = 1;
			if( LPD->pc->FL )		/* file leader */
				output_string( &output, LPD->pc->FL );

			path = pathname( LPD->pc->SD, line+1 );
			print_file( LPD, path, filename, *line, banner );
			LPD->status->jobno++;
			banner = 0;
			free( path );

			YIELD;
			if( LPD->pc->FT )		/* file trailer */
				output_string( &output, LPD->pc->FT );

		}
	}
	if( filename )
		free( filename );
	if( previous_file )
		free( previous_file );
	YIELD;

	/*
	 * Finish up job by printing banner if last
	 * and deleting specified files
	 */
	rewind( cfp );
	while( fgets( line, 80, cfp ) != NULL ) {
		rip( line );
		if( *line == 'L'
		    && (!LPD->pc->SH || LPD->pc->AB) && LPD->pc->HL ) {
			if( need_wait && LPD->pc->SS )
				pause_queue( LPD );
			else
				need_wait = 1;
			print_banner( LPD, &output, line+1 );
		} else if( *line == 'U' ) {
			path = pathname( LPD->pc->SD, line+1 );
			remove( path );
			free( path );
		}
	}
	fclose( cfp );
	YIELD;

	/*
	 * delete control file
	 */
	path = pathname( LPD->pc->SD, cf_name );
	remove( path );
	free( path );

	fclose( cfp );
	free_job( LPD->job );

	if( LPD->pc->JT )			/* job trailer */
		output_string( &output, LPD->pc->JT );

	LPD->status->current_job = NULL;

	/* Caller prints TRailer and closes print device if necessary */

	return 0;
}

/*
 * Print the given data file in the specified format
 */
static int
print_file( LPD, df_name, filename, format, banner )
struct LPDtrans *LPD;
char *df_name, *filename, format;
int banner;				/* 0 if none printed, 1 if printed */
{
	struct LPDfilter *fptr;
	struct filter_parms *parms;
	struct pr_parms *pr_parms;
	char *filter;
	char *file_mode;
	void (*process) __ARGS((int unused,void *parms,void *unused2));
	FILE *fp;
	int s[2], sync;

#ifdef LPD_DEBUG
	tprintf( "print_file( %s, %c )\n", df_name, format );
	tflush();
#endif

	switch( format ) {
		case 'f':		/* ordinary file */
		case 'p':		/* text file to print with pr(1) */
			filter = LPD->pc->IF;
			file_mode = READ_TEXT;
			break;
		case 'l':		/* text file with control characters */
			filter = LPD->pc->IF;
			file_mode = READ_BINARY;
			break;
		case 't':		/* troff(1) file */
			filter = LPD->pc->TF;
			file_mode = READ_BINARY;
			break;
		case 'n':		/* ditroff(1) file */
			filter = LPD->pc->NF;
			file_mode = READ_BINARY;
			break;
		case 'd':		/* dvi file */
			filter = LPD->pc->DF;
			file_mode = READ_BINARY;
			break;
		case 'g':		/* plot(1G) file */
			filter = LPD->pc->GF;
			file_mode = READ_BINARY;
			break;
		case 'v':		/* plain raster file */
			filter = LPD->pc->VF;
			file_mode = READ_BINARY;
			break;
		case 'c':		/* cifplot file */
			filter = LPD->pc->CF;
			file_mode = READ_BINARY;
			break;
	}
	if( LPD->pc->PW == 0 ) {		/* binary only printer */
		format = 'l';
		filter = LPD->pc->IF;
		file_mode = READ_BINARY;
	}
	if( filter == NULL ) {
		lpd_log( LPD, "print_file: filter for %cf not available", format );
		return -1;
	}

	/*
	 * locate filter info
	 */
	process = NULLVFP;
	fptr = filter_list;
	while( fptr->name != NULL ) {
		if( strcmp( fptr->name, filter ) == 0 ) {
			process = fptr->process;
			break;
		}
		fptr++;
	}
	if( process == NULLVFP ) {
		lpd_log( LPD, "print_file: filter (%s) for %cf not found", filter, format );
		return -1;
	}
	YIELD;

	/*
	 * Open data file to print
	 */
	if( (fp = fopen( df_name, file_mode )) == NULL ) {
		lpd_log( LPD, "print_file: %s could not be opened", df_name );
		return -1;
	}

	if( format == 'p' ) {		/* pr(1) filter to IF */
		pr_parms = callocw( 1, sizeof( struct pr_parms ) );

		socketpair( AF_LOCAL, SOCK_STREAM, 0, s );
		seteol( s[0], "\n" );
		seteol( s[1], "\n" );

		pr_parms->file_fp = fp;
		pr_parms->out_s = s[0];
		if( LPD->job->title )
			pr_parms->title = LPD->job->title;
		else
			pr_parms->title = filename;
		pr_parms->length = LPD->pc->PL;
		pr_parms->width = LPD->job->width;

		newproc("LPD pr filter",2048,pr_filter,0,(void *)pr_parms,NULL,0);
	}
	YIELD;

	/*
	 * Set up parameters for output filter
	 */
	parms = mallocw( sizeof( struct filter_parms ) );

	parms->printer_name = LPD->name;
	parms->width = LPD->job->width;
	parms->length = LPD->pc->PL;
	parms->literal = (format == 'l' || LPD->job->width == 0);
	parms->indent = LPD->job->indent;
	parms->class = LPD->job->class;
	parms->jobname = LPD->job->jobname;
	parms->account = LPD->job->account;
	parms->login_name = LPD->job->person;
	parms->host = LPD->job->hostname;
	parms->format = format;
	parms->af_file = LPD->pc->AF;
	parms->ps_start = LPD->pc->PS;
	parms->ps_end = LPD->pc->PE;
	if( format == 'p' ) {			/* needs input from pr(1) */
		parms->input.type = IO_SOCKET;
		parms->input.stream.socket = s[1];
	} else {				/* input from file */
		parms->input.type = IO_FILE;
		parms->input.stream.fp = fp;
	}
	parms->input.device = NULL;
	parms->pages = banner;
	if( LPD->status->type == INTERNAL ) {
		parms->output.type = IO_LOCAL;
		parms->output.stream.ifp = LPD->status->device.ifp;
	} else {
		parms->output.type = IO_FILE;
		parms->output.stream.fp = LPD->status->device.fp;
	}
	parms->output.device = LPD->device;
	parms->sync = &sync;

	newproc("LPD filter",2048,process,0,(void *)parms,NULL,0);

	pwait( &sync );
	fclose( fp );
	free( parms );
	if( format == 'p' )
		free( pr_parms );
	return 0;
}










