/*	Client routines for Network News Transfer Protocol -- RFC977
 *
 *	Copyright 1990 Anders Klemets - SM0RGV, All Rights Reserved.
 *	Permission granted for non-commercial copying and use, provided
 *	this notice is retained.
 *
 *	Changes copyright 1990 Bernie Roehl, All Rights Reserved.
 *	Permission granted for non-commercial copying and use, provided
 *	this notice is retained.
 *
 *  Revision history:
 *
 *	Jan 92	Bill Simpson -- fix leading '.' problem.
 *		Make optional time parameters, fix comments.
 *		Don't start process when low memory or wrong time.
 *		Correctly match FQDN (Fully Qualified Domain Name).
 *		Remove large numbers of duplicate trace print functions,
 *		and redirect to log rather than screen.
 *		Crude check to eliminate extra cross-posting which
 *		creates multitudes of unwanted groups.
 *		Use NNTP@host "From " line with human readable date
 *		(for View), followed by absolute seconds (for expiry).
 *		Reduce stack and heap usage.
 *		(changes Copyright 1992 William Allen Simpson)
 *
 *	July 19, 1990 pa0gri Delinted and cleaned up. (calls and includes)
 *
 *	May 11, 1990 - br checked for invalid chars in news filenames
 *
 *	May 10, 1990 - br changed date stamp in 'From ' lines to
 *            seconds since GMT (to make parsing and expiry easier)
 *
 *	May 9, 1990 - br added locking of nntp.dat and history files,
 *            second parameter to NNTP DIR, fixed bug in updating of
 *            nntp.dat
 *
 *	early May, 1990 -- br added NNTP TRACE, NNTP DIR,
 *            server-specific newsgroups and connection windows,
 *            locking of newsgroup files using mlock() and rmlock(),
 *            date stamping of 'From ' lines, increased stack space,
 *            updating of nntp.dat only on successful sessions.
 */
#include <ctype.h>
#include <stdio.h>
#include <time.h>
#include <string.h>
#ifdef	__TURBOC__
#include <dir.h>
#endif
#include "global.h"
#include "timer.h"
#include "proc.h"
#include "socket.h"
#include "usock.h"
#include "netuser.h"
#include "cmdparse.h"
#include "commands.h"
#include "dirutil.h"
#include "files.h"
#include "mailutil.h"
#include "smtp.h"

/* Tracing levels:
	0 - no tracing
	1 - serious errors reported
	2 - transient errors reported
	3 - session progress reported
	4 - complete article text reported (this is excessive)
 */
static unsigned short nntptrace = 1;

#define NNTPSTKSIZE	3072

static struct nntpservers {
	struct nntpservers *next;
	struct timer timer;
	int lowtime, hightime;  /* for connect window */
	char *hostname;
	char *groups;
};
#define NULLNNTP	(struct nntpservers *)NULL

struct nntpservers *Nntpservers = NULLNNTP;

#define MAXGROUPDIRS 10

static struct grouploc {
	char *prefix;        /* e.g. comp, rec, net, talk, alt ... */
	char *directory;     /* directory where these groups should be */
	} groupdirs[MAXGROUPDIRS] = { NULL, NULL };

static char *Nntpgroups = NULLCHAR;

static char *News_spool = NULL;
static int np_all = 0;  /* non-zero if Newsdir is a malloc'ed space */

static char *validchars = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
			  "abcdefghijklmnopqrstuvwxyz"
			  "0123456789-_";

static void nntptick __ARGS((void *tp));
static void nntp_job __ARGS((int i1,void *tp,void *v1));
static int getreply __ARGS((int s,char *rbuf));
static int getarticle __ARGS((int s,char *tbuf,struct nntpservers *np,char *msgid));

static int dogroups __ARGS((int argc,char *argv[],void *p));
static int doadds __ARGS((int argc,char *argv[],void *p));
static int dodrops __ARGS((int argc,char *argv[],void *p));
static int dokicks __ARGS((int argc,char *argv[],void *p));
static int dolists __ARGS((int argc,char *argv[],void *p));
static int donntrace __ARGS((int argc,char *argv[],void *p));
static int dondir __ARGS((int argc,char *argv[],void *p));

static struct cmds Nntpcmds[] = {
	"addserver",    doadds,         0, 2, "nntp addserver <nntpserver>"
			" [<seconds>] [hh:mm-hh:mm] [<grouplist>]",
	"directory",    dondir,         0, 0, NULLCHAR,
	"dropserver",   dodrops,        0, 2, "nntp dropserver <nntpserver>",
	"groups",       dogroups,       0, 0, NULLCHAR,
	"kick",         dokicks,        0, 2, "nntp kick <nntpserver>",
	"list",         dolists,        0, 0, NULLCHAR,
	"trace",        donntrace,      0, 0, NULLCHAR,
	NULLCHAR,
};


int
donntp(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	return subcmd(Nntpcmds,argc,argv,p);
}


static int donntrace(argc, argv, p)
int argc;
char *argv[];
void *p;
{
	return setshort(&nntptrace,"NNTP tracing",argc,argv);
}

        
static int
doadds(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	struct nntpservers *np;
	char *fullname = domainsuffix(argv[1]);
	int i;

	for (np = Nntpservers; np != NULLNNTP; np = np->next)
		if(stricmp(np->hostname,fullname) == 0)
			break;

	if (np == NULLNNTP) {
		np = (struct nntpservers *) callocw(1,sizeof(struct nntpservers));
		np->hostname = fullname;
		np->next = Nntpservers;
		Nntpservers = np;
		np->groups = NULLCHAR;
		np->lowtime = np->hightime = -1;
		np->timer.func = nntptick;  /* what to call on timeout */
		np->timer.arg = (void *)np;
	} else {
		free( fullname );
	}

	i = 2;
	if ( argc > i
	  && isdigit(*argv[i])
	  && strchr(argv[i],'-') == NULLCHAR ) {
		/* set timer duration */
		set_timer(&np->timer,atol(argv[i++])*1000L);
	}

	if ( argc > i
	  && isdigit(*argv[i]) ) {
		int lh, ll, hh, hl;
		sscanf(argv[i++], "%d:%d-%d:%d", &lh, &ll, &hh, &hl);
		np->lowtime = lh * 100 + ll;
		np->hightime = hh * 100 + hl;
	}

	if ( argc > i ) {
		if (np->groups == NULLCHAR) {
			np->groups = mallocw(RLINELEN);
			*np->groups = '\0';
		}
		for ( ; i < argc; ++i) {
			if ((strlen(np->groups)+strlen(argv[i])+2) >= RLINELEN) {
				tprintf("Group list too long!  Group '%s' ignored!\n", argv[i]);
			} else {  /* it's a group, and it fits... add it to list */
				if (*np->groups != '\0')
					strcat(np->groups, ",");
				strcat(np->groups, argv[i]);
			}
		}
		if (*np->groups == '\0') {	/* No groups specified? */
			free(np->groups);
			np->groups = NULLCHAR;
		}
	}
	start_timer(&np->timer);	    /* and fire it up */
	return 0;
}


static int
dodrops(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	struct nntpservers *np, *npprev = NULLNNTP;
	char *fullname = domainsuffix(argv[1]);

	for (np = Nntpservers; np != NULLNNTP; npprev = np, np = np->next) {
		if(strnicmp(np->hostname,fullname,strlen(fullname)) == 0) {
			stop_timer(&np->timer);
			free(np->hostname);
			free(np->groups);

			if(npprev != NULLNNTP)
				npprev->next = np->next;
			else
				Nntpservers = np->next;
			free((char *)np);
			return 0;
		}
	}
	tprintf("No such server enabled.\n");
	return -1;
}


static int
dolists(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	struct nntpservers *np;
	for (np = Nntpservers; np != NULLNNTP; np = np->next) {
		char xbuf[80];
		if (np->lowtime != -1 && np->hightime != -1)
			sprintf(xbuf, " %02d:%02d-%02d:%02d",
				np->lowtime/100,
				np->lowtime%100,
				np->hightime/100,
				np->hightime%100);
		else
			xbuf[0] = '\0';
		tprintf("%-32s (%lu/%lu%s) %s\n",
			np->hostname,
			read_timer(&np->timer) /1000L,
			dur_timer(&np->timer) /1000L,
			xbuf,
			np->groups ? np->groups : "");
	}
	return 0;
}


static int
dokicks(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	struct nntpservers *np;
	char *fullname = domainsuffix(argv[1]);

	for (np = Nntpservers; np != NULLNNTP; np = np->next) {
		if(strnicmp(np->hostname,fullname,strlen(fullname)) == 0) {
			/* If the timer is not running, the timeout function
			 * has already been called, and we don't want to call
			 * it again.
			 */
			if ( np->timer.duration == 0
			  || run_timer(&np->timer) ) {
				stop_timer(&np->timer);
				newproc("NNTP client", NNTPSTKSIZE, nntp_job,
					0, np, NULL,0);
			}
			return 0;
		}
	}
	tprintf("No such server enabled.\n");
	return -1;
}


static int dondir(argc, argv, p)
int argc;
char *argv[];
void *p;
{
	if (argc < 2) {
		int i;
		tprintf("spool: %s\n", News_spool ? News_spool : Mailspool);
		tprintf("control: %s\n", Newsdir);
		for (i = 0; i < MAXGROUPDIRS; ++i)
			if (groupdirs[i].prefix)
				tprintf("%-10.10s %s\n",
					groupdirs[i].prefix,
					groupdirs[i].directory);
	} else {
		char *p;
		if ((p = strchr(argv[1], '=')) != NULLCHAR) {
			/* set a groupdir */
			int i;
			*p++ = '\0';
			for (i = 0; i < MAXGROUPDIRS; ++i)
				if (groupdirs[i].prefix)
					if (!strnicmp(groupdirs[i].prefix, argv[1], strlen(argv[1]))) {
						if (groupdirs[i].directory) {
							free(groupdirs[i].directory);
							groupdirs[i].directory = NULLCHAR;
							}
						if (*p == '\0') {
							free(groupdirs[i].prefix);
							groupdirs[i].prefix = NULLCHAR;
						} else
							groupdirs[i].directory = strdup(p);
						return 0;
					}
			if (*p == '\0')  /* trashing a group that's not there */
				return 0;
			for (i = 0; i < MAXGROUPDIRS; ++i){
				if (groupdirs[i].prefix == NULLCHAR) {
					groupdirs[i].prefix = strdup(argv[1]);
					if (groupdirs[i].directory) {
						free(groupdirs[i].directory);
						groupdirs[i].directory = NULL;
					}
					groupdirs[i].directory = strdup(p);
					return 0;
				}
			}
			tprintf("Directory table full\n");
		}
		else {  /* no '=', so just set default */
			if (News_spool)
				free(News_spool);
			News_spool = strdup(argv[1]);
		}
		if (argc > 2) {  /* they specified a newsdir as well */
			if (np_all)
				free(Newsdir);
			Newsdir = strdup(argv[2]);
			np_all = 1;
		}
	}
	return 0;
}
	

static int
dogroups(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	int i;
	if(argc < 2) {
		if ( Nntpgroups == NULLCHAR
		  || strcmp(Nntpgroups,"*") == 0)
			tprintf("All groups are currently enabled.\n");
		else
			tprintf("Currently enabled newsgroups:\n%s\n",
				Nntpgroups);
		return 0;
	}
	if(Nntpgroups == NULLCHAR)
		Nntpgroups = mallocw(RLINELEN);
	*Nntpgroups = '\0';
	for(i=1; i < argc; ++i) {
		if(i > 1)
			strcat(Nntpgroups,",");
		strcat(Nntpgroups,argv[i]);
	}
	return 0;
}


/* This is the routine that gets called every so often to connect to
 * NNTP servers.
 */
static void
nntptick(tp)
void *tp;
{
	struct nntpservers *np = (struct nntpservers *) tp;
	struct tm *ltm;
	time_t t;
	int now;

	if(availmem() - NNTPSTKSIZE < Memthresh){
		/* Memory is tight, don't do anything */
		if (nntptrace >= 2)
			log(-1,"NNTP tick exit -- low memory");
		start_timer(&np->timer);
		return;
	}

	time(&t);	/* more portable than gettime() */
	ltm = localtime(&t);
	now = ltm->tm_hour * 100 + ltm->tm_min;
	if (np->lowtime < np->hightime) {  /* doesn't cross midnight */
		if (now < np->lowtime || now >= np->hightime) {
			if (nntptrace >= 2)
				log(-1,"NNTP window to '%s' not open", np->hostname);
			start_timer(&np->timer);
			return;
		}
	} else {
		if (now < np->lowtime && now >= np->hightime) {
			if (nntptrace >= 2)
				log(-1,"NNTP window to '%s' not open", np->hostname);
			start_timer(&np->timer);
			return;
		}
	}

	newproc("NNTP client", NNTPSTKSIZE, nntp_job, 0, tp, NULL,0);
}


static void
nntp_job(i1,tp,v1)
int i1;
void *tp, *v1;
{
	struct nntpservers *np = (struct nntpservers *) tp;
	struct sockaddr_in fsocket;
	struct tm *ltm;
	time_t t;
	FILE *fp, *tmpf;
	int s = -1, i;
/*	long pos; */
	char rbuf[RLINELEN];
	char tbuf[TLINELEN];
	char *cp;
	char *lastdate = NULLCHAR;
	char *want_groups;

	if ( (fsocket.sin_addr.s_addr = resolve(np->hostname)) == 0L ) {
		/* No IP address found */
		if (nntptrace >= 1)
			log(-1,"NNTP can't resolve host '%s'", np->hostname);
		start_timer(&np->timer);
		return;
	}

	fsocket.sin_family = AF_INET;
	fsocket.sin_port = IPPORT_NNTP;

	s = socket(AF_INET,SOCK_STREAM,0);
	sockmode(s,SOCK_ASCII);

	if(connect(s,(char *)&fsocket,SOCKSIZE) == -1){
		cp = sockerr(s);
		if (nntptrace >= 2)
			log(s,"NNTP Connect failed: %s",
				cp != NULLCHAR ? cp : "");
		goto quit;
	}

	log(s,"NNTP Connected to mailhost %s", np->hostname);

	/* Eat the banner */
	i = getreply(s,rbuf);
	if(i == -1 || i >= 400) {
		if (nntptrace >= 1)
			log(s,"NNTP negative reply on banner (%d)",i);
		goto quit;
	}

	if (mlock(Newsdir, "nntp")) {
		if (nntptrace >= 2)
			log(s,"NNTP cannot lock %s", "nntp");
		goto quit;
	}

	sprintf(rbuf,"%s/nntp.dat",Newsdir);
	if((fp = fopen(rbuf,APPEND_TEXT)) == NULLFILE) {
		if (nntptrace >= 1)
			log(s,"NNTP cannot open %s",rbuf);
		rmlock(Newsdir, "nntp");
		goto quit;
	}
	rewind(fp);

/*	for(pos=0L; fgets(rbuf,RLINELEN,fp) != NULLCHAR;pos=ftell(fp)) { */
	while ( fgets(rbuf,RLINELEN,fp) != NULLCHAR ) {
		if((cp = strchr(rbuf,' ')) == NULLCHAR)
			continue;	/* something wrong with this line, skip it */
		*cp = '\0';
		if(strnicmp(rbuf,np->hostname,strlen(rbuf)) == 0) {
			rip(cp+1);
			lastdate = strdup(cp+1);
			break;
		}
	}
	fclose(fp);
	rmlock(Newsdir, "nntp");

	if(lastdate == NULLCHAR)
		lastdate = strdup("700101 000000");

	/* snapshot the time for use later in re-writing nntp.dat */
	time(&t);
	ltm = localtime(&t);
				
	/* Get a list of new message-id's */
	if (np->groups) {
		want_groups = np->groups;
	} else if (Nntpgroups != NULLCHAR) {
		want_groups = Nntpgroups;
	} else {
		want_groups = "*";
	}
	if (nntptrace >= 3)
		log(s,"==>NEWNEWS %s %s",
			want_groups,
			lastdate);
	usprintf(s,"NEWNEWS %s %s\n",
		want_groups,
		lastdate);
	free(lastdate);

	/* Get the response */
	if((i = getreply(s,rbuf)) != 230) { /* protocol error */
		if (nntptrace >= 1)
			log(s,"NNTP protocol error (%d)",i);
		goto quit;
	}
	if((tmpf = tmpfile()) == NULLFILE) {
		if (nntptrace >= 1)
			log(s,"NNTP cannot create %s", "tmp file");
		goto quit;
	}
	if ( recvmail(s,rbuf,RLINELEN,tmpf,nntptrace) == -1 ) {
		fclose(tmpf);
		goto quit;
	}

	/* Open the history file */
	if (mlock(Newsdir, "history")) {
		if ( nntptrace >= 2 )
			log(s,"NNTP cannot lock %s", "history");
		fclose(tmpf);
		goto quit;
	}
	sprintf(rbuf,"%s/history",Newsdir);
	if((fp = fopen(rbuf,APPEND_TEXT)) == NULLFILE) {
		if ( nntptrace >= 1 )
			log(s,"NNTP cannot open %s",rbuf);
		fclose(tmpf);
		rmlock(Newsdir, "history");
		goto quit;
	}

	/* search through the history file for matching message id's */
	rewind(tmpf);
	while(fgets(rbuf,RLINELEN,tmpf) != NULLCHAR) {
		int want = TRUE;

		rewind(fp);
		while(fgets(tbuf,RLINELEN,fp) != NULLCHAR) {
			if(stricmp(tbuf,rbuf) == 0) {
				want = FALSE;
				break;
			}
			pwait(NULL);
		}
		if (want) {
			/* not found, get the article */
			if(getarticle(s,tbuf,np,rbuf) == -1) {
                                if (nntptrace >= 2)
					log(s,"NNTP could not get article");
				fclose(fp);
				fclose(tmpf);
				rmlock(Newsdir, "history");
				goto quit;
			}
			fprintf(fp,"%s",rbuf); /* add the new message id */
		}
	}
	fclose(fp);
	fclose(tmpf);
	rmlock(Newsdir, "history");

	if (nntptrace >= 3)
		log(s,"==>QUIT");
	usprintf(s,"QUIT\n");

	/* Eat the response */
	getreply(s,rbuf);

	/* NOW, update the nntp.dat file */
	while (mlock(Newsdir, "nntp")) {
		pause( 2000L );
	}

	sprintf(rbuf,"%s/nntp.dat",Newsdir);
	fp = fopen(rbuf,READ_TEXT);

	sprintf(tbuf, "%s/nntp.tmp",Newsdir);
	if ((tmpf = fopen(tbuf, WRITE_TEXT)) == NULLFILE)
		if (nntptrace >= 1)
			log(s,"NNTP cannot create %s",tbuf);

	if (fp == NULLFILE || tmpf == NULLFILE) {
		if (nntptrace >= 2)
			log(s,"NNTP cannot update %s",rbuf);
		if (fp)
			fclose(fp);
		if (tmpf)
			fclose(tmpf);
		rmlock(Newsdir, "nntp");
		goto quit;
	}

	while (fgets(rbuf, RLINELEN, fp)) {
		if (strnicmp(rbuf, np->hostname, strlen(np->hostname)) != 0)
			fputs(rbuf, tmpf);
	}

	fprintf(tmpf,"%s %02d%02d%02d %02d%02d%02d\n",
		np->hostname,
		ltm->tm_year%100,
		ltm->tm_mon+1,
		ltm->tm_mday,
		ltm->tm_hour,
		ltm->tm_min,
		ltm->tm_sec);
	fclose(fp);
	fclose(tmpf);
	pwait(NULL);	/* give other tasks a chance */

	sprintf(rbuf, "%s/nntp.dat", Newsdir);
	sprintf(tbuf, "%s/nntp.tmp", Newsdir);
	unlink(rbuf);
	rename(tbuf, rbuf);
	rmlock(Newsdir, "nntp");

quit:
	log(s,"NNTP daemon exiting");
	close_s(s);

	/* Restart timer */
	start_timer(&np->timer);
	return;
}


static int
getreply(s, rbuf)
int s;
char *rbuf;
{
	int response;

	while(recvline(s,rbuf,RLINELEN) != -1) {
		/* skip informative messages and blank lines */
		if(rbuf[0] == '\0' || rbuf[0] == '1')
			continue;
		sscanf(rbuf,"%d",&response);
		if (nntptrace >= 3) {
			rip(rbuf);
			log(s,"<==%s", rbuf);
		}
		return response;
	}
	if (nntptrace >= 3)
		log(s,"***No response");
	return -1;
}

static int
getarticle(s,tbuf,np,msgid)
int s;
char *tbuf;		/* scratch TLINELEN buffer */
struct nntpservers *np;
char *msgid;
{
	FILE *fp;
	FILE *tfp;
	char *groupline = NULLCHAR;
	char *group_p;
	char *nextgroup;
	int r;

	if (nntptrace >= 3)
		log(s,"==>ARTICLE %s", msgid);
	usprintf(s,"ARTICLE %s", msgid);

	r = getreply(s,tbuf);
	if(r == -1 || r >= 500) {
		return -1;
	}
	if(r >= 400) {
		return 0;
	}

	if((tfp = tmpfile()) == NULLFILE) {
		if (nntptrace >= 1)
			log(s,"NNTP cannot create %s", "temp file for article");
		return -1;
	}
	if(recvmail(s,tbuf,TLINELEN,tfp,nntptrace) == -1) {
		fclose(tfp);
		return -1;
	}

	rewind(tfp);
	while(fgets(tbuf,TLINELEN,tfp) != NULLCHAR) {
		if(strncmp(tbuf,"Newsgroups: ",12) == 0) {
			rip(tbuf);
			groupline = strdup( &tbuf[12] );
			break;
		}
		if ( strcmp(tbuf,"\n") == 0 )
			break;
	}

	strcpy(tbuf,msgid);	/* Get a copy we can munge */
	rip(tbuf);		/* remove trailing new-line */

	if ( groupline == NULLCHAR ) {
		if ( nntptrace >= 2 ) {
			log(s,"NNTP article %s missing Newsgroups",tbuf);
		}
		fclose(tfp);
		return 0;
	}

	if ( nntptrace >= 2 ) {
		log(s,"NNTP article %s in %s",tbuf,groupline);
	}

	/* Under the current file arrangement, each article is copied
	 * into every cross-posted group file.	This wastes a lot of space.
	 * In an attempt to ameliorate the loss, we perform a simple check
	 * on cross-posting to keep from creating unwanted sub-trees. {WAS}
	 */
	for ( group_p = groupline; group_p != NULLCHAR;
					group_p = nextgroup ) {
		char *pattern_p;
		char *nextpattern;
		char *cp;

		/* ignore leading spaces */
		while ( *group_p == ' ' ) {
			group_p++;
		}

		/* copy group */
		if ( (nextgroup = strchr( group_p, ',' )) != NULLCHAR ) {
			*nextgroup++ = '\0';
		}

		/* check to see if this group is in the group list */
		for ( pattern_p = Nntpgroups; pattern_p != NULLCHAR;
						pattern_p = nextpattern ) {
			if ( (nextpattern = strchr( pattern_p, ',' )) != NULLCHAR ) {
				*nextpattern = '\0';
				if ( wildmat( group_p, pattern_p, NULLCHARP ) ) {
					*nextpattern++ = ',';
					break;
				}
				*nextpattern++ = ',';
			} else {
				if ( wildmat( group_p, pattern_p, NULLCHARP ) )
					break;
			}
		}

		if ( Nntpgroups != NULLCHAR  &&  pattern_p == NULLCHAR ) {
			/* group not in grouplist */
			if (nntptrace >= 3)
				log(s,"NNTP news group '%s'"
					" not in '%s'",
					group_p,
					Nntpgroups);
			continue;
		}

                /* reuse nextpattern and pattern_p */
		nextpattern = News_spool ? News_spool : Mailspool;

		/* convert groupname to pathname */
		for ( cp = group_p; *cp != '\0'; cp++ ) {
			if ( strchr(validchars, *cp) != NULLCHAR ) {
				/* do nothing */
			} else if ( *cp == '.' ) {
				*cp = '\0';
				/* create a subdirectory, if necessary */
                                pattern_p = pathname(nextpattern, group_p);
#ifdef __TURBOC__
				mkdir(pattern_p);
#else
				mkdir(pattern_p,0755);
#endif
                                free( pattern_p );
				*cp = '/';
			} else {
				*cp = '_';
			}
		}

		pattern_p = pathname( nextpattern, group_p );
		strcpy( tbuf, pattern_p );	/* temporary copy */

		if ((cp = strrchr(pattern_p, '/')) != NULLCHAR) {
			*cp++ = '\0';
			group_p = strdup( cp ); /* save copy of file name */
		}

		if (mlock(pattern_p,group_p)) {
			if (nntptrace >= 2)
				log(s,"NNTP cannot lock %s", tbuf);
			free(group_p);
			free(pattern_p);
			continue;
		}
		strcat(tbuf,".txt");

		/* open the news file */
		if (nntptrace >= 3)
			log(s,"NNTP Writing article to %s", tbuf);

		if((fp = fopen(tbuf,APPEND_TEXT)) != NULLFILE) {
			time_t tt;
			char *ct;

			/* looks like normal mail header,
			 * plus message identifier and GMT time
			 */
			time(&tt);
			ct = ctime(&tt);
			rip(ct);
			strcpy(tbuf,msgid);	/* Get a copy we can munge */
			rip(tbuf);		/* remove trailing new-line */
			fprintf(fp,"From NNTP@%s %s %s %ld\n",
				np->hostname,
				ct,
				tbuf,
				tt);

			rewind(tfp);
			while(fgets(tbuf,TLINELEN,tfp) != NULLCHAR) {
				fputs(tbuf,fp);
				pwait(NULL);
			}
			putc('\n',fp);
			fclose(fp);
		}
		rmlock(pattern_p, group_p);
		free(group_p);
		free(pattern_p);
	}
	free( groupline );
	fclose(tfp);
	return 0;
}


