/* This module implements the serial line framing method used by
 * net/rom nodes.  This allows the net/rom software to talk to
 * an actual net/rom over its serial interface, which is useful
 * if we want to do packet switching for multi-line wormholes.
 *
 * Copyright 1992 William Allen Simpson
 *	complete re-write to match PPP design
 *	more efficient, elegant, no static structures
 *	partly based on a previous design by Dan Frank, W9NK
 *	which was based on the SLIP design by Phil Karn, KA9Q
 */
#include <stdio.h>
#include "global.h"
#include "mbuf.h"
#include "iface.h"
#include "ax25.h"
#include "nrs.h"
#include "trace.h"
#include "commands.h"

static int nrs_raw __ARGS((struct iface *iface, struct mbuf *data));
static void nrs_recv __ARGS((int dev,void *v1,void *v2));
static void nrs_free __ARGS((struct iface *));

#define NRS_INTER	0	/* between packets */
#define NRS_INTRA	1	/* seen STX, within packet */
#define NRS_ESCAPE	2	/* seen DLE, while in packet */
#define NRS_CSUM	3	/* seen ETX, waiting for checksum */


/* Send a raw net/rom serial frame */
static int
nrs_raw(ifp,bp)
struct iface *ifp;
struct mbuf *bp;
{
	struct mbuf *vbp;
	register char *cp;
	register int c;
	unsigned char csum = 0;

	dump(ifp,IF_TRACE_OUT,CL_AX25,bp);
	ifp->OutOctets += len_p(bp);
	ifp->rawsndcnt++;
	ifp->lastsent = secclock();

	/* Allocate output mbuf that's twice as long as the packet.
	 * This is a worst-case guess (consider a packet full of STX's!)
	 * Add five bytes for STX, ETX, checksum, and two nulls.
	 */
	if ( (vbp = alloc_mbuf((int16)(2*len_p(bp) + 5))) == NULLBUF ) {
		ifp->OutDiscards++;
		free_p(bp);
		return -1;
	}
	cp = vbp->data;

	*cp++ = STX;

	/* Copy input to output, escaping special characters */
	while ((c = PULLCHAR(&bp)) != -1) {
		switch(c){
		case STX:
		case ETX:
		case DLE:
			*cp++ = DLE;
			/* fallthru */
		default:
			*cp++ = c;
		}
		csum += c;
	}

	/* Tie off the packet */
	*cp++ = ETX;
	*cp++ = csum;
	*cp++ = NUL;
	*cp++ = NUL;
	vbp->cnt = cp - vbp->data;

	if (ifp->trace & IF_TRACE_RAW)
		raw_dump(ifp,IF_TRACE_OUT /* -1? */,vbp);
	return (*ifp->iosend)(ifp->d_index,vbp);
}


/****************************************************************************/
/* Packetize net/rom input from device */
/* (process started by nrs_init) */
static void
nrs_recv(dev,v1,v2)
int dev;
void *v1;
void *v2;
{
	struct iface *ifp = v1;
	byte_t csum = 0;
	struct mbuf *raw_bp = NULLBUF;
	register struct mbuf *head_bp = NULLBUF;
	char *cp;			/* next byte in mbuf */
	register int c;
#define MODE_TOSS	0x01
#define MODE_ESCAPED	0x02
#define MODE_CHECKSUM	0x04
	int mode = MODE_TOSS;

	while ( (c = (*ifp->ioget)(dev)) != -1 ) {
		if ( raw_bp != NULLBUF	&&  raw_bp->cnt < raw_bp->size ) {
			*raw_bp->data++ = c;
			raw_bp->cnt++;
		}

		/*
		 * (The order of the following tests is important.)
		 */
		if ( mode & MODE_ESCAPED ) {
			mode &= ~MODE_ESCAPED;
		} else if ( mode & MODE_CHECKSUM ) {
			if ( raw_bp != NULLBUF ) {
				raw_bp->data = (char *)(raw_bp + 1);
				if ( raw_bp->cnt != 1 ) {
					raw_dump( ifp, IF_TRACE_IN, raw_bp );
				}
				if ( ifp->trace & IF_TRACE_RAW ) {
					raw_bp->cnt = 0;
				} else {
					free_p(raw_bp);
					raw_bp = NULLBUF;
				}
			} else if ( ifp->trace & IF_TRACE_RAW ) {
				raw_bp = alloc_mbuf( 2 * ifp->mru);
			}

			if ( uchar(c) == csum ) {
				net_route( ifp, CL_AX25, head_bp );
			} else {
				free_p(head_bp);	/* bad checksum */
				ifp->InErrors++;
			}
			head_bp = NULLBUF;
			mode = MODE_TOSS;
			continue;
		} else if ( c == STX ) {
			if ( raw_bp != NULLBUF ) {
			} else if ( ifp->trace & IF_TRACE_RAW ) {
				raw_bp = alloc_mbuf( 2 * ifp->mru);
			}

			if ( !(mode & MODE_TOSS) ) {
				/* If we see an STX in a packet,
				 * assume the current packet was trashed.
				 */
				free_p(head_bp);
				ifp->InErrors++;
			}

			/* setup for next buffer */
			head_bp = NULLBUF;
			csum = 0;
			mode = FALSE;
			continue;
		} else if ( c == DLE ) {
			/* before TOSS so we correctly toss escapes */
			mode |= MODE_ESCAPED;
			continue;
		} else if ( mode & MODE_TOSS ) {
			/* before ETX so we don't test checksums */
			continue;
		} else if ( c == ETX ) {
			mode |= MODE_CHECKSUM;
			continue;
		}

		/* We reach here with a byte for the buffer.
		 * Make sure there is room for it.
		 */
		if ( head_bp == NULLBUF ) {
			if ( (head_bp = alloc_mbuf( ifp->mru )) == NULLBUF ) {
				ifp->InDiscards++;
				mode |= MODE_TOSS;
				continue;
			}
			cp = head_bp->data;
			head_bp->cnt = 0;
		} else if ( head_bp->cnt >= head_bp->size ) {
			/* too much data */
			free_p(head_bp);
			ifp->InErrors++;
			head_bp = NULLBUF;
			mode |= MODE_TOSS;
			continue;
		}

		/* Store the byte, increment counts */
		*cp++ = c;
		head_bp->cnt++;
		csum += uchar(c);	/* add to checksum */
	}

	/* clean up afterward */
	free_p(head_bp);
	ifp->rxproc = NULLPROC;
}


/* display status of active net/rom serial interfaces */
int
donrstat(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	struct iface *ifp;

	tprintf("Interface      RcvBytes   RcvPackets  Errors\n");

	for (ifp = Ifaces;ifp != NULLIF;ifp = ifp->next) {
		if ( ifp->e_type == CL_AX25 ) {
			tprintf(" %-15s %10lu,%10lu,%6u\n",
				ifp->name,
				ifp->InOctets,
				ifp->InUnicast,
				ifp->InErrors);
		}
	}
	return 0;
}


/****************************************************************************/
/* Initialize control structures for a NRS serial interface */
/* (called from setencap) */
int
nrs_init(ifp)
struct iface *ifp;
{
	char *ifn;

	ifp->edv = NULL;
	ifp->raw = nrs_raw;
	ifp->detach = nrs_free;

	ax25_init(ifp);

	ifp->rxproc = newproc( ifn = if_name( ifp, " receive" ),
			256, nrs_recv, ifp->d_index, ifp, NULL, 0);
	free(ifn);
	return 0;
}


static void
nrs_free(ifp)
struct iface *ifp;
{
	alert( ifp->rxproc, -1 );

	ifp->edv = NULL;
}

