/*
   rc.c : Copyright Paul Healy, EI9GL, 1992.

   Derived from bm. Resource filehandling - initialises the internal
   bm user environment.

   Copyright 1986 Bdale Garbee, All Rights Reserved.
   Permission granted for non-commercial copying and use, provided
   this notice is retained.
   Copyright 1987 1988 Dave Trulli NN2Z, All Rights Reserved.
   Permission granted for non-commercial copying and use, provided
   this notice is retained.

   911217 : Added this header
   920828 : Modifed the tok array to allow for bsearch'ing
   920906 : Changed where the resource file will be looked for (slightly)
   921108 : Added check for spooldir, add offer to make missing dirs
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "config.h"
#include "rc.h"
#include "pc.h"
#include "misc.h"
#include "buf.h"
#include "mailfile.h"
#include "header.h"
#include "nntp.h"
#include "mkpath.h"

static int parsefile(char *name);

/*
 * Look for the rc file in:
 *  . given by the environment variable BMHRC
 *  . given by the environment variable BMRC
 *  . in the current directory
 *  . in the directory given by the HOME environment variable
 *  . in the / directory
 *                                (look in the directory that bmh.exe is in ?)
 * Return NULL if the file can't be found.
 */
char *
getruncom(char *runcom)
{
   char *p;
   /*
    *  check for BMHRC/BMRC in the ENV
    */
   if ((p = getenv("BMHRC")) != NULL)
      return strcpy(runcom, p);
   else if ( (p = getenv("BMRC")) != NULL ) 
      return strcpy(runcom, p);
   else {
      /*
       * Now try guessing likely places
       */
      if (access(RUNCOM, 0) == 0)
         return strcpy(runcom, RUNCOM);

      if ((p = getenv("HOME")) != NULL) {       /* Try $HOME/RUNCOM */
         sprintf(runcom, "%s/%s", p, RUNCOM);
         if (access(runcom, 0) == 0)
            return runcom;
         }

      sprintf(runcom, "/%s", RUNCOM);
      if (access(runcom, 0) == 0)
         return runcom;
      }
      
   return NULL;
}

/*
 * commands, and their defaults, which are valid in bm.rc / .bmrc
 */
static struct tok {
   char str[11],
       *val;
   UserVar variable;
} rccmds[] = {  
  "alias",    "/alias",        alias,    /* alias file */
  "bufsize",  BUFSIZE,         bufsize,  /* internal mail file buffer */
  "columns",  "80",            maxcol,   /* number of screen character columns */
  "current",  "/current.bmh",  current,  /* file holding current folder/msg */
#ifdef DAYFIRST
  "dayfirst", "y",             dayfirst, /* ddmmyy instead of mmddyy */
#else
  "dayfirst", "n",             dayfirst, /* mmddyy instead of ddmmyy */
#endif
  "edit",     NULL,            editor,   /* user's favorite text editor */
  "editline", "y",             editline, /* call DOS fn 0A for console input */
  "folder",   NULL,            folder,   /* directory for saving read mail */
  "fullname", NULL,            fullname, /* fullname of user */
  "host",     NULL,            hostname, /* name of this host */
  "index",    "n",             index,    /* re-create jnos-style index after a change to mail file */
  "maxlet",   "300",           maxlet,   /* maximum number of msgs in a file */
  "mbox",     "mbox",          savebox,  /* name of the mbox text file */
  "mqueue",   "/spool/mqueue", mqueue,   /* defined mqueue outbound directory */
  "multitask","y",             multitask,/* try and give up timeslices to a multitasker */
  "newsdir",  "/spool/news",   newsdir,  /* news directory */
  "newshost", NULL,            newshost, /* news posting hostname for path line */
  "organ",    NULL,            organ,    /* nntp organization header */
  "quote",    ": ",            quote,    /* string to prefix quoted message */
  "record",   NULL,            record,   /* record outbound mail in this file */
  "reply",    NULL,            replyto,  /* address for reply-to header */
  "rewrite",  NULL,            rewrite,  /* rewrite file */
  "rows",     "24",            maxrow,   /* number of screen character rows */
  "screen",   "bios",          screen,   /* screen access mode under Turbo C */
  "signature",NULL,            signature,/* location of sig file */
  "smtp",     "/spool/mail",   maildir,  /* mail directory */
  "source",   NULL,            source,   /* pull in rc info from another file */
  "spooldir", "/spool",        spooldir, /* spool directory - areas file etc */
  "swap",     "n",             swap,     /* msdos: swap out bmh program when editing or listing */
  "tmpdir",   NULL,            tmpdir,   /* directory for temporary files */
  "tsbids",   "y",             tsbids,   /* add time stamp to bids in history file */
  "user",     NULL,            username, /* name of this user from rc file */
  "viewer",   NULL,            viewer,   /* program to view files with */
  "zone",     NULL,            zone      /* timezone variable to override automatic choice(s) */
};

static const int rccmdsize = sizeof(rccmds) / sizeof(struct tok);

/*
 * return the rc type from a line of the configuration file
 */
static int
rctype(char *s)
{
   /*
    * remove trailing colons here
    */
   int len = strlen(s);

   if (len > 0) {
      struct tok *p;

      if (s[len-1] == ':')
         s[len-1] = 0;
   
      p = bsearch(s, rccmds, rccmdsize, sizeof(struct tok), (CMPFN) strcmp);

      return p == NULL ? -1 : p->variable;
      }
   else
      return -1;
}

/*
 * check for important settings
 */
static int
checkrc(UserVar u)
{
   if (rccmds[u].val == NULL) {
      fprintf(stderr, "bmh: %s not set\n", rccmds[u].str);
      return -1;
      }

   switch (u) {
      case spooldir:
      case maildir:
      case mqueue:
         if(access(rccmds[u].val, 0))
            return makepath(rccmds[u].val);
         break;
      case hostname:
         break;
      case screen:
         setvideo(rccmds[u].val);
         break;
      case source:
         if (parsefile(rccmds[u].val) == -1)
            return -1;
         break;
      default:
         break;
      }
   return 0;
}

static int
parserc(char *s)
{
   int type;
   char *argv[MAXARGS];

   if ( (parse(s, argv, MAXARGS) < 2 ) || (type = rctype(argv[0])) == -1) {
      fprintf(stderr, "bmh: invalid command '%s'\n", argv[0]);
      return -1;
      }
   else {
      rccmds[type].val = strdup(argv[1]);
      if (checkrc(type) == -1)
         return -1;
      }
   return 0;
}

static int
parsefile(char *name)
{
   FILE *rcfp;
   char rcline[LINELEN];

   if ( (rcfp = fopen(name, "r")) == NULL) {
      fprintf(stderr, "bmh: cannot open '%s', check your installation\n", name);
      return -1;
   }

   while (fgets(rcline, sizeof(rcline), rcfp) != NULL) {
      if (*rcline == '#' || *rcline == '\n' ||
          *rcline == '\0' || *rcline == ';')
         continue;
      rip(rcline);
      if (parserc(rcline) == -1) {
         fclose(rcfp);
         return -1;
         }
   }
   return fclose(rcfp);
}

static int NotLoaded = 1;

char *
getrc(UserVar u)
{
   if (NotLoaded)
      loadconfig();

   return rccmds[u].val;
}

/*
 * returns -1 if config file couldn't be parsed correctly. 0 otherwise.
 */
int
loadconfig(void)
{
   char runcom[MAXNAMLEN];

   if (!NotLoaded)
      return 0;
   else
      NotLoaded = 0;

   if (getruncom(runcom) == NULL) {
      fprintf(stderr, "bmh: can't find rc file\n");
      return -1;
      }

   if (parsefile(runcom) == -1)
      return -1;
   
   if ( (checkrc(maildir) == -1)  ||
        (checkrc(mqueue) == -1)   ||
        (checkrc(hostname) == -1) ||
        (checkrc(spooldir) == -1)
      )
      return -1;

   return 0;
}

