/*			RSPF - Radio Shortest Path First
 *			    Anders Klemets, SM0RGV.
 *  Several modifications by Lars Pettersson, SM6RPZ. d8lars@dtek.chalmers.se
 *	    This code may be used for non-commercial purposes only.
 *
 * 890918 - Version 2.0  SM0RGV
 * 900816 - Version 2.1  SM0RGV
 * 920820 - Version 2.1a SM6RPZ
 *
 * History of modifications by SM6RPZ -
 * Changes also made to the ax25 routing table and the arp
 * table (added interface etc, se some comments below).
 * 1992-05-23 - Small changes here and there.
 *	      - A try to correct the behaviour of partial updates with
 *		sequence numbers greater than the old sequence number
 *		and subsequence number greater than 0, i.e. we have lost
 *		the full update with the new sequence. This caused a loss of
 *		all routes going through the reporting router. A better
 *		approach would be to keep the old routes and keep the old
 *		sequence and subsequence number, this is done now.
 *	      - The code will now change interface for stations hopping
 *		between our interfaces. This CAN cause oscillation if we
 *		receive the same station on two or more interfaces.
 * 1992-05-25 - Added interface to ax25 route and the arp table.
 * 1992-05-26 - RSPF_RTIME changed to ipReasmTimeout (realy not needed).
 * 1992-05-27 - rr_delete() delteted :-) It was used in rspfcheck() to delete
 *		a router entry, but it's better not to delete the entry
 *		because we might reach the router through another router.
 *		If not, it will be deleted in makeroutes().
 * 1992-06-05 - Added different horizons for different routes.
 *	      - Added the function rt_clookup(). This function looks in the
 *		routing table for a direct route with less then 32 bits
 *		covering the target.
 * 1992-06-06 - Fixed problem when a station has a manual ARP entry on another
 *		interface than the interface used now. This caused problem
 *		before, the code tried to ping the station on the wrong
 *		interface.
 * 1992-06-27 - RSPF refuse list added. Will stop RRH:s and updates from
 *		stations on the list.
 * 1992-07-08 - rspfping(): If an adjacency won't answer our pings when the
 *		suspecttimer has timed out we will wait 60 seconds and then
 *		delete the entry. A bad news update will then also be sent.
 *		In the old code it took suspectimer seconds before this
 *		happend (normaly 900 seconds).
 *	      - goodbadnews(): If an adjacencies address is included in our
 *		group route a good/bad news bulletin will not be sent out.
 *		Se section IV.6. in k1io's protocol.
 * 1992-07-13 - Moved interface check from rspfmain() to rspf_input().
 *	      - Changed rspfarpupcall() into rspfupcall() and removed
 *		rspfrouteupcall. Rspfupcall now used for all upcalls.
 *	      -	Riface is put on queue instead of iface, we already had
 *		this one handy in rspfupcall() and rspf_input().
 * 1992-07-14 - Riface added to adjacency structure.
 *	      - rt_add() etc. in rspfmain() moved to rspfping().
 *	      - rtif_lookup() changed into riface_lookup(). It now only
 *		checks if a given interface is a RSPF interface.
 *	      - Added adj_lookup(). It has two parameters, you can
 *		either look for an address or an interface. It returns
 *		the adjacency structure.
 * 1992-07-16 - In add_rspfadj(). If we find an adjacency on another
 *		interface, but with an inferior cost and the old one is
 *		in the OK-state, we don't bother about it. It can't be
 *		better than the one we have.
 *	      - A new route is only added in rspfping() if it realy is needed.
 *	      - Test of rt_merge() in makeroutes(), seemes to work ok but it
 *		would be better to never add routes already covered by another
 *		route with less bits, it would save some CPU time.
 * 1992-07-28 - In add_rspfadj(). If we already have a manual, 32 bit route
 *		pointing to another interface to the adjacency we don't
 *		accept the new one.
 * 1992-07-29 - When a update has been received makeroutes() is only called,
 *		if needed, once at the end of update_in() and not after
 *		each processed router bulletin as before.
 * 1992-07-30 - Some more changes to node_in(). If the station sending us a
 *		routing update has old information about us we will send it
 *		our latest information. In the old code we would only send
 *		them new information if the sequence number was larger.
 *	      -	Restructure of add_rspfadj() (a new switch-statement).
 * 1992-07-31 - Some changes to the subsequence handling. Full routing updates
 *		should always have a subsequence of zero and only partial
 *		updates should have other values.
 *	      -	In makeadjupdate() the rr->subseq value was used before,
 *		this was wrong, it should use the node-header value instead
 *		(which should be zero). If the latest subsequence is greater
 *		than 0, it could otherwise be misstaken for a partial update
 *		advertising a new route and cause extra load on the network!
 * 1992-08-01 - Can't I ever get this right?! Changes to node_in(). If the
 *		first update we receive is a partial one and the next is a
 *		full one, or a partial update with lower subsequence, we will
 *		swap data and merge the old update into the new update.
 *	      -	In makeadjupdate() we must make sure that the updates we send
 *		out are full ones otherwise we will cause extra load on the
 *		network. If the first update received is partial update we now
 *		inform our other router adjacencies about this (was not done
 *		before).
 *	      - Moved initiation of timer.func and timer.arg to add_rspfadj().
 *		If we run rspfcheck and had no more Nrspfproceses available we
 *		started a timer without any function and argument!
 * 1992-08-02 - Changes to partialupdate(). If the stored update is a partial
 *		one, we also save bad adjacencies (cost = 255). This will
 *		enable us to remove bad adjacencies when we receive the full
 *		update (see above). Note thet these will not be sent on, the
 *		changes to makeadjupdate() make sure of this (see above).
 * 1992-08-09 - Added rspfiface structure to update_in() to sense new routers
 *		a bit faster. If we receive a update from a unknown router
 *		we make a upcall, which will run checkrspf(). If we receive a
 *		fragmented update we will use NULLRIFACE instead as we might
 *		have received the fragments on different interfaces (is this
 *		possible?), and we will not do a upcall.
 * 1992-08-10 - In partialupdate() we always make sure that the new adjacency
 *		we are about to merge into our old data has a equal or higher
 *		horizon than any former entry, otherwise we use the old one.
 *	      - We use this in node_in() when we have received a packet	with
 *		the same sequence and subsequence (equal to zero) as our old
 *		data. This will make sure of that we aways have the routes with
 *		the longest horizons.
 * 1992-08-14 - The order of our adjacencies in makeownupdate() have been
 *		reversed. This will ensure that the horizon check in node_in()
 *		realy finds the one with the longest horizon.
 * 1992-08-18 - Cost and heard count saved if we receive an ARP packet or RRH
 *		on the same interface as the stored adjacency.
 *	      - Heard count set in rspfmain() only if we receive a RRH packet.
 * 1992-08-26 - Released to ucsd.edu
 * 1992-10-13 - Fixed bug when sending updates, horizon was not decremented.
 *	      - Fixed bug in partialupdate(), link count was not updated
 *		correctly (old links was counted several times...)
 * 1992-10-16 - Fixed bug in sendrrh(). Data was not duplicated so we got
 *		invalid frees.
 *	      -	sendonerspf() removed and associated code put into the
 *		calling code.
 * 1992-10-30 -	Updates to be sent to a update sending station (answers to
 *		polls etc.) are all appended and sent first when all incoming
 *		updates have been processed. Changes in update_in() and
 *		node_in(). This will minimize the update traffic.
 * 1992-11-02 - Horizon for routers are now always set to Router_horizon.
 * 1992-11-03 - Sequence number incremented in rspfcheck() instead of
 *		makeownupdate(). Makeownupdate() called only once in
 *		rspfcheck().
 * 1993-04-14 - BIG BUG corrected in makeroutes(). rp could be undefined
 *		in the for-statement removing all non-manual routes. We
 *		therefore can't refere to it with rp = rp->next. We first
 *		have to save rp->next in rp1 and then free rp and use rp1
 *		to recurse through the list.
 */
#include "global.h"
#include "config.h"
#include "mbuf.h"
#include "proc.h"
#include "timer.h"
#include "netuser.h"
#include "internet.h"
#include "pktdrvr.h"
#include "ip.h"
#include "iface.h"
#include "ax25.h"
#include "arp.h"
#include "icmp.h"
#include "socket.h"
#include "rspf.h"

/* If defined we will update the suspecttimer each time we hear the station. */
#undef	RSPF_HEARD	1

struct rspf_stat Rspf_stat;
struct rspfreasm *Rspfreasmq = NULLRREASM;
struct rspfiface *Rspfifaces = NULLRIFACE;
struct rspfadj *Adjs = NULLADJ;
struct rspfrouter *Rspfrouters = NULLRROUTER;
struct rspf_refuse *Rspf_refuse = NULLREFUSE;	/* sm6rpz */
struct mbuf *Rspfinq = NULLBUF;
struct timer Rspfreasmt, Susptimer;

char *Rrh_message = NULLCHAR;
int Keeprouter = 0;
int Nrspfproc = 0;
short Rspfseq = -32768L + 1;
unsigned char Rspfsubseq = 0;
unsigned short Rspfpingmax = 5;
unsigned char Router_horizon = 1;	/* sm6rpz */
unsigned char Portable_horizon = 1;	/* sm6rpz */
unsigned char Group_horizon = 1;	/* sm6rpz */
unsigned char Local_horizon = 1;	/* sm6rpz */
unsigned char Non_rspf_horizon = 0;	/* sm6rpz */
int16 Rspf_trace = 0;			/* sm6rpz */

static int16 Envelopeid = 0;

static char *findlowroute __ARGS((int32 addr,int *low,int add,int32 *resaddr,int *resbits));
static void makeroutes __ARGS((void));
static void partialupdate __ARGS((struct rspfrouter *rr,struct mbuf *bp));
static void rspfcsum __ARGS((struct mbuf **bpp,int32 dest));
static void update_in __ARGS((struct mbuf *bp,int32 addr,struct rspfiface *riface));
static int node_in __ARGS((struct mbuf *bp,int32 addr,int fragment,struct mbuf **update,int *nodecnt));
static void sendtoall __ARGS((struct mbuf *bp,int nodecnt,struct rspfiface *riface));
static int sendupdate __ARGS((struct mbuf *bp,int nodecnt,int32 addr));
static int isnewer __ARGS((short a,short b));
static void del_reasm __ARGS((struct rspfreasm *re));
static void reasmtimeout __ARGS((void *t));
static struct mbuf *rspfreasm __ARGS((struct mbuf *bp,struct rspfpacketh *rph,int32 addr));
static struct mbuf *fragmenter __ARGS((struct mbuf *bp,int nodes,int16 mtu));
static struct mbuf *makeadjupdate __ARGS((int32 addr,struct rspfrouter *rr));
static void pushadj __ARGS((struct mbuf **bpp,int32 addr,int bits));
static void sendrrh __ARGS((struct rspfiface *ri));
static void sendrspf __ARGS((void));
static void goodbadnews __ARGS((struct rspfadj *adj));
static void add_rspfadj __ARGS((struct rspfadj *adj));
static void rspfcheck __ARGS((struct rspfadj *adj));
static void rspfping __ARGS((int notused,void *p,void *v));
static void sendpartial __ARGS((int bad,void *p,void *v));  /* sm6rpz */
static struct route *rt_clookup __ARGS((int32 target));	    /* sm6rpz */
static struct rspfadj *adj_lookup __ARGS((int32 addr,struct iface *iface));/* sm6rpz */

/* IP passes its RSPF packets to this function */
void
rspf_input(iface,ip,bp,rxbroadcast)
struct iface *iface;
register struct ip *ip;
register struct mbuf *bp;
int rxbroadcast;
{
    struct rspfiface *riface;
    struct pseudo_header ph;	/* Pseudo-header for checksumming */

    if(bp == NULLBUF)
        return;
    if(ismyaddr(ip->source)) {
	free_p(bp);
	return;			/* Ignore our own packets if they get here */
    }
    if((riface = riface_lookup(iface)) == NULLRIFACE) {
	/* Not a RSPF interface, or RSPF main loop not running */
        free_p(bp);
	if(Rspfifaces != NULLRIFACE)
	    ++Rspf_stat.norspfiface;
	return;			
    }
    ph.length = ip->length - IPLEN - ip->optlen;
    ph.source = ip->source;
    ph.dest = ip->dest;
    ph.protocol = ip->protocol;
    if(cksum(&ph,bp,ph.length)) {
        free_p(bp);		/* Checksum failed, ignore packet completely */
        ++Rspf_stat.badcsum; 
        return;
    }
    bp = pushdown(bp,1 + sizeof(ip->source) + sizeof(riface));
    *bp->data = RSPFE_PACKET;
    memcpy(bp->data + 1,&ip->source,sizeof(ip->source));
    memcpy(bp->data + 1 + sizeof(ip->source),&riface,sizeof(riface));
    enqueue(&Rspfinq,bp);
}

/* Main loop in RSPF process */
void
rspfmain(v,v1,v2)
int v;
void *v1,*v2;
{
#ifdef	AX25
    register struct arp_tab *atp;
    register struct lq *alp;
#endif
    register struct rspf_refuse *rl;	/* The refuse list -- sm6rpz */
    union rspf rspf;			/* Internal packet structure */
    struct rspfadj *adj;		/* Adjacency */
    struct mbuf *bp, *tbp;
    struct rspfiface *riface;
    int32 source;			/* Source address (or whatever...) */
    int cmd;

    for(;;) {
	if(Rspfinq == NULLBUF)
	    pwait(&Rspfinq);
	bp = dequeue(&Rspfinq);		/* at least 5 bytes */
	cmd = PULLCHAR(&bp);		/* get internal event indicator */
	pullup(&bp,(char *)&source,sizeof(source));
	switch(cmd) {
	case RSPFE_RRH:			/* It is time to send a new RRH */
	    sendrrh((struct rspfiface *)source);
	    continue;
	case RSPFE_CHECK:		/* An adjacency has becomed suspect */
	    rspfcheck((struct rspfadj *)source);
	    continue;
	case RSPFE_UPDATE:		/* Time to send a new routing update */
	    sendrspf();
	    continue;
	case RSPFE_UPCALL:		/* Upcall from ARP, route add etc. */
	    add_rspfadj((struct rspfadj *)source);
	    continue;
	case RSPFE_PACKET:		/* A RSPF packet was received */
	    pullup(&bp,(char *)&riface,sizeof(riface));
	    if(ntohrspf(&rspf,&bp) == -1) {
	        free_p(bp);
	        continue;
	    }
	    switch(rspf.hdr.type) {	/* packet types */
	    case RSPF_RRH:
		free_p(bp);
		/* Refuse if router is on the refuse list -- sm6rpz */
		for(rl = Rspf_refuse; rl != NULLREFUSE; rl = rl->next)
		    if(rl->target == source)
			break;
		adj = (struct rspfadj *)callocw(1,sizeof(struct rspfadj));
		if(rl != NULLREFUSE) {	/* Refuse the packet. */
		    ++Rspf_stat.refused;
		    if(Rspf_trace > 0)
			printf("RSPF RRH refused from %s\n", inet_ntoa(source));
		    /* Proceed as if an ARP request. */
		} else {		/* Accept the packet. */
		    ++Rspf_stat.rrhin;
		    if(Rspf_trace > 0)
			printf("RSPF RRH received from %s on %s\n",
			    inet_ntoa(source), riface->iface->name);
		    adj->seq = rspf.rrh.seq;
		    adj->tos = (rspf.rrh.flags & RSPFMODE) ? DELAY:RELIABILITY;
#ifdef	AX25
		    /* Try to find the AX25 callsign and update the heard count */
		    if((atp = arp_lookup(ARP_AX25,source,riface->iface)) != NULLARP
		       && atp->state == ARP_VALID && dur_timer(&atp->timer)
		       && (alp = al_lookup(riface->iface,atp->hw_addr,0)) != NULLLQ
		      )
			adj->heard = alp->currxcnt;
#endif
		}
		adj->addr = source;
		adj->riface = riface;
		adj->cost = riface->quality;	/* Default cost */
		adj->state = RSPF_TENTATIVE;
		add_rspfadj(adj);
		continue;
	    case RSPF_FULLPKT:
		++Rspf_stat.updatein;
		if(Rspf_trace > 0)
		    printf("RSPF update received from %s on %s\n",
			    inet_ntoa(source), riface->iface->name);
		/* Feed the packet through the reassembly queue */
		if((tbp = rspfreasm(bp,&rspf.pkthdr,source)) != NULLBUF)
		    update_in(tbp,source,riface);
		continue;
	    }
	    break;
	}
	free_p(bp);			/* Clean up. */
    }
}

/* This function is called every time a RRH should be broadcasted. A RRH
 * is sent on the interface given by ri or on every RSPF interface. The
 * broadcast address of the interface must be routed onto the same interface
 * using a static route, i.e.
 * ifc <ifc> broadcast <broadcast_address>
 * otherwise there will be no RRH sent on that interface. Broadcast_address
 * must be different on all interfaces but the exact address is not that
 * important, it is never used, all packets sent to the layer 2 broadcast
 * address is considered to be a broadcast.
 */
static void
sendrrh(ri)
struct rspfiface *ri;		    /* Interface to use, if specified. */
{
    struct pseudo_header ph;
    struct mbuf *bp, *data = NULLBUF, *tmp;
    struct rspfiface *riface;
    struct route *rp;
    struct rrh rrh;
    struct arp_tab *ap;		    /* sm6rpz */
    int i;			    /* sm6rpz */

    for(riface = Rspfifaces; riface != NULLRIFACE; riface = riface->next) {
	if(ri != NULLRIFACE && riface != ri)
	    continue;
	if((rp = rt_lookup(riface->iface->broadcast)) != NULLROUTE &&
	  rp->iface == riface->iface) {
	    rrh.version = RSPF_VERSION;
	    rrh.addr = riface->iface->addr;
	    rrh.seq = riface->iface->rawsndcnt;
	    /* Default is to use the same mode as the interface */
	    rrh.flags = (Rspfownmode == -1) ?
			    !(riface->iface->flags & CONNECT_MODE) :
			    !(Rspfownmode & CONNECT_MODE);
	    ph.length = 0;
	    if(Rrh_message != NULLCHAR) {
		data = ambufw(strlen(Rrh_message));
		memcpy(data->data,Rrh_message,strlen(Rrh_message));
		ph.length = data->cnt = strlen(Rrh_message);
	    }
	    ph.length += RRHLEN;
	    ph.source = riface->iface->addr;
	    ph.protocol = RSPF_PTCL;
	    /* This code will send RRH's to all valid, not published,
	     * manual entries with a manual route or an automatic route
	     * routed to the same interface as the ARP entry. By doing
	     * so we will sense routers working via digipeaters or netrom
	     * (they will also send RRH's the same way as we do).
	     */
	    for(i = 0; i < HASHMOD; ++i) {	/* scan ARP table */
		for(ap = Arp_tab[i]; ap != NULLARP; ap = ap->next) {
		    if(ap->iface == riface->iface
		       && ap->state == ARP_VALID
		       && !ap->pub
		       && !dur_timer(&ap->timer)
 		       && (rp = rt_lookup(ap->ip_addr)) != NULLROUTE
		       && ((rp->flags & RTPRIVATE) && !dur_timer(&rp->timer)
		           || (rp->gateway == 0L || rp->gateway == rp->target)
		           && rp->iface == ap->iface
			  )
		      ) {
			if(dup_p(&tmp,data,0,len_p(data)) != len_p(data))
			    free_p(tmp);
			else {
			    ph.dest = ap->ip_addr;
			    ++rrh.seq;
			    bp = htonrrh(&rrh,tmp,&ph);
			    ++Rspf_stat.rrhout;
			    ip_send(riface->iface->addr,ap->ip_addr,RSPF_PTCL,
				    0,2,bp,0,0,0);
			}
		    }
		}
	    }
	    ph.dest = riface->iface->broadcast;
	    ++rrh.seq;
	    bp = htonrrh(&rrh,data,&ph);
	    ++Rspf_stat.rrhout;
	    ip_send(riface->iface->addr,riface->iface->broadcast,RSPF_PTCL,
			0,2,bp,0,0,0);
	}
    }
}

/* This function is called whenever an RRH packet has been
 * received or when a new entry is added to the ARP table.
 */
static void
add_rspfadj(adj)
register struct rspfadj *adj;
{
    register struct rspfadj *oldadj, *tmp;
    register struct route *rp;
    register int dif, origdif;
#ifndef	RSPF_HEARD
    register long int oldtimer;
#endif

    if(adj == NULLADJ)
	return;				/* sanity, but it happens... */

    if((rp = rt_blookup(adj->addr,32)) != NULLROUTE
       && (rp->flags & RTPRIVATE)
       && !dur_timer(&rp->timer)
       && rp->iface != adj->riface->iface
       && rp->metric < adj->cost
      ) {
        /* We already have a manual, 32 bit route with
	 * better cost on another interface, so we are
	 * not interested in this new one. -- sm6rpz
         */
        free((char *)adj);
        return;
    }

    /* Insert the new adjacency. */
    adj->next = Adjs;
    Adjs = adj;

    /* Find the previous copy of the adjacency, if there was any. */
    for(oldadj = Adjs; oldadj->next != NULLADJ; oldadj = oldadj->next)
	if(oldadj->next->addr == adj->addr) {
	    tmp = oldadj->next;
	    if(tmp->riface != adj->riface) {
		if(tmp->state == RSPF_OK && tmp->cost < adj->cost) {
		    /* We already have a better route to this station on
		     * another interface, so we are not interested in this
		     * new one. -- sm6rpz
		     */
		    Adjs = adj->next;	/* Restore Adjs pointer. */
		    free((char *)adj);
		    return;
		}
	    } else {			/* Save some of the old values. */
		adj->cost = tmp->cost;
		if(!adj->seq) {		/* Not a RRH, save these too. */
		    adj->seq = tmp->seq;
		    adj->tos = tmp->tos;
		    adj->heard = tmp->heard;
		}
	    }
#ifndef	RSPF_HEARD
	    /* Save the old adjacency timer value. */
	    oldtimer = read_timer(&tmp->timer);
#endif
	    stop_timer(&tmp->timer);
	    /* Remove the old adjacency from the list. */
	    oldadj->next = tmp->next;
	    oldadj = tmp;
	    break;
	}

    if(oldadj->addr != adj->addr || oldadj == adj)
	oldadj = NULLADJ;		/* No previous copy found */

    /* Fill in timer values. */
    adj->timer.func = rspfevent;
    adj->timer.arg = (void *)&adj->timer;
    set_timer(&adj->timer, dur_timer(&Susptimer));

    switch(adj->state) {
    case RSPF_TENTATIVE:
	/* We have received a RRH, an ARP request or
	 * added an ARP entry or a route via gateway.
	 */
	if(oldadj != NULLADJ) {
	    adj->okcnt = oldadj->okcnt;
#ifdef	AX25
	    if(oldadj->riface == adj->riface
	       && oldadj->seq && adj->seq != oldadj->seq
	       && adj->heard && oldadj->heard
	      ) {
		/* The wrapping of the modulus is taken into account here */
		origdif = (oldadj->seq > (MAXINT16 - 100) && adj->seq < 100) ?
		    adj->seq + MAXINT16 - oldadj->seq :	adj->seq - oldadj->seq;
		dif = (oldadj->heard > (MAXINT32 - 100L) && adj->heard < 100L)?
		    adj->heard + MAXINT32 - oldadj->heard :
		    adj->heard - oldadj->heard;
		/* At this point, "origdif" equals the difference in sequence
		 * numbers between the two latest RRH packets, i.e. the number
		 * of AX.25 frames the station has sent since the last RRH
		 * packet we heard. "dif" equals the number of AX.25 frames
		 * that we actually heard from that station during the same
		 * period. This algorithm can be improved, see 2.1 spec.
		 */
		if(origdif > 0 && dif >= 0 && origdif >= dif)
		    adj->cost = max(adj->riface->quality,
				 adj->riface->quality * 2 -
				 (1 + (adj->riface->quality * dif) / origdif));
	    }
#endif	/* AX25 */
	    switch(oldadj->state) {
	    case RSPF_BAD:
		/* If old state was Bad, don't announce this adjacency until
		 * it is known to be OK, to prevent announcing an adjacency
		 * with a state transition loop such as OK -> Suspect -> Bad ->
		 * Tentative -> Suspect -> Bad -> Tentative -> ... 
		 */
		adj->okcnt = 0;
		rspfcheck(adj);			/* Enter Suspect state. */
		break;
	    case RSPF_TENTATIVE:
		start_timer(&oldadj->timer);	/* Start the old timer again */
		/* Fall through. */
	    case RSPF_SUSPECT:
		/* PINGer process is running (or will be). */
		if(oldadj->riface == adj->riface && adj->seq && adj->seq != oldadj->seq) {
		    /* Update only if we have received a RRH and the interface
		     * is the same as the old one. RRH's always give tos and
		     * seq so update them also. Cost will only be updated if
		     * the cost has become worse. PING has to prove that the
		     * route is good until we change the cost to the better.
		     */
		    oldadj->seq = adj->seq;
		    oldadj->tos = adj->tos;
		    oldadj->heard = adj->heard;
		    if(oldadj->cost < adj->cost)
			oldadj->cost = adj->cost;
		}
		oldadj->next = adj->next;	/* adj is at start of chain. */
		Adjs = oldadj;
		free((char *)adj);
		return;
	    case RSPF_OK:
		adj->state = RSPF_OK;		/* Remain in OK state. */
		/* We have here received a RRH or an ARP request. These do not
		 * give us any information if the route is bidirectional. But
		 * according to k1io's protocol we should restart the suspect-
		 * timer in those situations. Se section II.3 etc. (code is
		 * also added in update_in() )  One problem exist, manually
		 * added routes and ARP's shoudn't realy count here, that
		 * should be fixed somehow. -- sm6rpz
		 *
		 * Because the code isn't fully working on heard only routes
		 * this section is not defined (RSPF_HEARD not defined).
		 */
#ifndef	RSPF_HEARD
		/* Use the time left on the old timer if it hasn't expired. */
		if(oldtimer) {
		    set_timer(&adj->timer, oldtimer);
		    start_timer(&adj->timer);
		    /* Set duration to Susptimer so we will get the right
		     * duration the next time we restart the timer.
		     */
		    set_timer(&adj->timer, dur_timer(&Susptimer));
		} else			    /* Old timer have expired. */
#endif
		    start_timer(&adj->timer);
		break;
	    }
	} else {			    /* oldadj == NULLADJ, new one! */
	    rspfcheck(adj);		    /* Enter Suspect state. */
	    makeroutes();		    /* Routing table may change. */
	}
	break;
    case RSPF_OK:			    /* We have received an ARP reply. */
	if(oldadj != NULLADJ) {
	    adj->okcnt += oldadj->okcnt;
	    if(oldadj->state == RSPF_SUSPECT) {
		killproc(oldadj->pinger);
		--Nrspfproc;		    /* Bug fix - N1BEE */
	    }
	}
	if(oldadj == NULLADJ || oldadj->riface != adj->riface)
	    makeroutes();		    /* Routing table may change. */
	if(adj->okcnt == 1)		    /* A previously unkown route... */
	    goodbadnews(adj);		    /* ...that is good news :-) */
	start_timer(&adj->timer);
	break;
    }
    free((char *)oldadj);
}

/* Take appropriate action if an adjacency is Bad or Tentative */
static void
rspfcheck(adj)
register struct rspfadj *adj;
{
    register struct rspfadj *prev;

    if(adj == NULLADJ)
	return;

    switch(adj->state) {
    case RSPF_OK:
	adj->state = RSPF_TENTATIVE;	/* Fall through. */
    case RSPF_TENTATIVE:
	/* Wait for half the time of suspecttimer if problems occur - sm6rpz */
	set_timer(&adj->timer, dur_timer(&Susptimer) / 2L);
	if(Nrspfproc < RSPF_PROCMAX)
	    adj->pinger = newproc("RSPF ping",256,rspfping,0,adj,NULL,0);
	else				/* Try later */
	    start_timer(&adj->timer);
	break;
    case RSPF_BAD:
	adj->cost = 255;
	if(adj->okcnt)
	    goodbadnews(adj);		/* This is bad news... :-( */
	for(prev = Adjs; prev != NULLADJ; prev = prev->next)
	    if(prev->next == adj)
		break;
	if(prev != NULLADJ)
	    prev->next = adj->next;	/* Unlink */
	else
	    Adjs = adj->next;
	stop_timer(&adj->timer);	/* just in case */
	free((char *)adj);		/* delete the adjacency */
	makeroutes();			/* update the routing table */
	break;
    }
}

/* Send a ping to a suspect or tentative adjacency. */
static void
rspfping(notused,p,v)
int notused;
void *p, *v;
{
    int s, fromlen, pcnt;
    struct route *rp;
    struct icmp icmp;
    struct sockaddr_in from;
    struct mbuf *bp;
    struct rspfadj *adj = (struct rspfadj *)p;

    adj->state = RSPF_SUSPECT;		/* We have entered suspect state... */
    Nrspfproc++;			/* ...and started a new process. */

    pause(((ptol(p) & 7) + 1) * 1000L);	/* Pause for 1 to 8 seconds */

    /* If we don't have an entry in the routing table for this station, or
     * if the entry is on the wrong interface (routes on other interfaces
     * with lower costs are handled in add_rspfadj()) or is via a gateway
     * not equal to the target then add a correct, private, route.
     *
     * These PINGs should probably be changed to go directly to layer 2,
     * bypassing IP. In that case we need no new routes here.
     */
    if((rp = rt_lookup(adj->addr)) == NULLROUTE
       || rp->iface != adj->riface->iface
       || rp->gateway != 0L && rp->gateway != rp->target
      ) {
        if(Rspf_trace > 0) {
	    printf("RSPF new route added to %s/32 on %s with cost %u\n",
		    inet_ntoa(adj->addr),
		    adj->riface->iface->name,
		    adj->riface->quality);
	}
	rt_add(adj->addr,32,0L,adj->riface->iface,(int32)adj->riface->quality,0L,1);
    }

    if((s = socket(AF_INET,SOCK_RAW,ICMP_PTCL)) == -1) {
	--Nrspfproc;
	start_timer(&adj->timer);
	adj->state = RSPF_TENTATIVE;
	return;
    }

    fromlen = sizeof(from);
    for(pcnt = 0; pcnt < Rspfpingmax; ++pcnt) {
	pingem(s, adj->addr, 0, (int16)s, 0);
	/* Now collect the reply */
	alarm(60 * 1000L);	/* Let each ping timeout after 1 minute. */
	for(;;) {
	    if(recv_mbuf(s,&bp,0,(char *)&from,&fromlen) == -1) {
		if(errno == EALARM)	/* We timed out */
		    break;
		alarm(0L);
		close_s(s);
		--Nrspfproc;
		start_timer(&adj->timer);
		adj->state = RSPF_TENTATIVE;
		return;
	    }
	    ntohicmp(&icmp,&bp);
	    free_p(bp);
	    if(icmp.type != ICMP_ECHO_REPLY
	       || from.sin_addr.s_addr != adj->addr
	       || icmp.args.echo.id != s)
		continue;		/* Ignore other people's responses */
	    alarm(0L);
	    close_s(s);
	    if(++adj->okcnt == 1)
		goodbadnews(adj);	/* Good news :-) */
	    --Nrspfproc;
	    set_timer(&adj->timer,dur_timer(&Susptimer));
	    start_timer(&adj->timer);
	    adj->state = RSPF_OK;	/* Finally change state */
	    return;
	}
    }
    close_s(s);				/* We give up. */
    --Nrspfproc;
    set_timer(&adj->timer,60 * 1000L);	/* Wait 1 minute before deleting */
    start_timer(&adj->timer);
    adj->state = RSPF_BAD;		/* Mark the adjacency as bad. */
}

/* Upcalls come in two flavours:
 * 1. If an ARP reply has been received we know for certain that
 *    bidirectional communication is possible, i.e. reply = 1.
 * 2. If an ARP request is received, or if we have manually added an ARP
 *    entry or a route entry is added via a gateway we have yet to
 *    determine if the link is bidirectional by pinging, i.e. reply = 0.
 *
 * Called by:	arp_input()	    in arp.c
 *		doarpadd()	    in arpcmd.c ("arp add" command)
 *		doadd()		    in ipcmd.c  ("route add" command)
 *		rspfsuspect()	    in rspf.c
 *		update_in()	    in rspf.c
 */
void
rspfupcall(addr,hardware,iface,reply)
int32 addr;			/* Address of the station causing the upcall */
int16 hardware;			/* Hardware type or NHWTYPES */
struct iface *iface;		/* Interface used */
char reply;			/* 1 if reply, 0 otherwise -- sm6rpz */
{
    struct rspfadj *adj;
    struct mbuf *bp;
    struct rspfiface *riface;

    if(ismyaddr(addr))
	return;			/* Ignore our own packets if they get here */

    /* Is it coming from a RSPF interface? */
    if((riface = riface_lookup(iface)) == NULLRIFACE)
	return;
    /* Proceed only if the hardware type is relevant for
     * the interface onto which IP datagrams are routed.
     */
    switch(hardware) {
    case ARP_NETROM:
	if(riface->iface->type != CL_NETROM)
	    return;
	break;
    case ARP_AX25:
	if(riface->iface->type != CL_AX25)
	    return;
	break;
    case ARP_ETHER:
	if(riface->iface->type != CL_ETHERNET)
	    return;
	break;
    case NHWTYPES:
	break;			/* "Any interface type is ok" type entry */
    default:
	return;
    }
    if((adj = (struct rspfadj *)calloc(1,sizeof(struct rspfadj))) == NULLADJ)
	return;
    adj->addr = addr;
    adj->riface = riface;
    adj->cost = riface->quality;/* default cost */
    if(reply) {			/* ARP reply received */
	adj->state = RSPF_OK;
	adj->okcnt++;
    } else	/* An ARP request or manual ARP or route via gateway entry. */
	adj->state = RSPF_TENTATIVE;
    if((bp = alloc_mbuf(1 + sizeof(adj))) == NULLBUF) {
	free((char *)adj);
	return;
    }
    *bp->data = RSPFE_UPCALL;
    memcpy(bp->data + 1, &adj, sizeof(adj));
    bp->cnt = bp->size;
    enqueue(&Rspfinq,bp);
}

/* When the suspect timer expires, we scan through the routing table
 * for all manual 32 bit specific routes through a gateway that are
 * not an adjacency, and calls rspfupcall(). A similar thing is
 * done for all manual ARP entries. This will make sure that a station
 * that was not reachable when the "route add" or "arp add" command
 * was executed will eventually be discovered if it joins the network.
 */
void
rspfsuspect(t)
void *t;
{
    register struct route *rp;
    register struct arp_tab *ap;
    register int i;

    for(i = 0; i < HASHMOD; i++) 	/* Check the routing table */
	for(rp = Routes[31][i]; rp != NULLROUTE; rp = rp->next)
	    if(!(rp->flags & RTPRIVATE) && !dur_timer(&rp->timer)
	       && rp->gateway != 0L && rp->gateway != rp->target
	       && adj_lookup(rp->target,NULLIF) == NULLADJ
	      )
		rspfupcall(rp->target,NHWTYPES,rp->iface,0);
    for(i = 0; i < HASHMOD; ++i) 	/* scan the ARP table */
	for(ap = Arp_tab[i]; ap != NULLARP; ap = ap->next)
	    if(!ap->pub && ap->state == ARP_VALID && !dur_timer(&ap->timer)
	       && adj_lookup(ap->ip_addr,NULLIF) == NULLADJ
	      )
		rspfupcall(ap->ip_addr,ap->hardware,ap->iface,0);
    start_timer(&Susptimer);		/* restart the timer */
}

/* This function will check if the given interface is used for RSPF. If it
 * is, it returns the matching rspfiface structure or NULLRIFACE if not.
 */
struct rspfiface *
riface_lookup(iface)
register struct iface *iface;
{
    register struct rspfiface *riface;

    for(riface = Rspfifaces; riface != NULLRIFACE; riface = riface->next)
	if(riface->iface == iface)
	    return riface;
    return NULLRIFACE;
}

/* Send good or bad news partial routing updates. A cost of 255
 * should be interpreted as bad news by the remote station.
 */
static void
goodbadnews(adj)
register struct rspfadj *adj;
{
    struct rspfnodeh nodeh;
    struct rspflinkh linkh;
    register struct rspfiface *riface;
    register struct rspfrouter *rr;
    struct mbuf *bp, *tbp, *wbp;
    int nodecnt = 1;

    linkh.horizon =
	(adj->seq || rr_lookup(adj->addr) != NULLRROUTER) ? Router_horizon :
	(rt_clookup(adj->addr) == NULLROUTE) ? Portable_horizon :
	Local_horizon;
    if(linkh.horizon != Local_horizon) {
	bp = ambufw(5);
	bp->cnt = bp->size;
	*bp->data = 32;	    /* the number of significant bits in the address */
	put32(bp->data+1,adj->addr);
	linkh.erp = adj->riface->erp;
        linkh.cost = adj->cost;
        linkh.adjn = 1;
        tbp = htonrspflink(&linkh,bp);
        nodeh.seq = Rspfseq;
        nodeh.subseq = ++Rspfsubseq;
        nodeh.links = 1;
        for(riface = Rspfifaces; riface != NULLRIFACE; riface = riface->next) {
	    if(dup_p(&wbp,tbp,0,len_p(tbp)) != len_p(tbp)) {
		free_p(wbp);
		continue;
	    }
	    nodeh.addr = riface->iface->addr;
	    bp = htonrspfnode(&nodeh,wbp);
	    sendtoall(bp,1,riface);
	}
        free_p(tbp);
    }
    if(adj->cost == 255 || adj->okcnt != 1)
	return;		/* bad adjacency or old adjacency becoming good */
    /* When two RSPF stations learn about each others existence, one of
     * them will usually have received a RRH from the other. The one that
     * received the RRH will send the peer a routing update immediately.
     * So in the code below, if no RRH has been received (seq is 0) and no
     * routing update has yet been received, we should expect to receive a
     * routing update shortly if the adjacency is running RSPF.
     * This could fail though if two RSPF stations learn about each other
     * solely through ARP upcalls and no RRH's or Updates were exchanged
     * prior to or during the establishment of the adjacency.
     */
    if(!adj->seq && rr_lookup(adj->addr) == NULLRROUTER) {
	/* No RRH or routing update received yet so
	 * we don't know if this is a router or not.
	 */
	if(adj->state != RSPF_SUSPECT)
	    return;		/* running in RSPF process, give up */
	/* If we get here we are running in a PING-process
	 * so we can relax and wait 60 seconds for an update.
	 */
	alarm(60 * 1000L);
	if(pwait(adj) != 0)
	    return;		/* still no sign of RSPF activity */
	alarm(0L);
    }
    /* This is a router, send it our latest update. */
    ++adj->okcnt;		/* we don't want to get here again */
    if((bp = makeownupdate(adj->riface->iface->addr)) == NULLBUF)
	return;
    for(rr = Rspfrouters; rr != NULLRROUTER; rr = rr->next)
	if((tbp = makeadjupdate(get32(rr->data->data),rr)) != NULLBUF) {
	     append(&bp,tbp);
	     nodecnt++;
	}
    sendupdate(bp,nodecnt,adj->addr);
}

/* This function is called whenever it is time to send a new RSPF update */
static void
sendrspf()
{
    struct rspfnodeh nodeh;
    struct rspfrouter *rr;
    struct mbuf *bp, *tbp, *wbp, *tmp = NULLBUF;
    struct rspfiface *riface;
    int nodecnt = 1;

    for(rr = Rspfrouters; rr != NULLRROUTER; rr = rr->next)
	if(!rr->sent)		/* don't send stale data */
	    if((bp = makeadjupdate(get32(rr->data->data),rr)) != NULLBUF) {
		append(&tmp,bp);
		nodecnt++;
	    }
    /* Increase sequence number, clear subseq. number. */
    Rspfseq = (Rspfseq == 32768L-2 || Rspfseq == -32768L+1) ? 0 : Rspfseq + 1;
    Rspfsubseq = 0;
    if((bp = makeownupdate(Ip_addr)) == NULLBUF)
	return;
    ntohrspfnode(&nodeh,&bp);
    for(riface = Rspfifaces; riface != NULLRIFACE; riface = riface->next) {
	if(adj_lookup(0L,riface->iface) != NULLADJ) {
	    if(dup_p(&wbp,tmp,0,len_p(tmp)) != len_p(tmp))
		free_p(wbp);
	    else if(dup_p(&tbp,bp,0,len_p(bp)) != len_p(bp)) {
		free_p(tbp);
		free_p(wbp);
	    } else {
		nodeh.addr = (riface->iface->forw != NULLIF) ?
			 riface->iface->forw->addr : riface->iface->addr;
		tbp = htonrspfnode(&nodeh,tbp);	/* Prepend the node header */
		append(&tbp,wbp);
		sendtoall(tbp,nodecnt,riface);
	    }
	}
    }
    free_p(bp);
    free_p(tmp);
    for(rr = Rspfrouters; rr != NULLRROUTER; rr = rr->next)
	rr->sent = 1;		/* Mark router data as propagated */
}

/* Send update news to all stations on all interfaces EXCEPT iface -- sm6rpz */
static void
sendpartial(bad,p,v)
int bad;
void *p, *v;
{
    struct mbuf *wbp;
    register struct rspfiface *riface;
    struct mbuf *bp = (struct mbuf *)p;
    register struct iface *iface = (struct iface *)v;

    if(bad) {			/* Running in a new process! */
	/* Perhapes we should here also check to see if we have gotten
	 * good news about this station and then inhibit this process.
	 */
	pause(60 * 1000L);	/* Wait 1 minute if bad news */
    }

    for(riface = Rspfifaces; riface != NULLRIFACE; riface = riface->next) {
	if(riface->iface != iface && adj_lookup(0L,riface->iface) != NULLADJ) {
	    if(dup_p(&wbp,bp,0,len_p(bp)) != len_p(bp))
		free_p(wbp);
	    else		/* Send the news to all stations on riface */
		sendtoall(wbp,1,riface); 
	}
    }
    free_p(bp);
}

/* send an update to all adjacencies on an RSPF interface */
static void
sendtoall(bp,nodecnt,riface)
struct mbuf *bp;
int nodecnt;			/* number of reporting nodes in update */
register struct rspfiface *riface;/* interface to send to */
{
    register struct route *rp;	/* sm6rpz */
    register struct arp_tab *ap;/* sm6rpz */
    register struct rspfadj *adj;
    struct mbuf *wbp;
    int broadcast = 0;		/* Send to broadcast address or not. */

    for(adj = Adjs; adj != NULLADJ; adj = adj->next)
	/* If adj->seq is 0 we have never received an RRH from the
	 * adjacency, and if there is no stored routing update, then
	 * it is not known if the adjacency understands RSPF.
	 */
	if(riface == adj->riface
	   && (adj->seq || rr_lookup(adj->addr) != NULLRROUTER)
	  ) {
	    /* Send private copies to routers with manual ARP:s also,
 	     * these might not be reached by broadcasts only -- sm6rpz
 	     */
	    if(adj->cost != 255
	       && adj->okcnt
	       && (((ap = arp_lookup(NHWTYPES,adj->addr,adj->riface->iface)) != NULLARP
		    && ap->state == ARP_VALID
		    && !ap->pub
		    && !dur_timer(&ap->timer)
		    && (rp = rt_lookup(ap->ip_addr)) != NULLROUTE
		    && ((rp->flags & RTPRIVATE) && !dur_timer(&rp->timer)
		        || (rp->gateway == 0L || rp->gateway == rp->target)
		        && rp->iface == ap->iface
		       )
	           )
		   || (adj->tos & RELIABILITY) && !(adj->tos & DELAY)
		  )
	      ) {
		    if(dup_p(&wbp,bp,0,len_p(bp)) != len_p(bp)) {
			free_p(wbp);
			continue;
		    }
		    sendupdate(wbp,nodecnt,adj->addr); /* private copy */
	    } else
		++broadcast;	/* send to broadcast IP address instead */
	}
    if(broadcast)
	if(dup_p(&wbp,bp,0,len_p(bp)) != len_p(bp))
	    free_p(wbp);
	else
	    sendupdate(wbp,nodecnt,riface->iface->broadcast);
    free_p(bp);
}     

/* This function sends a routing update to the
 * adjacencies that we have recevied RRH's from.
 */
static int
sendupdate(bp,nodecnt,addr)
struct mbuf *bp;	/* Full packet, except for the packet header */
int nodecnt;		/* Number of reporting nodes in the packet */
int32 addr;		/* Station to send update to */
{
    struct rspfadj *adj;
    struct mbuf *tmp;
    int tos = 0;

    /* See if we are sending to a known adjacency, use its TOS in that case. */
    if((adj = adj_lookup(addr,NULLIF)) != NULLADJ)
	tos = adj->tos;
    if((tmp = fragmenter(bp,nodecnt,ip_mtu(addr) - IPLEN)) == NULLBUF)
	return -1;
    while((bp = dequeue(&tmp)) != NULLBUF) {
	rspfcsum(&bp,addr);	/* Calculate the checksum */
	++Rspf_stat.updateout;
	ip_send(INADDR_ANY,addr,RSPF_PTCL,INET_CTL | tos,2,bp,0,0,0);
    }
    return 0;
}

/* Fragment a large mbuf if necessary into packets of maximum mtu bytes.
 * Each packet is prepended a RSPF routing update header, without the checksum.
 */
static struct mbuf *
fragmenter(bp,nodes,mtu)
struct mbuf *bp;/* Packet to send, without packet header. */
int nodes;	/* The number of reporting nodes in the routing update. */
int16 mtu;	/* The maximum transmission unit. */
{
    struct rspfnodeh nodeh;
    struct rspflinkh linkh;
    struct rspfpacketh pkth;
    struct mbuf *tbp, *tmp, *bpq = NULLBUF;
    int n, nodecnt, linkcnt, adjcnt, nodemax, sync;
    char *cp, fragn = 1;

    if(mtu < RSPFPKTLEN + RSPFNODELEN || bp == NULLBUF
       || len_p(bp) < RSPFNODELEN) {
	free_p(bp);
	return NULLBUF;
    }
    ++Envelopeid;
    nodemax = nodes;		/* total number of nodes in envelope */
    nodecnt = nodes;		/* nodes left to packetize */
    linkcnt = adjcnt = 0;
    while(len_p(bp)) {
	n = min(mtu,len_p(bp) + RSPFPKTLEN);	/* length of this packet */
	n -= RSPFPKTLEN;
	tbp = NULLBUF;
	sync = 0;
	if(adjcnt) {
	    tbp = ambufw(min(adjcnt * 5, n / 5 * 5));
	    tbp->cnt = tbp->size;
	    cp = tbp->data;
	}
	while(n > 0) {
	    while(adjcnt) {
		if((n -= 5) < 0)
		    break;
		pullup(&bp,cp,5);
		cp += 5;
		adjcnt--;
	    }
	    if(linkcnt && n > 0) {
		if((n -= RSPFLINKLEN) < 0)
		    break;
		ntohrspflink(&linkh,&bp);
		adjcnt = linkh.adjn;
		tmp = htonrspflink(&linkh,NULLBUF);
		append(&tbp,tmp);
		tmp = ambufw(min(5 * adjcnt, n / 5 * 5));
		tmp->cnt = tmp->size;
		cp = tmp->data;
		append(&tbp,tmp);
		linkcnt--;
		continue;
	    }
	    if(nodecnt && !linkcnt && n > 0) {
		if((n -= RSPFNODELEN) < 0)
		    break;
		if(nodecnt == nodes)		/* Set the sync octet */
		    sync = len_p(tbp) + 4;
		ntohrspfnode(&nodeh,&bp);
		linkcnt = nodeh.links;
		tmp = htonrspfnode(&nodeh,NULLBUF);
		append(&tbp,tmp);
		nodecnt--;
	    }
	}
	pkth.version = RSPF_VERSION;	/* The version number */
	pkth.type = RSPF_FULLPKT;	/* The packet type */
	pkth.fragn = fragn++;		/* The fragment number */
	pkth.fragtot = 0;		/* The total number of fragments */
	pkth.csum = 0;			/* The checksum */
	pkth.sync = sync < 256 ? sync : 0;/* The sync octet */
	pkth.nodes = nodemax;		/* The number of nodes in envelope */
	pkth.envid = Envelopeid;	/* The envelope-ID */
	tmp = htonrspf(&pkth,tbp);	/* add packet header */
	enqueue(&bpq,tmp);		/* add packet to chain */
	nodes = nodecnt;		/* number of nodes left */
    }
    free_p(bp);
    for(tbp = bpq; tbp != NULLBUF; tbp = tbp->anext)
	*(tbp->data + 3) = len_q(bpq);	/* Set the fragment total counter */
    return bpq;
}

/* Calculate the checksum on an RSPF routing update packet */
static void
rspfcsum(bpp,dest)
register struct mbuf **bpp;
register int32 dest;
{
    struct pseudo_header ph;
    register int16 checksum;

    ph.length = len_p(*bpp);
    ph.source = locaddr(dest);
    ph.dest = dest;
    ph.protocol = RSPF_PTCL;
    if((checksum = cksum(&ph,*bpp,ph.length)) == 0)
	checksum = 0xffffffffL;
    /* This assumes that the checksum really is at this position */
    put16((*bpp)->data+4,checksum);
}

/* This function creates our own routing update and returns it in an mbuf */
struct mbuf *
makeownupdate(addr)
int32 addr;			/* Address on the destinations interface. */
{
    register struct rspfadj *adj;
    struct rspflinkh linkh;
    struct rspfnodeh nodeh;
    register struct rspfiface *riface;
    struct rspfiface rifdefault;
    struct mbuf *bp = NULLBUF, *tbp, *tmp;
    register struct route *rp;
    register int i, adjcnt, lnkcnt, bits;
    register int32 prev, high, cur;

    /* Set default values to apply to non-RSPF interfaces */
    rifdefault.erp = 0;
    rifdefault.quality = 0;

    /* Our adjacencies has to be sorted into groups of the same horizon,
     * ERP factor and cost. This is done by combining these numbers into a
     * single one, modulo 256 and take the highest number first.
     */
    high = 255*65536L+255*256+255;
    lnkcnt = 0;
    for(;;) {
	prev = high;
	high = 0;
	for(adj = Adjs; adj != NULLADJ; adj = adj->next)
	    /* Include all adjacencies that have been or are in OK state
	     * in the update. Bad adjacencies are also included to give
	     * them a chance to recover. Hopelessly Bad adjacencies are
	     * eventually deleted elsewhere.
	     */
	    if(adj->okcnt) {
		cur = ((adj->seq || rr_lookup(adj->addr) != NULLRROUTER) ?
		       Router_horizon : (rt_clookup(adj->addr) == NULLROUTE) ?
		       Portable_horizon : Local_horizon
		      )*65536L+adj->riface->erp*256+adj->cost;
		if(cur < prev && cur > high)
		    high = cur;
	    }
	/* Add any manual public, 1-31 bit specific routes. Use the
	 * route metric only if it is greater than the default
	 * quality to lessen a possible "wormhole" effect.
	 */
	for(bits = 1; bits <= 31; bits++)
	    for(i = 0; i < HASHMOD; i++)
		for(rp = Routes[bits-1][i]; rp != NULLROUTE; rp = rp->next)
		    if(!(rp->flags & RTPRIVATE) && !dur_timer(&rp->timer)) {
			if((riface = riface_lookup(rp->iface)) == NULLRIFACE)
			    riface = &rifdefault;
			cur = Group_horizon*65536L+riface->erp*256+
			    (rp->metric > riface->quality ?
			     rp->metric : riface->quality);
			if(cur < prev && cur > high)
			    high = cur;
		     }
	/* Add any 32 bit routes on interfaces not using RSPF */
	for(i = 0; i < HASHMOD; i++)
	    for(rp = Routes[31][i]; rp != NULLROUTE; rp = rp->next)
		if(!(rp->flags & RTPRIVATE) && riface_lookup(rp->iface)
		     == NULLRIFACE) {
		    cur = Non_rspf_horizon*65536L+rifdefault.erp*256+
			    (rp->metric > rifdefault.quality ?
			     rp->metric : rifdefault.quality);
		    if(cur < prev && cur > high)
			high = cur;
		}
	if(high == 0) /* All done */
	    break;
	/* now start adding the routes */
	adjcnt = 0;
	tbp = NULLBUF;
	for(adj = Adjs; adj != NULLADJ; adj = adj->next)
	    if(adj->okcnt
	       && ((adj->seq || rr_lookup(adj->addr) != NULLRROUTER) ?
		   Router_horizon : (rt_clookup(adj->addr) == NULLROUTE) ?
		   Portable_horizon : Local_horizon)*65536L+
		   adj->riface->erp*256+adj->cost == high
	      ) {
		pushadj(&tbp,adj->addr,32);
		adjcnt++;
	    }
	/* Manual, public, 1-31 bit routes */
	for(bits = 1; bits <= 31; bits++)
	    for(i = 0; i < HASHMOD; i++)
		for(rp = Routes[bits-1][i]; rp != NULLROUTE; rp = rp->next)
		    if(!(rp->flags & RTPRIVATE) && !dur_timer(&rp->timer)) {
			if((riface = riface_lookup(rp->iface)) == NULLRIFACE)
			       riface = &rifdefault;
			/* Manually entered local routes only */
			if(Group_horizon*65536L+riface->erp*256+
			   (rp->metric > riface->quality ? rp->metric :
			   riface->quality) == high) {
				pushadj(&tbp,rp->target,bits);
				adjcnt++;
			}
		    }
	/* 32 bit specific routes not on interfaces using RSPF */
	for(i = 0; i < HASHMOD; i++)
	     for(rp = Routes[31][i]; rp != NULLROUTE; rp = rp->next)
		 if(!(rp->flags & RTPRIVATE) && riface_lookup(rp->iface)
		     == NULLRIFACE)
		    if(Non_rspf_horizon*65536L+rifdefault.erp*256+
			  (rp->metric > rifdefault.quality ? rp->metric :
			   rifdefault.quality) == high) {
			pushadj(&tbp,rp->target,bits);
			adjcnt++;
		    }
	if(adjcnt) {
	    /* Prepend the link header */
	    linkh.horizon = ((high >> 16) & 255);/* Horizon value */
	    if(linkh.horizon == 0) {
		free_p(tbp);
		continue;	/* Don't bother about these ones. -- sm6rpz */
	    }
	    linkh.erp = ((high >> 8) & 255);	/* ERP factor */
	    linkh.cost = (high & 255);		/* Link cost */
	    linkh.adjn = adjcnt;		/* Number of adjacencies */
	    tmp = htonrspflink(&linkh,tbp);
	    append(&bp,tmp);
	    lnkcnt++;
	}
    }
    if(!lnkcnt) {				/* No links at all? */
	free_p(bp);
	return NULLBUF;
    }
    nodeh.addr = addr;
    nodeh.seq = Rspfseq;
    nodeh.subseq = 0;	/* Subsequence always 0 on full routing updates. */
    nodeh.links = lnkcnt;
    return htonrspfnode(&nodeh,bp);		/* Prepend the node header */
}

/* Prepends an adjacency to bpp. Allocates bpp in large chunk for efficency */
static void
pushadj(bpp,addr,bits)
register struct mbuf **bpp;
int32 addr;
int bits;
{
    if(bpp == NULLBUFP)
	return;
    if(*bpp == NULLBUF) {
	*bpp = ambufw(60);		/* good for 12 adjacencies */
	(*bpp)->data += 55;
	(*bpp)->cnt = 5;
    } else
	*bpp = pushdown(*bpp,5);
    *(*bpp)->data = bits;
    put32((*bpp)->data+1,addr);
}

/* This function returns the latest routing update in network format from
 * the adjacency denoted by addr.
 */
static struct mbuf *
makeadjupdate(addr,rr)
int32 addr;
struct rspfrouter *rr;	/* pointer to stored routing entry, if known */
{
    struct mbuf *tmp, *tmp2, *tmp3, *bp = NULLBUF;
    struct rspfnodeh nodeh;
    struct rspflinkh linkh;
    int linkcnt = 0;

    if(rr == NULLRROUTER && (rr = rr_lookup(addr)) == NULLRROUTER)
	return NULLBUF;
    if(dup_p(&tmp,rr->data,0,len_p(rr->data)) != len_p(rr->data)) {
	free_p(tmp);
	return NULLBUF;
    }

    ntohrspfnode(&nodeh,&tmp);		/* Get the node header */

    if(nodeh.subseq) {			/* This is a partial update and we */
	free_p(tmp);			/* should here only send full ones. */
	return NULLBUF;
    }

    while(nodeh.links--) {
	ntohrspflink(&linkh,&tmp);
	/* Decrement and check the horizon value */
	if(--linkh.horizon > 0) {
	    /* Duplicate the adjacencies */
	    if(dup_p(&tmp2,tmp,0,5*linkh.adjn) != 5*linkh.adjn) {
		free_p(tmp);
		free_p(tmp2);
		free_p(bp);
		return NULLBUF;
	    }
	    /* Prepend the link header */
	    tmp3 = htonrspflink(&linkh,tmp2);
	    append(&bp,tmp3);
	    linkcnt++;
	}
	pullup(&tmp,NULLCHAR,5*linkh.adjn); /* Skip the adjacencies */
    }
    free_p(tmp);
    if(!linkcnt) {			/* No links at all? */
	free_p(bp);
	return NULLBUF;
    }
    nodeh.links = linkcnt;
    return htonrspfnode(&nodeh,bp);	/* Prepend the node header */
}

/* Returns 1 if sequence number a is newer than b. Sequence numbers start
 * at -32768 + 1 and then continues with 0 and the positive integer numbers
 * until reaching 32768 - 2 after which they continue with 0.
 * Algorithm taken from RFC-1131 but fixed bug when a or b is 0.
 */
static int
isnewer(a,b)
register short a,b;
{
    if((b < 0 && a > b) ||
       (a >= 0 && b >= 0 &&	/* bug corrected here */
       (((32768L - 1)/2 > (a-b) && (a-b) > 0) || (a-b) < -(32768L - 1)/2)))
	return 1;
    return 0;
}
     
/* Reassemble possibly fragmented packets into a single large one */
static struct mbuf *
rspfreasm(bp,rph,addr)
struct mbuf *bp;	    /* Routing update packet without packet header */
register struct rspfpacketh *rph;    /* Packet header */
register int32 addr;	    /* Address of station we got the packet from */
{
    union rspf rspf;
    register struct rspfreasm *re, *rtmp;
    struct mbuf *tbp, *bp1, *bp2;

    for(re = Rspfreasmq; re != NULLRREASM; re = re->next) {
	if(re->addr == addr) {		/* found another fragment */
	    if(dup_p(&tbp,re->data,0,RSPFPKTLEN) != RSPFPKTLEN) {
		free_p(tbp);
		free_p(bp);
		return NULLBUF;
	    }
	    ntohrspf(&rspf,&tbp);	/* get its packet header */
	    if(rph->envid != rspf.pkthdr.envid) {
		del_reasm(re);		/* an obsolete entry, delete it */
		break;
	    }
	    /* Now try to find a place where this fragment fits in the chain */
	    bp1 = re->data;
	    bp2 = NULLBUF;
	    while(rph->fragn > rspf.pkthdr.fragn && bp1->anext != NULLBUF) {
		bp2 = bp1;
		bp1 = bp1->anext;
		if(dup_p(&tbp,bp1,0,RSPFPKTLEN) != RSPFPKTLEN) {
		    free_p(bp);
		    free_p(tbp);
		    return NULLBUF;
		}
		ntohrspf(&rspf,&tbp);
	    }
	    if(rspf.pkthdr.fragn == rph->fragn) {
		free_p(bp);
		return NULLBUF;		/* We already had this one */
	    }
	    bp = htonrspf(rph,bp);	/* Put the packet header back on */
	    /* Insert the fragment at the right place in the chain */
	    if(rph->fragn > rspf.pkthdr.fragn) {
		bp1->anext = bp;
		bp->anext = NULLBUF;
	    } else {
		if(bp2 != NULLBUF)
		    bp2->anext = bp;
		else
		    re->data = bp;
		bp->anext = bp1;
	    }
	    if(len_q(re->data) == rspf.pkthdr.fragtot) { 
		bp1 = NULLBUF;			/* we have all the fragments */
		while((bp2 = dequeue(&re->data)) != NULLBUF) {
		    pullup(&bp2,NULLCHAR,RSPFPKTLEN); /* strip the headers */
		    append(&bp1,bp2);		/* and form a single packet */
		}
		del_reasm(re);
		stop_timer(&Rspfreasmt);
		reasmtimeout(NULL);	/* restarts timer if necessary */
		return bp1;		/* return the completed packet */
	    }
	    re->time = secclock();	/* Update the timestamp */
	    goto timstart;		/* We have to wait for fragments */
	}
    }
    /* At this point we know that there is no matching entry in the
     * reassembly queue (any more)
     */
    if(rph->fragtot == 1)	/* Simple case, an un-fragmented packet */
	return bp;
    tbp = htonrspf(rph,bp);	/* Put the packet header back on */
    rtmp = (struct rspfreasm *)callocw(1,sizeof(struct rspfreasm));
    /* The values are filled in */
    rtmp->addr = addr;
    rtmp->time = secclock();
    rtmp->data = tbp;
    rtmp->next = Rspfreasmq;
    Rspfreasmq = rtmp;
timstart:					/* Start the timer if needed */
    if(!run_timer(&Rspfreasmt)) {
	Rspfreasmt.func = reasmtimeout;
	Rspfreasmt.arg = NULL;
	set_timer(&Rspfreasmt, ipReasmTimeout * 1000L);
	start_timer(&Rspfreasmt);
    }
    return NULLBUF;
}

/* Delete a reassembly descriptor from the queue */
static void
del_reasm(re)
register struct rspfreasm *re;
{
    register struct rspfreasm *r, *prev = NULLRREASM;

    for(r = Rspfreasmq; r != NULLRREASM; prev = r, r = r->next) {
	if(r == re) {
	    free_q(&re->data);
	    if(prev != NULLRREASM)
		prev->next = re->next;
	    else
		Rspfreasmq = re->next;
	    free((char *)re);
	    break;
	}
    }
}

/* When the reassembly timer times out, this function tries to make use of
 * any fragments in the reassembly queue.
 */
static void
reasmtimeout(t)
void *t;
{
    union rspf rspf;
    register struct rspfreasm *re;
    struct mbuf *bp, *tbp;
    int last = 0;
    int32 low;
    extern int32 clock __ARGS((void));

    re = Rspfreasmq;
    while(re != NULLRREASM) {
	if((secclock() - re->time) >= ipReasmTimeout) {
	    /* Try to use what we have */
	    bp = NULLBUF;
	    while((tbp = dequeue(&re->data)) != NULLBUF) {
		ntohrspf(&rspf,&tbp);
		if(rspf.pkthdr.fragn != (last+1)) { /* a missing fragment */
		    if(bp != NULLBUF)
			update_in(bp,re->addr,NULLRIFACE);
		    bp = NULLBUF;
		    if(rspf.pkthdr.sync)
			pullup(&tbp,NULLCHAR,rspf.pkthdr.sync - 4);
		    else {	/* no sync possible in this fragment */
			free_p(tbp);
			continue;
		    }
		}
		append(&bp,tbp);
		last = rspf.pkthdr.fragn;
	    }
	    if(bp != NULLBUF)
		update_in(bp,re->addr,NULLRIFACE);
	    del_reasm(re);
	    re = Rspfreasmq;	/* start over again */
        } else
	     re = re->next;
    }
    /* Find the oldest fragment and restart the reassembly timer */
    low = 0;
    for(re = Rspfreasmq; re != NULLRREASM; re = re->next)
	if(re->time < low || low == 0)
	    low = re->time;
    if(low) {
	set_timer(&Rspfreasmt, ipReasmTimeout * 1000L - clock() + low);
	if(dur_timer(&Rspfreasmt) > 0L)	/* just to be safe */
	    start_timer(&Rspfreasmt);
    }
}

/* Handle incoming routing updates (a reassembled envelope) */
static void
update_in(bp,addr,riface)
struct mbuf *bp;	/* Reassembled data packet, without packet header */
register int32 addr;	/* Senders address (in host order) */
register struct rspfiface *riface;  /* Used if a unknown router. */
{
    struct rspfnodeh nodeh;
    struct rspflinkh linkh;
    struct mbuf *tbp, *tmp, *b, *update = NULLBUF;
    register int linkcnt = 0, adjcnt = 0;
    int nodecnt = 0;
    register int16 offset = 0;
    char make_routes = 0;   /* if > 0 makeroutes() should be run -- sm6rpz */
    register struct rspfadj *adj;

    tbp = b = NULLBUF;
    /* Check to see if the sender is an adjacency.
     * One may argue that updates from non-adjacencies should not be
     * accepted since they will not contribute to the routing table.
     * But it might happen that the sender will very shortly become
     * an adjacency, and then its routing update will come handy. By
     * increasing Keeprouter, routing updates from disjoint routers
     * will not be not be purged when makeroutes() is called this time.
     */
    if((adj = adj_lookup(addr,NULLIF)) == NULLADJ) {
	if(riface != NULLRIFACE)	/* Could be a new router so check it */
	    rspfupcall(addr,NHWTYPES,riface->iface,0);
	++Rspf_stat.noadjupdate;
	Keeprouter += 2;
    } else {
	psignal(adj,1);		/* Alert anyone waiting for the update. */
#ifdef	RSPF_HEARD
	/* Adjacency is still alive so restart the suspect timer,
 	 * see add_rspfadj() for more information. Do this only
 	 * if the timer is already running -- sm6rpz
 	 */
	if(adj->state == RSPF_OK && read_timer(&adj->timer))
	    start_timer(&adj->timer);
#endif
    }
    while(offset < len_p(bp)) {
	if(adjcnt) {
	    if(dup_p(&tmp,bp,offset,5) == 5) {
		append(&tbp,tmp);
		offset += 5;
		adjcnt--;
		continue;
	    } else {
		free_p(tmp);		/* Bug fix -- sm6rpz */
		break;
	    }
	}
	if(tbp != NULLBUF) {
	    tmp = htonrspflink(&linkh,tbp);
	    append(&b,tmp);
	    tbp = NULLBUF;
	}
	if(linkcnt) {
	    if(dup_p(&tbp,bp,offset,RSPFLINKLEN) == RSPFLINKLEN) {
		ntohrspflink(&linkh,&tbp);
		adjcnt = linkh.adjn;
		offset += RSPFLINKLEN;
		linkcnt--;
		continue;
	    } else
		break;
	}
	if(b != NULLBUF) {
	    tmp = htonrspfnode(&nodeh,b);
	    /* One complete entry ready for processing. */
	    make_routes += node_in(tmp,addr,0,&update,&nodecnt);
	    b = NULLBUF;
	}
	if(dup_p(&tmp,bp,offset,RSPFNODELEN) == RSPFNODELEN) {
	    ntohrspfnode(&nodeh,&tmp);
	    linkcnt = nodeh.links;
	    offset += RSPFNODELEN;
	} else {
	    free_p(bp);
	    free_p(tmp);
	    return;
	}
    }
    if(tbp != NULLBUF) {
	/* Adjust the adjacency counter in the link header. */
	linkh.adjn -= adjcnt;
	tmp = htonrspflink(&linkh,tbp);
	append(&b,tmp);
    }
    free_p(bp);
    /* Adjust the link counter in the node header. */
    nodeh.links -= linkcnt;
    tmp = htonrspfnode(&nodeh,b);
    /* If linkcnt != 0 or adjcnt != 0 we have a fragmented entry. */
    if(node_in(tmp,addr,(linkcnt || adjcnt),&update,&nodecnt) || make_routes)
	makeroutes();
    if(nodecnt)		/* Should we send updates to the received update? */
	sendupdate(update,nodecnt,addr);
}

/* Returns 1 if makeroutes() should be run, 0 otherwise. -- sm6rpz */
static int
node_in(bp,addr,fragment,update,nodecnt)
struct mbuf *bp;	/* A single update, starting with the node header */
register int32 addr;	/* Address of station we got packet from */
int fragment;		/* True if we got a fragment of a router update */
struct mbuf **update;	/* Holds eventual update to be sent. */
int *nodecnt;		/* Numbers of nodes in update to be sent. */
{
    struct mbuf *tbp;
    struct rspfnodeh nodeh, rnodeh;
    register struct rspfrouter *rr;
    register struct route *rp;		/* sm6rpz */
    struct rspflinkh linkh;		/* sm6rpz */
    register struct rspf_refuse *rl;	/* sm6rpz */

    ntohrspfnode(&nodeh,&bp);

    /* Refuse if router is on the refuse list -- sm6rpz */
    for(rl = Rspf_refuse; rl != NULLREFUSE; rl = rl->next)
	if(rl->target == nodeh.addr) {
	    free_p(bp);
	    ++Rspf_stat.refused;
	    if(Rspf_trace > 0)
		printf("RSPF update refused from %s\n",inet_ntoa(nodeh.addr));
	    return 0;
	}

    if(ismyaddr(nodeh.addr)) {
	free_p(bp); 		/* We already know our own adjacencies! */
	if(nodeh.seq == Rspfseq)
	    return 0;			/* Our latest update. */
	if(isnewer(nodeh.seq,Rspfseq)) {
	    /* Old "epoch"-data present in the network, catch up. */
	    Rspfseq = (nodeh.seq == 32768L - 2 || nodeh.seq == -32768L + 1) ?
		      0 : nodeh.seq + 1;
	    Rspfsubseq = 0;		/* BUG FIX -- sm6rpz */
	}
	/* The other station has a old routing update, send him the latest. */
	++Rspf_stat.oldreport;
	if((addr = locaddr(addr)) == 0L)
	    addr = Ip_addr;
	if((tbp = makeownupdate(addr)) != NULLBUF) {
	    append(update,tbp);
	    ++*nodecnt;
	}
	return 0;
    }
    if((rr = rr_lookup(nodeh.addr)) != NULLRROUTER) {
	if(dup_p(&tbp,rr->data,0,RSPFNODELEN) != RSPFNODELEN) {
	    free_p(bp);
	    free_p(tbp);
	    return 0;
        }
	ntohrspfnode(&rnodeh,&tbp);
	if(isnewer(rnodeh.seq,nodeh.seq)) {
	    /* We got a old routing update (poll ?), send him the latest. */
	    free_p(bp);
	    ++Rspf_stat.oldreport;
	    if((tbp = makeadjupdate(nodeh.addr,NULLRROUTER)) != NULLBUF) {
		append(update,tbp);
		++*nodecnt;
	    }
	    return 0;
	}
	if(nodeh.seq == rnodeh.seq && nodeh.subseq <= rr->subseq) {
	    if(rnodeh.subseq && rnodeh.subseq > nodeh.subseq) {
		/* The first update we got from this station was a partial
		 * update. We have now received the full routing update,
		 * or a partial routing update with a lower subsequence, so
		 * swap the data and run partialupdate(). This will make it
		 * possible to build up a full routing update even if we
		 * have missed it. -- sm6rpz
		 */
		tbp = rr->data;
		pullup(&tbp,NULLCHAR,RSPFNODELEN);
		rr->data = htonrspfnode(&nodeh,bp);
		partialupdate(rr,tbp);
		return 1;		/* Makeroutes() should be run. */
	    }
	    if(rr->subseq == 0 && nodeh.subseq == 0) {
		/* We must here make sure that the update we have is the one
		 * that has traversed the shortest path. Se section IV.3 in the
		 * protocol. This is taken care of in partialupdate().
		 */
		partialupdate(rr,bp);
		/* No need to run makeroutes. */
	    } else
		free_p(bp);		/* We already have this one, ignore. */
	    return 0;
	}
	if(fragment) {
	    /* We have received a fragment of a larger routing update, use
	     * what we have got and poll the reporting router for the rest.
	     */
	    partialupdate(rr,bp);
	    /* Send a "poll" packet. */
	    --nodeh.seq;
	    nodeh.subseq = 0;
	    nodeh.links = 0;
	    tbp = htonrspfnode(&nodeh,NULLBUF);
	    append(update,tbp);
	    ++*nodecnt;
	    ++Rspf_stat.outpolls;
	    return 1;			/* Makeroutes() should be run. */
	}
	if(nodeh.subseq) {		/* Partial update, good or bad news. */
	    rr->subseq = nodeh.subseq;
	    if(dup_p(&tbp,bp,0,len_p(bp)) != len_p(bp))
		free_p(tbp);
	    else
		partialupdate(rr,tbp);
	}
    } else {				/* New router */
	/* Allocate a new router structure. */
	rr = (struct rspfrouter *)callocw(1,sizeof(struct rspfrouter));
	rr->next = Rspfrouters;
	Rspfrouters = rr;
    }
    if(nodeh.subseq) {			/* Partial update, good or bad news. */
	/* We should here also inform our adjacencies about the news (bad
	 * news should travel slower). Look at from which interface we have
	 * received the update. Transmit the update on all other interfaces
	 * (we don't want to send the data on the arriving interface because
	 * the likelihood that all stations there already have received the
	 * news are big).
	 * When we receive bad news we must save these for 60 seconds. We
	 * therefore must start a new process allowing us to continue
	 * processing RSPF packets while waiting. -- sm6rpz
	 */
	if((rp = rt_lookup(addr)) != NULLROUTE) {
	    if(dup_p(&tbp,bp,0,len_p(bp)) != len_p(bp))
		free_p(tbp);
	    else {
		ntohrspflink(&linkh,&tbp);
		if(--linkh.horizon > 0) {
		    /* Put link and node header on again */
		    tbp = htonrspflink(&linkh,tbp);
		    tbp = htonrspfnode(&nodeh,tbp);
		    if(linkh.cost < 255)	/* good news :-) */
			sendpartial(0,tbp,rp->iface);
		    else			/* bad news :-( */
			newproc("RSPF tx_partial",256,sendpartial,1,
				tbp,rp->iface,0);
		} else
		    free_p(tbp);
	    }
	}
	if(rr->time) {			/* Non-zero only for known routers. */
	    free_p(bp);			/* Clean up. */
	    return 1;			/* Makeroutes() should be run. */
	}
    }
    free_p(rr->data);
    rr->sent = 0;
    rr->subseq = nodeh.subseq;		/* Save eventual subsequence. */
    rr->time = secclock();
    rr->data = htonrspfnode(&nodeh,bp);
    return 1;				/* Makeroutes() should be run. */
}

/* Return the matching RSPF route entry for addr (in host order) */
struct rspfrouter *
rr_lookup(addr)
register int32 addr;
{
    register struct rspfrouter *rr;

    for(rr = Rspfrouters; rr != NULLRROUTER; rr = rr->next)
	/* this assumes that the first word of the header is the address */
	if(rr->data != NULLBUF && get32(rr->data->data) == addr)
	    return rr;
    return NULLRROUTER;
}

/* Handle incoming partial routing updates.
 * Adjacencies from bp will be merged into rr->data
 */
static void
partialupdate(rr,bp)
register struct rspfrouter *rr; /* current node entry in routing table */
struct mbuf *bp;		/* data packet, starting with link header */
{
    struct rspflinkh linkh, rlinkh;
    struct rspfnodeh rnodeh;
    struct mbuf *wbp, *tbp, *tmp, *b;
    register int adjcnt = 0, radjcnt, rlinkcnt = 0, bits, rbits, added = 0;
    register int32 raddr;
    int32 addr;

    if(bp == NULLBUF)
	return;

    rlinkh.adjn = 0;
    rr->time = secclock();
    rr->sent = 0;

    /* Make a working copy of the stored routing update */
    if(dup_p(&wbp,rr->data,0,len_p(rr->data)) != len_p(rr->data)) {
	free_p(wbp);
	free_p(bp);
	return;
    }
    ntohrspfnode(&rnodeh,&wbp);
    /* Save old link count */
    rlinkcnt = rnodeh.links;		/* Bug fix -- sm6rpz 1992-10-13 */

    while(len_p(bp) > 0) {
	if(!adjcnt) {
	    if(ntohrspflink(&linkh,&bp) == -1) {
		free_p(wbp);
		free_p(bp);
		return;
	    }
	    adjcnt = linkh.adjn;
	} else {
	    bits = PULLCHAR(&bp);	/* get one adjacency to merge */
	    if(pullup(&bp,(char *)&addr,4) != 4) {
		free_p(wbp);
		free_p(bp);
		return;
	    }
	    addr = get32((char *)&addr); /* convert to host order */
	    --adjcnt;
	    radjcnt = 0;
	    rlinkcnt = 0;		/* Bug fix -- sm6rpz 1992-10-13 */
	    b = tbp = NULLBUF;
	    for(;;) {			/* Search through stored update. */
		if(!radjcnt || len_p(wbp) == 0) {
		    if(tbp != NULLBUF) {
			rlinkh.adjn -= radjcnt;
			tmp = htonrspflink(&rlinkh,tbp);
			++rlinkcnt;
			append(&b,tmp);
			tbp = NULLBUF;
		    }
		    if(len_p(wbp) == 0)
			break;
		}
	    	if(!radjcnt) {
		    ntohrspflink(&rlinkh,&wbp);
		    radjcnt = rlinkh.adjn;
		    if(rlinkh.horizon == linkh.horizon
		       && rlinkh.cost == linkh.cost
		       && rlinkh.erp == linkh.erp
		       && !added) {
			pushadj(&tbp,addr,bits);	/* Add new adjacency */
			++rlinkh.adjn;
			added = 1;
		    }
		} else {
		    rbits = PULLCHAR(&wbp);
		    raddr = pull32(&wbp);
		    --radjcnt;
		    if(bits != rbits || addr != raddr)
			pushadj(&tbp,raddr,rbits);	/* Put old adj back */
		    else if(rlinkh.horizon > linkh.horizon
			    && rlinkh.cost == linkh.cost
			    && rlinkh.erp == linkh.erp
			    && !added) {
			pushadj(&tbp,raddr,rbits);	/* Put old adj back */
			added = 1;		/* Mark the new one as added */
		    } else
			--rlinkh.adjn;		/* Deleted old adjacency */
	        }
	    }
	    wbp = b;	/* wbp now keeps link headers and adjacencies. */
	    if((linkh.cost < 255 || rnodeh.subseq) && !added) {
		/* Append the new link. Bad adjacencies are also added, if
		 * the old data is a partial update, so that we can remove
		 * adjacencies when we receive the full update. -- sm6rpz
		 */
		++rlinkcnt; /* We will add one extra link - BUG FIX -- sm6rpz */
		tmp = ambufw(5);
		*tmp->data = bits;
		put32(tmp->data + 1,addr);
		tmp->cnt = tmp->size;
		tbp = htonrspflink(&linkh,tmp);
		append(&wbp,tbp);
	    }
	    added = 0;
        }
    }
    free_p(rr->data);
    rnodeh.links = rlinkcnt;
    rr->data = htonrspfnode(&rnodeh,wbp);
}

/* The shortest path first algorithm */
static void
makeroutes()
{
    register struct route *rp, *rp2, *saved[HASHMOD];
    register struct rspfadj *adj, *lowadj, *gateway;
    register struct rspflinkh linkh;
    register struct rspfrouter *rr, *rrprev;
    register int i, adjcnt;
    register char *lowp, *r;
    struct mbuf *bp;
    int32 lowaddr;
    int bits, low;

    if(Keeprouter)	/* If false, purge unreachable router entries. */
	--Keeprouter;

    /* Remove all non-manual routes. */
    for(bits = 1; bits <= 32; bits++)
	for(i = 0; i < HASHMOD; i++)
/*	    for(rp = Routes[bits-1][i]; rp != NULLROUTE; rp = rp->next) */
	    for(rp = Routes[bits-1][i]; rp != NULLROUTE; rp = rp2) {
		rp2 = rp->next;		/* BIG BUG FIX -- sm6rpz */
		if(dur_timer(&rp->timer))
		    rt_drop(rp->target,bits);
		/* rp will be undefined here if(dur_timer(&rp->timer) */
	    }

    if((rp = rt_blookup(0L,0)) != NULLROUTE && dur_timer(&rp->timer))
	rt_drop(0L,0);			/* Delete non-manual default route. */

    /* Temporarily remove all 32-bit specific manual routes. This
     * will make it possible to prevent loops in findlowroute()
     */
    for(i = 0; i < HASHMOD; i++) {
	saved[i] = Routes[31][i];
	Routes[31][i] = NULLROUTE;
    }

    for(;;) {
	lowadj = NULLADJ;
	lowp = NULLCHAR;
	low = 255;
	for(adj = Adjs; adj != NULLADJ; adj = adj->next) {
	    if(!adj->added) {
		if(adj->cost <= low) {
		    low = adj->cost;
		    lowadj = adj;
		    lowp = NULLCHAR;
		}
	    } else
		if((r = findlowroute(adj->addr,&low,adj->cost,&lowaddr,&bits))
		  != NULLCHAR) {
		    lowp = r;
		    gateway = adj;
		    lowadj = NULLADJ;
	        }
	}
	if(lowadj != NULLADJ) {
	    lowadj->added++;
	    rp = rt_add(lowadj->addr,32,0L,lowadj->riface->iface,
			(int32)lowadj->cost,0L,0);
	    /* Set the timer to a dummy value. This makes it possible to
	     * distinguish between manual and RSPF-generated routes. The
	     * timer is never started however since RSPF handles the
	     * expiration of routes itself.
	     */
	    set_timer(&rp->timer,MSPTICK);
	    continue;
	}
	if(lowp != NULLCHAR) {
	    /* Check if we already have this one */
	    rp = rt_blookup(lowaddr,bits);
	    /* rp->metric > (int32)low changed to rp->metric >= (int32)low */
	    if((rp == NULLROUTE || rp->metric >= (int32)low) && !ismyaddr(lowaddr)) {
		/* This is a new route, or a route with a lower,
		 * or equal, cost than the previous route.
		 */
		rp = rt_add(lowaddr,(unsigned)bits,gateway->addr,
			    gateway->riface->iface,(int32)low,0L,0);
		set_timer(&rp->timer,MSPTICK); /* a dummy value */
	    }
	    *lowp |= 128;	/* mark the route as added in any case */
	} else
	    break;		/* no more routes */
    }

    /* Add the saved 32 bit routes, if there isn't now a better route */
    for(i = 0; i < HASHMOD; i++) {
	rp = saved[i];
	while(rp != NULLROUTE) {
	    rp2 = rt_blookup(rp->target,32);
	    /* rp2->metric >= rp->metric changed to rp2->metric > rp->metric */
	    if(rp2 == NULLROUTE || rp2->metric > rp->metric) {
		rp2 = rt_add(rp->target,32,rp->gateway,rp->iface,rp->metric,
			       0L,rp->flags & RTPRIVATE);
		rp2->uses = rp->uses;
	    }
	    rp2 = rp->next;
	    stop_timer(&rp->timer);	/* Stop timer before free() -- KZ1F */
	    free((char *)rp);
	    rp = rp2;
	}
    }

    /* Reset added flag */
    for(adj = Adjs; adj != NULLADJ; adj = adj->next)
	adj->added = 0;

    for(rr = Rspfrouters; rr != NULLRROUTER; rr = rr->next) {
	i = len_p(rr->data) - RSPFNODELEN;
	/* Jump past the node header. */
	if(dup_p(&bp,rr->data,RSPFNODELEN,i) != i) {
	    free_p(bp);
	    continue;
	}
	adjcnt = 0;
	while(i > 0) {
	    if(adjcnt) {
		if(!Keeprouter && (*bp->data & 128) == 0) {
		    /* This router has an adjacency that was not added. That
		     * means that the router is unreachable. Mark the
		     * stored routing update for deletion.
		     */
		    free_p(bp);
		    free_p(rr->data);
		    rr->data = NULLBUF;	/* indicates disposal */
		    break;
		}
		*bp->data &= ~128;	/* Clear the "added" bit */
		pullup(&bp,NULLCHAR,5);
		i -= 5;
		--adjcnt;
		continue;
	    }
	    ntohrspflink(&linkh,&bp);
	    adjcnt = linkh.adjn;
	    i -= RSPFLINKLEN;
        }
    }

    rt_merge(Rspf_trace);	/* remove routes covered by a group route */

    if(!Keeprouter) {
	/* Delete any routers that were discovered as being unreachable */
	rrprev = NULLRROUTER;
	rr = Rspfrouters;
	while(rr != NULLRROUTER) {
	    if(rr->data == NULLBUF) {
		if(rrprev != NULLRROUTER)
		    rrprev->next = rr->next;
		else
		    Rspfrouters = rr->next;
		free((char *)rr);
		rr = (rrprev != NULLRROUTER) ? rrprev->next : Rspfrouters;
	    } else {
		rrprev = rr;
		rr = rr->next;
	    }
	}
    }
}

/* Find a route from addr with the lowest cost. Returns a pointer to the
 * buffer that keeps the significant bit count of the selected route.
 */
static char *
findlowroute(addr,low,add,resaddr,resbits)
register int32 addr;		/* The node whose routes will be examined */
register int *low;		/* Lowest cost so far */
register int add;		/* Cost of this node */
register int32 *resaddr;	/* Resulting route stored here */
register int *resbits;		/* Significant bits of resulting route */
{
    struct mbuf *bp;
    register struct rspfrouter *rr;
    struct rspflinkh linkh;
    register struct route *rp;
    register char *r, *retval = NULLCHAR;
    register int adjcnt = 0;

    if((rr = rr_lookup(addr)) == NULLRROUTER)
	return NULLCHAR;
    if((rp = rt_blookup(addr,32)) != NULLROUTE && rp->metric < add)
	return NULLCHAR;	/* already added this one, no loops thanks */
    if(dup_p(&bp,rr->data,RSPFNODELEN,len_p(rr->data) - RSPFNODELEN) !=
       len_p(rr->data) - RSPFNODELEN) {
	free_p(bp);
	return NULLCHAR;
    }
    linkh.cost = 0;
    while(len_p(bp) > 0) {
	if(adjcnt) {
	    if(*bp->data & 128) {
		(void) PULLCHAR(&bp);
		if((r = findlowroute(pull32(&bp),low,add+linkh.cost,resaddr,
		  resbits)) != NULLCHAR)
		    retval = r;
	    } else {
		*low = add + linkh.cost;
		retval = bp->data;
		*resbits = PULLCHAR(&bp);
		*resaddr = pull32(&bp);
		pullup(&bp,NULLCHAR,5 * (adjcnt - 1));
		adjcnt = 1;	/* No need to check the rest of this link */
	    }
	    --adjcnt;
	    continue;
	}
	ntohrspflink(&linkh,&bp);
	if((add + linkh.cost) <= *low)
	    adjcnt = linkh.adjn;
	else
	    pullup(&bp,NULLCHAR,5*linkh.adjn);
    }
    return retval;
}

/* Search through the routing table for a manual, non private entry
 * with less then 32 bits, covering our target, return that route or
 * NULLROUTE if not found -- sm6rpz
 */
static struct route *
rt_clookup(target)
register int32 target;
{
    register struct route *rp;
    register int bits, i;

    for(bits = 31; bits > 0; bits--) {
	target &= ~0L << (32 - bits); /* Mask off target according to width */
	for(i = 0; i < HASHMOD; i++)
	    for(rp = Routes[bits - 1][i]; rp != NULLROUTE; rp = rp->next)
		if(!(rp->flags & RTPRIVATE) && !dur_timer(&rp->timer)
		  && rp->target == target)
		    return rp;
    }
    return NULLROUTE;
}

/* Search through the list af adjacencies for addr or iface. The iface
 * option is used to see if we have any adjacencies on the given interface.
 * Addr should be set to 0L to enable this. -- sm6rpz
 */
static struct rspfadj *
adj_lookup(addr,iface)
register int32 addr;
register struct iface *iface;
{
    register struct rspfadj *adj;

    for(adj = Adjs; adj != NULLADJ; adj = adj->next)
	if(addr != 0L && adj->addr == addr ||
	   iface != NULLIF && adj->riface->iface == iface)
	    return adj;
    return NULLADJ;
}
