/* SLIP (Serial Line IP) encapsulation and control routines.
 * Copyright 1992 William Allen Simpson
 *	partly based on a previous design by Phil Karn, KA9Q
 *
 *	- Feb 1991	Bill_Simpson@um.cc.umich.edu
 *			reflect changes to header compression calls
 *			revise status display
 *	- Jul 1991	Bill_Simpson@um.cc.umich.edu
 *			complete re-write to match PPP design
 *			more efficient, elegant, no static structures
 *	- Dec 1991	Bill_Simpson@um.cc.umich.edu
 *			move common variables into interface structure
 */
#include <stdio.h>
#include "global.h"
#include "mbuf.h"
#include "iface.h"
#include "slhc.h"
#include "i8250.h"
#include "slip.h"
#include "trace.h"

static void slip_log __ARGS((struct slip_s *slip_p, char *comment));
static void slip_error __ARGS((struct slip_s *slip_p, struct mbuf *bp, char *comment));
static void slip_skipped __ARGS((struct slip_s *slip_p, struct mbuf *bp, char *comment));

static int slip_send __ARGS((struct mbuf *bp,struct iface *iface,int32 gateway,
	int prec,int del,int tput,int rel));

static void slip_recv __ARGS((int dev,void *p1,void *p2));

static void slip_show __ARGS((struct iface *iface));
static void slip_free __ARGS((struct iface *iface));


/****************************************************************************/
/* General log routine */
static void
slip_log( slip_p, comment )
struct slip_s *slip_p;
char *comment;
{
	if (slip_p->trace)
		trace_log(slip_p->iface,"SLIP %s",
			comment);
}


/* Bad packet */
static void
slip_error(slip_p, bp, comment)
struct slip_s *slip_p;
struct mbuf *bp;
char *comment;
{
	free_p(bp);
	slip_log( slip_p, comment );
}


/* Unknown type input packet */
static void
slip_skipped(slip_p, bp, comment)
struct slip_s *slip_p;
struct mbuf *bp;
char *comment;
{
	slhc_toss( slip_p->slhcp );
	slip_error( slip_p, bp, comment );
}


/****************************************************************************/
/* Send routine for point-to-point slip */
static int
slip_send(bp,ifp,gateway,prec,del,tput,rel)
struct mbuf *bp;	/* Buffer to send */
struct iface *ifp;	/* Pointer to interface control block */
int32 gateway;		/* Ignored (SLIP is point-to-point) */
int prec;
int del;
int tput;
int rel;
{
	struct slip_s *slip_p;

	if (ifp == NULLIF
	 || (slip_p = ifp->edv) == NULL){
		free_p(bp);
		return -1;
	}

	if (slip_p->flags & SLIP_VJWANT) {
		/* Will attempt TCP/IP header compression.
		 * Note that the mbuf referenced by bp may change,
		 * therefore the ip version is referenced by its location
		 * at the first byte of the (possibly new) packet.
		 */
		int type = slhc_compress( slip_p->slhcp, &bp,
			slip_p->flags & SLIP_VJSLOT);
		bp->data[0] |= type;
	}
	return (*ifp->raw)(ifp,bp);
}


/* Send a raw slip frame */
int
slip_raw(ifp,bp)
struct iface *ifp;
struct mbuf *bp;
{
	struct slip_s *slip_p = ifp->edv;
	struct mbuf *vbp;
	register char *cp;
	register int c;

	dump(ifp,IF_TRACE_OUT,ifp->e_type,bp);
	ifp->OutOctets += len_p(bp);
	ifp->rawsndcnt++;
	ifp->lastsent = secclock();

	/* Allocate output mbuf that's twice as long as the packet.
	 * This is a worst-case guess (consider a packet full of FR_ENDs!)
	 */
	if ( (vbp = alloc_mbuf((int16)(2*len_p(bp) + 2))) == NULLBUF ) {
		slip_error( slip_p, bp, Nospace );
		ifp->OutDiscards++;
		return -1;
	}
	cp = vbp->data;

	/* No need to send an opening flag if the previous packet is still
	 * being transmitted.
	 */
#if 0
	if ( (*ifp->outqlen)(ifp) == 0 ) {
#endif
	if ( Asy[ifp->d_index].dma.flags == 0 ) {
		/* Flush out any line garbage */
		*cp++ = FR_END;
		slip_p->OutOpenFlag++;
	}

	/* Copy input to output, escaping special characters */
	while ((c = PULLCHAR(&bp)) != -1) {
		switch(c){
		case FR_ESC:
			*cp++ = FR_ESC;
			*cp++ = T_FR_ESC;
			break;
		case FR_END:
			*cp++ = FR_ESC;
			*cp++ = T_FR_END;
			break;
		default:
			*cp++ = c;
		}
	}

	/* Tie off the packet */
	*cp++ = FR_END;
	vbp->cnt = cp - vbp->data;

	if (ifp->trace & IF_TRACE_RAW)
		raw_dump(ifp,IF_TRACE_OUT /* -1? */,vbp);
	return (*ifp->iosend)(ifp->d_index,vbp);
}


/****************************************************************************/
/* Packetize SLIP input from device */
/* (process started by slip_init) */
static void
slip_recv(dev,p1,p2)
int dev;
void *p1;
void *p2;
{
	struct iface *ifp = p1;
	struct slip_s *slip_p = ifp->edv;
	struct mbuf *raw_bp = NULLBUF;
	register struct mbuf *head_bp = NULLBUF;
	char *cp;			/* next byte in mbuf */
	register int c;
#define MODE_TOSS	0x01
#define MODE_ESCAPED	0x02
	int mode = MODE_TOSS;

	while ( (c = (*ifp->ioget)(dev)) != -1 ) {
		if ( raw_bp != NULLBUF	&&  raw_bp->cnt < raw_bp->size ) {
			*raw_bp->data++ = c;
			raw_bp->cnt++;
		}

		if ( c == FR_END ) {
			if ( raw_bp != NULLBUF ) {
				raw_bp->data = (char *)(raw_bp + 1);
				if ( raw_bp->cnt != 1 ) {
					raw_dump( ifp, IF_TRACE_IN, raw_bp );
				}
				if ( ifp->trace & IF_TRACE_RAW ) {
					raw_bp->cnt = 0;
				} else {
					free_p(raw_bp);
					raw_bp = NULLBUF;
				}
			} else if ( ifp->trace & IF_TRACE_RAW ) {
				raw_bp = alloc_mbuf( 2 * ifp->mru);
			}

			if ( mode & MODE_ESCAPED ) {
				slip_skipped( slip_p, head_bp,
					"deliberate cancellation" );
				ifp->InDiscards++;
			} else if ( mode & MODE_TOSS ) {
				free_p( head_bp );
			} else if ( head_bp != NULLBUF ) {
				net_route( ifp, ifp->e_type, head_bp );
			} else {
				slip_p->InOpenFlag++;
			}

			/* setup for next buffer */
			head_bp = NULLBUF;
			mode = FALSE;
			continue;
		}

		/* We reach here for every byte inside a frame.
		 * Check for escape sequence.
		 */
		if ( mode & MODE_ESCAPED ) {
			mode &= ~MODE_ESCAPED;

			if ( c == T_FR_ESC ) {
				c = FR_ESC;
			} else if ( c == T_FR_END ) {
				c = FR_END;
			} else {
				slip_skipped( slip_p, head_bp,
					"Illegal escape sequence" );
				ifp->InErrors++;
				head_bp = NULLBUF;
				mode |= MODE_TOSS;
				continue;
			}
		} else if ( c == FR_ESC ) {
			mode |= MODE_ESCAPED;
			continue;
		}

		/* We reach here with a byte for the buffer.
		 * Make sure there is room for it.
		 */
		if ( head_bp == NULLBUF ) {
			if ( mode & MODE_TOSS ) {
				continue;
			}
			if ( (head_bp = alloc_mbuf( ifp->mru )) == NULLBUF ) {
				slip_skipped( slip_p, head_bp, Nospace );
				ifp->InDiscards++;
				mode |= MODE_TOSS;
				continue;
			}
			cp = head_bp->data;
			head_bp->cnt = 0;
		} else if ( head_bp->cnt >= head_bp->size ) {
			/* too much data */
			slip_skipped( slip_p, head_bp, "MRU exceeded" );
			ifp->InErrors++;
                        head_bp = NULLBUF;
			mode |= MODE_TOSS;
			continue;
		}

		/* Store the byte, increment counts */
		*cp++ = c;
		head_bp->cnt++;
	}

	/* clean up afterward */
	free_p(head_bp);
	ifp->rxproc = NULLPROC;
}


/* Process incoming SLIP packets */
/* (called from network task) */
void
slip_proc( ifp, bp )
struct iface *ifp;
struct mbuf *bp;
{
	struct slip_s *slip_p;
	register int c;

	if ( ifp == NULLIF ) {
		log(-1, "slip_proc: missing iface" );
		return;
	}
	if ( bp == NULLBUF ) {
		trace_log(ifp, "slip_proc: missing buffer" );
		return;
	}

	slip_p = ifp->edv;
	ifp->InOctets += bp->cnt;

	if ((c = bp->data[0]) & SL_TYPE_COMPRESSED_TCP) {
		if ( !(slip_p->flags & SLIP_VJWANT) ) {
			slip_error( slip_p, bp,
				"unexpected VJ compressed packet" );
			ifp->InDiscards++;
			return;
		}
		if ( slhc_uncompress(slip_p->slhcp, &bp) < IPLEN+TCPLEN ) {
			slip_error( slip_p, bp,
				"compressed packet failed" );
			ifp->InErrors++;
			return;
		}
	} else if (c >= SL_TYPE_UNCOMPRESSED_TCP) {
		if ( !(slip_p->flags & (SLIP_VJWANT | SLIP_VJWILL)) ) {
			slip_error( slip_p, bp,
				"unexpected VJ uncompressed packet" );
			ifp->InDiscards++;
			return;
		}
		bp->data[0] &= 0x4f;
		if ( slhc_remember(slip_p->slhcp, &bp) < IPLEN+TCPLEN ) {
			slip_error( slip_p, bp,
				"uncompressed packet failed" );
			ifp->InErrors++;
			return;
		}
		if ( slip_p->flags & SLIP_VJWILL ) {
			slip_p->flags |= SLIP_VJWANT;
		}
	}

	ip_route(ifp,bp,0);
}



/* Show serial line status */
static void
slip_show(ifp)
struct iface *ifp;
{
	struct slip_s *slip_p = ifp->edv;

	slhc_i_status(slip_p->slhcp);
	slhc_o_status(slip_p->slhcp);
}


/****************************************************************************/
/* Initialize SLIP control structures for a Point-to-Point interface */
/* (called from setencap) */
int
slip_init(ifp)
struct iface *ifp;
{
	struct slip_s *slip_p;
	char *ifn;

	slip_p = callocw(1,sizeof(struct slip_s));

	ifp->edv = slip_p;
	ifp->send = slip_send;
	ifp->put = NULL;		/* ARP isn't used */
	ifp->raw = slip_raw;
	ifp->show = slip_show;
	ifp->detach = slip_free;

	slip_p->iface = ifp;
	slip_p->flags = ifp->flags;

	if ( slip_p->flags & (SLIP_VJWANT | SLIP_VJWILL) ) {
		slip_p->slhcp = slhc_init( 16, 16 );
	}

	ifp->rxproc = newproc( ifn = if_name( ifp, " receive" ),
			320, slip_recv, ifp->d_index, ifp, NULL, 0);
	free(ifn);
	return 0;
}


static void
slip_free(ifp)
struct iface *ifp;
{
	struct slip_s *slip_p = ifp->edv;

	alert( ifp->rxproc, -1 );

	slhc_free( slip_p->slhcp );
	free( slip_p );
	ifp->edv = NULL;
}

