/*
 *	CLIENT routines for Simple Mail Transfer Protocol ala RFC821
 *	A.D. Barksdale Garbee II, aka Bdale, N3EUA
 *	Copyright 1986 Bdale Garbee, All Rights Reserved.
 *	Permission granted for non-commercial copying and use, provided
 *	this notice is retained.
 * 	Modified 14 June 1987 by P. Karn for symbolic target addresses,
 *	also rebuilt locking mechanism
 *	Copyright 1987 1988 David Trulli, All Rights Reserved.
 *	Permission granted for non-commercial copying and use, provided
 *	this notice is retained.
 *
 *	Jan 92	Bill Simpson -- fix leading '.' problem.
 *		Use correct RFC821 line sizes, defined with symbols.
 *		Eliminate many stub routines and straighten code.
 *		Drastically reduce static data usage.
 *		(changes Copyright 1992 William Allen Simpson)
 */
#include <stdio.h>
#include <time.h>
#ifdef UNIX
#include <sys/types.h>
#endif
#ifdef	AMIGA
#include <stat.h>
#else
#include <sys/stat.h>
#endif
#ifdef	__TURBOC__
#include <io.h>
#endif

#include "global.h"
#ifdef	ANSIPROTO
#include <stdarg.h>
#endif
#include "timer.h"
#include "proc.h"
#include "socket.h"
#include "session.h"
#include "netuser.h"
#include "cmdparse.h"
#include "commands.h"
#include "dirutil.h"
#include "files.h"
#include "mailutil.h"
#include "smtp.h"

#define SMTPSTKSIZE	3072
#define MAXSESSIONS	10		/* most connections allowed */

static struct timer Smtpcli_t;
static int32 Gateway;

#ifdef SMTPTRACE
static unsigned short Smtptrace = 0;		/* used for trace level */
static int dosmtptrace __ARGS((int argc,char *argv[],void *p));
#endif

static int Smtpbatch;

/* used as a queue entry for a single message */
struct jobq_s {
	struct	jobq_s *next;	/* pointer to next mail job for this system */
	char	jobname[9];	/* the prefix of the job file name */
	char	*from;		/* address of sender */
	struct recipient_s *to;	/* Linked list of recipients */
};
#define NULLJOB 	(struct jobq_s *)0

/* control structure used by a smtp client session */
struct smtpcli_s {
	struct smtpcli_s *next;
	int 	s;		/* connection socket */
	int32	ipdest;		/* address of forwarding system */
	char	*destname;	/* domain address of forwarding system */
	char	*wname; 	/* name of work file */
	char	*tname; 	/* name of text file */
	char	*tbuf;
	struct	jobq_s	*jobq;
	struct	recipient_s *errlog;
	int lock;		/* In use */
};
#define NULLSMTPCLI	(struct smtpcli_s *)0

static struct smtpcli_s *cli_sessions = NULLSMTPCLI; /* queue of client sessions  */
static unsigned short Smtpmaxcli = MAXSESSIONS;
static int Smtpsessions = 0;		/* client connections currently open */

static void del_job __ARGS((struct jobq_s *jp));
static void del_session __ARGS((struct smtpcli_s *cb));
static int dogateway __ARGS((int argc,char *argv[],void *p));
static int dosmtpmaxcli __ARGS((int argc,char *argv[],void *p));
static int dotimer __ARGS((int argc,char *argv[],void *p));
static int dosmtpkill __ARGS((int argc,char *argv[],void *p));
static int dosmtplist __ARGS((int argc,char *argv[],void *p));
static int dobatch __ARGS((int argc,char *argv[],void *p));
static int getresp __ARGS((struct smtpcli_s *ftp,int mincode));
static void logerr __ARGS((struct smtpcli_s *cb,char *line));
static void retmail __ARGS((struct smtpcli_s *cb));
static void sendcmd __ARGS((struct smtpcli_s *cb,char *fmt,...));
static int setsmtpmode __ARGS((int argc,char *argv[],void *p));
static struct jobq_s *setupjob __ARGS((struct smtpcli_s *cb,char *id,char *from));
static void smtp_send __ARGS((int unused,void *cb1,void *p));
static int smtpkick __ARGS((int argc,char *argv[],void *p));

static struct cmds Smtpcmds[] = {
	"batch",	dobatch,	0,	0,	NULLCHAR,
	"gateway",	dogateway,	0,	0,	NULLCHAR,
	"mode",		setsmtpmode,	0,	0,	NULLCHAR,
	"kick",         smtpkick,       2048,   0,      NULLCHAR,
	"kill",		dosmtpkill,	0,	2,	"kill <jobnumber>",
	"list",		dosmtplist,	0,	0,	NULLCHAR,
	"maxclients",	dosmtpmaxcli,	0,	0,	NULLCHAR,
	"timer",	dotimer,	0,	0,	NULLCHAR,
#ifdef SMTPTRACE
	"trace",	dosmtptrace,	0,	0,	NULLCHAR,
#endif
	NULLCHAR,
};

int
dosmtp(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	return subcmd(Smtpcmds,argc,argv,p);
}

static int
dobatch(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	return setbool(&Smtpbatch,"SMTP batching",argc,argv);
}
static int
dosmtpmaxcli(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	return setshort(&Smtpmaxcli,"Max clients",argc,argv);
}

static int
setsmtpmode(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	if (argc < 2) {
		tprintf("smtp mode: %s\n",
			(Smtpmode & QUEUE) ? "queue" : "route");
	} else {
		switch(*argv[1]) {
		case 'q':
			Smtpmode |= QUEUE;
			break;
		case 'r':
			Smtpmode &= ~QUEUE;
			break;
		default:
			tprintf("Usage: smtp mode [queue | route]\n");
			break;
		}
	}
	return 0;
}
static int
dogateway(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	int32 n;

	if(argc < 2){
		tprintf("%s\n",inet_ntoa(Gateway));
	} else if((n = resolve(argv[1])) == 0){
		tprintf(Badhost,argv[1]);
		return 1;
	} else
		Gateway = n;
	return 0;
}

#ifdef SMTPTRACE
static int
dosmtptrace(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	return setshort(&Smtptrace,"SMTP tracing",argc,argv);
}
#endif

/* list jobs waiting to be sent in the mqueue */
static int
dosmtplist(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	char filepath[FILE_PATH_SIZE];
	char filename[FILE_NAME_SIZE];
	char *filedot;
	char host[ULINELEN];
	char from[ULINELEN];
	char   to[ULINELEN];
	char status;
	struct stat sbuf;
	struct tm *tminfo, *localtime();
	FILE *fp;

	Current->flowmode = 1; /* Enable the more mechanism */

	tprintf("S     Job    Size Date  Time  Host                 From\n");
	filedir(Mailqueue,0,filename);
	while(filename[0] != '\0') {
		sprintf(filepath,"%s/%s",Mailqdir,filename);
		if ((fp = fopen(filepath,READ_TEXT)) == NULLFILE) {
			tprintf("Can't open %s: %s\n",filepath,sys_errlist[errno]);
			continue;
		}
		if ((filedot = strrchr(filename,'.')) != NULLCHAR)
			*filedot = '\0';

		sprintf(filepath,"%s/%s.lck",Mailqdir,filename);
		if (access(filepath,0))
			status = ' ';
		else
			status = 'L';

		sprintf(filepath,"%s/%s.txt",Mailqdir,filename);
		stat(filepath,&sbuf);
		tminfo = localtime(&sbuf.st_ctime);
		fgets(host,sizeof(host),fp);
		rip(host);
		fgets(from,sizeof(from),fp);
		rip(from);
		tprintf("%c %7s %9ld %02d/%02d %02d:%02d %-25s %-20s\n      ",
			status, filename, sbuf.st_size,
			tminfo->tm_mon+1,
			tminfo->tm_mday,
			tminfo->tm_hour,
			tminfo->tm_min,
			host,
			from);
		while (fgets(to,sizeof(to),fp) != NULLCHAR) {
			rip(to);
			tprintf("%s\n",to);
		}
		(void) fclose(fp);
		pwait(NULL);
		filedir(Mailqueue,1,filename);
	}
	Current->flowmode = 0;
	return 0;
}

/* kill a job in the mqueue */
static int
dosmtpkill(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	char filepath[FILE_PATH_SIZE];
	char *filedot;
	char c;

	sprintf(filepath,"%s/%s.lck",Mailqdir,argv[1]);
	filedot = strrchr(filepath,'.');
	if (!access(filepath,0)) {
		Current->ttystate.echo = Current->ttystate.edit = 0;
		c = keywait("Warning, the job is locked by SMTP. Remove (y/n)? ",0);
		Current->ttystate.echo = Current->ttystate.edit = 1;
		if (c != 'y')
			return 0;
		(void) unlink(filepath);
	}
	strcpy(filedot,".wrk");
	if (unlink(filepath))
		tprintf("Job id %s not found\n",argv[1]);
	strcpy(filedot,".txt");
	(void) unlink(filepath);
	return 0;
}

/* Set outbound spool scan interval */
static int
dotimer(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	if(argc < 2){
		tprintf("%lu/%lu\n",
			read_timer(&Smtpcli_t) / 1000L,
			dur_timer(&Smtpcli_t) / 1000L);
		return 0;
	}
	Smtpcli_t.func = (void (*)())smtptick;/* what to call on timeout */
	Smtpcli_t.arg = NULL;		/* dummy value */
	set_timer(&Smtpcli_t,atol(argv[1])*1000L);	/* set timer duration */
	start_timer(&Smtpcli_t);		/* and fire it up */
	return 0;
}

static int
smtpkick(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	int32 addr = 0;
	if(argc > 1 && (addr = resolve(argv[1])) == 0){
		tprintf(Badhost,argv[1]);
		return 1;
	}
	smtptick((void *)addr);
	return 0;
}

/* This is the routine that gets called every so often to do outgoing
 * mail processing. When called with a null argument, it runs the entire
 * queue; if called with a specific non-zero IP address from the remote
 * kick server, it only starts up sessions to that address.
 */
int
smtptick(t)
void *t;
{
	register struct smtpcli_s *cb;
	struct jobq_s *jp;
	char wfilename[FILE_NAME_SIZE];
	char *filedot;
	char host[ULINELEN];
	char from[ULINELEN];
	char to[ULINELEN];
	int32 destaddr;
	int32 target = (int32)t;
	FILE *wfile;

	if(availmem() - SMTPSTKSIZE < Memthresh){
		/* Memory is tight, don't do anything */
#ifdef SMTPTRACE
		if (Smtptrace >= 3)
			tprintf("SMTP tick exit -- low memory\n");
#endif
		start_timer(&Smtpcli_t);
		return 0;
	}
#ifdef SMTPTRACE
	if (Smtptrace >= 3)
		tprintf("smtp daemon entered, target = %s\n",inet_ntoa(target));
#endif
	for(filedir(Mailqueue,0,wfilename);wfilename[0] != '\0';
		filedir(Mailqueue,1,wfilename)){

		if ((filedot = strrchr(wfilename,'.')) != NULLCHAR)
			*filedot = '\0';

		/* lock this file from the smtp daemon */
		if (mlock(Mailqdir,wfilename))
			continue;

		sprintf(host,"%s/%s.wrk",Mailqdir,wfilename);
		if ((wfile = fopen(host,READ_TEXT)) == NULLFILE) {
			/* probably too many open files */
			(void) rmlock(Mailqdir,wfilename);

			/* continue to next message.
			 * The failure may be temporary
			 */
			continue;
		}

		(void) fgets(host,ULINELEN,wfile);  /* read target host */
		rip(host);

		if ((destaddr = mailroute(host)) == 0) {
			fclose(wfile);
			tprintf("** smtp: Unknown address %s\n",host);
			(void) rmlock(Mailqdir,wfilename);
			continue;
		}
		if(target != 0 && destaddr != target){
			fclose(wfile);
			(void) rmlock(Mailqdir,wfilename);
			continue;	/* Not the proper target of a kick */
		}

		for ( cb = cli_sessions; cb != NULLSMTPCLI; cb = cb->next ) {
			if ( cb->ipdest == destaddr) {
				break;
			}
		}

		if ( cb == NULLSMTPCLI ) {
			if (Smtpsessions >= Smtpmaxcli) {
#ifdef SMTPTRACE
				if (Smtptrace) {
					tprintf("smtp daemon: too many processes\n");
				}
#endif
				fclose(wfile);
				(void) rmlock(Mailqdir,wfilename);
				break;
			}

			cb = (struct smtpcli_s *)callocw(1,sizeof(struct smtpcli_s));
			cb->ipdest = destaddr;
			cb->destname = strdup(host);

			cb->next = cli_sessions;
			cli_sessions = cb;
			Smtpsessions++;
		} else {
			if(cb->lock){
				/* This system is already sending mail,
				 * let's not interfere with its send queue.
				 */
				fclose(wfile);
				(void) rmlock(Mailqdir,wfilename);
				continue;
			}
		}

		(void) fgets(from,sizeof(from),wfile);	     /* read from */
		rip(from);
		if ((jp = setupjob(cb,wfilename,from)) == NULLJOB) {
			fclose(wfile);
			del_session(cb);
			(void) rmlock(Mailqdir,wfilename);
			break;
		}
		while (fgets(to,sizeof(to),wfile) != NULLCHAR) {
			rip(to);
			if (add_rcpt(&jp->to,to,DOMAIN) == NULLRCPT) {
				fclose(wfile);
				del_session(cb);
			}
		}
		fclose(wfile);
#ifdef SMTPTRACE
		if (Smtptrace >= 2) {
			struct recipient_s *ap;
			tprintf("queue job %s From: %s To:",wfilename,from);
			for (ap = jp->to; ap != NULLRCPT; ap = ap->next)
				tprintf(" %s",ap->val);
			tprintf("\n");
		}
#endif
	}

	/* start sending that mail */
	for ( cb = cli_sessions; cb != NULLSMTPCLI; cb = cb->next ) {
		if(cb->lock)
			continue;

		newproc("smtp_send", SMTPSTKSIZE, smtp_send, 0, cb,NULL,0);

#ifdef SMTPTRACE
		if (Smtptrace) {
			tprintf("Trying Connection to %s\n",
				inet_ntoa(cb->ipdest));
		}
#endif
	}

	/* Restart timer */
	start_timer(&Smtpcli_t);
	return 0;
}


/* This is the master state machine that handles a single SMTP transaction.
 * It is called with a queue of jobs for a particular host.
 * The logic is complicated by the "Smtpbatch" variable, which controls
 * the batching of SMTP commands. If Smtpbatch is true, then many of the
 * SMTP commands are sent in one swell foop before waiting for any of
 * the responses. Unfortunately, this breaks many brain-damaged SMTP servers
 * out there, so provisions have to be made to operate SMTP in lock-step mode.
 */
static void
smtp_send(unused,cb1,p)
int unused;
void *cb1;
void *p;
{
	register struct smtpcli_s *cb = (struct smtpcli_s *)cb1;
	register struct recipient_s *tp;
	struct sockaddr_in fsocket;
	FILE *tfile;
	char tbuf[TLINELEN];	/* Output buffer */
	char *cp;
	int rcode;
	int rcpts;
	int goodrcpt;
	int init = 3;	/* responses to receive */

	cb->lock = TRUE;
	cb->tbuf = tbuf;	/* save address of buffer for commands */

	fsocket.sin_family = AF_INET;
	fsocket.sin_addr.s_addr = cb->ipdest;
	fsocket.sin_port = IPPORT_SMTP;

	cb->s = socket(AF_INET,SOCK_STREAM,0);
	sockmode(cb->s,SOCK_ASCII);
	setflush(cb->s,-1);	/* We'll explicitly flush before reading */
#ifdef SMTPTRACE
	if (Smtptrace) 
		tprintf("SMTP client Trying...\n");
#endif
	if(connect(cb->s,(char *)&fsocket,SOCKSIZE) == 0){
#ifdef SMTPTRACE
	if (Smtptrace) 
		tprintf("Connected\n");
#endif
		;
	} else {
		cp = sockerr(cb->s);
#ifdef SMTPTRACE
		if (Smtptrace) 
			tprintf("Connect failed: %s\n",cp != NULLCHAR ? cp : "");
#endif
		log(cb->s,"SMTP Connect failed: %s",
			cp != NULLCHAR ? cp : "");
	}
	if(!Smtpbatch){
		rcode = getresp(cb,200);
		if(rcode == -1 || rcode >= 400)
			goto quit;
	}
	/* Say HELO */
	sendcmd(cb,"HELO %s\n",Hostname);
	if(!Smtpbatch){
		rcode = getresp(cb,200);
		if(rcode == -1 || rcode >= 400)
			goto quit;
	}

	cb->wname = mallocw(strlen(Mailqdir) + 1 + FILE_NAME_SIZE);
	cb->tname = mallocw(strlen(Mailqdir) + 1 + FILE_NAME_SIZE);

	/* For each message... */
	while ( cb->jobq != NULLJOB ) {
		register struct jobq_s *jp = cb->jobq;

#ifdef SMTPTRACE
		if (Smtptrace >= 3) {
			tprintf("sending job %s\n",jp->jobname);
		}
#endif
		sprintf(cb->wname,"%s/%s.wrk",Mailqdir,jp->jobname);
		sprintf(cb->tname,"%s/%s.txt",Mailqdir,jp->jobname);

		/* if this file open fails, skip it */
		if ((tfile = fopen(cb->tname,READ_TEXT)) == NULLFILE) {
			cb->jobq = jp->next;
			del_job(jp);
			continue;
		}

		/* Send MAIL and RCPT commands */
		sendcmd(cb,"MAIL FROM:<%s>\n",jp->from);
		if(!Smtpbatch){
			rcode = getresp(cb,200);
			if(rcode == -1 || rcode >= 400)
				goto quit;
		}
		rcpts = 0;
		goodrcpt = FALSE;
		for (tp = jp->to; tp != NULLRCPT; tp = tp->next){
			sendcmd(cb,"RCPT TO:<%s>\n",tp->val);
			if(!Smtpbatch){
				rcode = getresp(cb,200);
				if(rcode == -1)
					goto quit;
				if(rcode < 400)
					goodrcpt = TRUE; /* At least one good */
			}
			rcpts++;
		}
		/* Send DATA command */
		sendcmd(cb,"DATA\n");
		if(!Smtpbatch){
			rcode = getresp(cb,200);
			if(rcode == -1 || rcode >= 400)
				goto quit;
		}
		if(Smtpbatch){
			/* Now wait for the responses to come back. The first time
			 * we do this, we wait first for the start banner and
			 * HELO response. In any case, we wait for the response to
			 * the MAIL command here.
			 */
			while( init-- > 0 ) {
				rcode = getresp(cb,200);
				if(rcode == -1 || rcode >= 400)
					goto quit;
			}
			init = 1;

			/* Now process the responses to the RCPT commands */
			while ( rcpts-- > 0 ) {
				rcode = getresp(cb,200);
				if(rcode == -1)
					goto quit;
				if(rcode < 400)
					goodrcpt = TRUE; /* At least one good */
			}
			/* And finally get the response to the DATA command.
			 * Some servers will return failure here if no recipients
			 * are valid, some won't.
			 */
			rcode = getresp(cb,200);
			if(rcode == -1 || rcode >= 400)
				goto quit;

			/* check for no good rcpt on the list */
			if (!goodrcpt){
				sendcmd(cb,".\n");  /* Get out of data mode */
				goto quit;
			}
		}

		/* Send the file */
		strcpy(tbuf,"\n");
		while(fgets(tbuf,sizeof(tbuf),tfile) != NULLCHAR) {
			/* Escape a '.' character at the beginning of a line */
			if ( tbuf[0] == '.' )
				usputc(cb->s,'.');
			usputs(cb->s,tbuf);
		}
		fclose(tfile);
		tfile = NULLFILE;

		/* Send the end-of-message command */
		if(tbuf[strlen(tbuf)-1] == '\n')
			sendcmd(cb,".\n");
		else
			sendcmd(cb,"\n.\n");

		/* Wait for the OK response */
		rcode = getresp(cb,200);
		if(rcode == -1)
			goto quit;
		if((rcode >= 200 && rcode < 300) || rcode >= 500){
			/* if a good transfer or permanent failure remove job */

			if (cb->errlog != NULLRCPT)
				retmail(cb);

			(void) unlink(cb->tname);	/* unlink textfile */
			(void) unlink(cb->wname);	/* unlink workfile */
			log(cb->s,"SMTP sent job %s To: %s From: %s",
				jp->jobname,
				jp->to->val,
				jp->from);
		}

		/* remove the error log of previous message */
		del_rcpt(cb->errlog);
		cb->errlog = NULLRCPT;

		cb->jobq = jp->next;
		del_job(jp);
	}
quit:
	sendcmd(cb,"QUIT\n");
	if (cb->errlog != NULLRCPT){
		retmail(cb);
		(void) unlink(cb->wname);	/* unlink workfile */
		(void) unlink(cb->tname);	/* unlink text */
	}
	(void) close_s(cb->s);
	if(tfile != NULLFILE)
		fclose(tfile);
	del_session(cb);
}


/* free the message struct and data */
static void
del_session(cb)
struct smtpcli_s *cb;
{
	register struct jobq_s *jp,*tp;

	if (cb == NULLSMTPCLI)
		return;

	if ( cli_sessions == cb ) {
		cli_sessions = cb->next;
	} else {
		register struct smtpcli_s *pcb;

		for ( pcb = cli_sessions; pcb != NULLSMTPCLI; pcb = pcb->next ) {
			if ( pcb->next == cb ) {
				pcb->next = cb->next;
				break;
			}
		}
	}
	Smtpsessions--;	/* number of connections active */

	free(cb->wname);
	free(cb->tname);
	free(cb->destname);
	for (jp = cb->jobq; jp != NULLJOB;jp = tp) {
			tp = jp->next;
			del_job(jp);
	}
	del_rcpt(cb->errlog);
	free((char *)cb);
}


/* add this job to control block queue */
static struct jobq_s *
setupjob(cb,id,from)
struct smtpcli_s *cb;
char *id,*from;
{
	register struct jobq_s *p1,*p2;

	p1 = (struct jobq_s *)callocw(1,sizeof(struct jobq_s));
	p1->from = strdup(from);
	strcpy(p1->jobname,id);

	/* now add to end of jobq */
	if ((p2 = cb->jobq) == NULLJOB)
		cb->jobq = p1;
	else {
		while(p2->next != NULLJOB)
			p2 = p2->next;
		p2->next = p1;
	}
	return p1;
}


static void
del_job(jp)
register struct jobq_s *jp;
{
	if ( *jp->jobname != '\0')
		(void) rmlock(Mailqdir,jp->jobname);
	free(jp->from);
	del_rcpt(jp->to);
	free((char *)jp);
}


/* stub for calling mdaemon to return message to sender */
static void
retmail(cb)
struct smtpcli_s *cb;
{
	FILE *infile;
#ifdef SMTPTRACE
	if (Smtptrace >= 3) {
		tprintf("smtp job %s returned to sender\n",cb->jobq->jobname);
	}
#endif
	if ((infile = fopen(cb->tname,READ_TEXT)) == NULLFILE)
		return;
	mdaemon(infile,cb->jobq->from,cb->errlog,1);
	fclose(infile);
}


/* save line in error list */
static void
logerr(cb,line)
struct smtpcli_s *cb;
char *line;
{
	register struct recipient_s *lp,*tp;

	tp = (struct recipient_s *)callocw(1,sizeof(struct recipient_s));
	tp->val = strdup(line);

	/* find end of list */
	if ((lp = cb->errlog) == NULLRCPT)
		cb->errlog = tp;
	else {
		while(lp->next != NULLRCPT)
			lp = lp->next;
		lp->next = tp;
	}
}

/* do a printf() on the socket with optional local tracing */
#ifdef	ANSIPROTO
static void
sendcmd(struct smtpcli_s *cb,char *fmt, ...)
{
	va_list args;

	va_start(args,fmt);
#ifdef	SMTPTRACE
	if(Smtptrace){
		tprintf("smtp sent: ");
		vprintf(fmt,args);
	}
#endif
	vsprintf(cb->tbuf,fmt,args);
	usputs(cb->s,cb->tbuf);
	va_end(args);
}
#else
static void
sendcmd(cb,fmt,arg1,arg2,arg3,arg4)
struct smtpcli_s *cb;
char *fmt;
int arg1,arg2,arg3,arg4;
{
#ifdef	SMTPTRACE
	if(Smtptrace){
		tprintf("smtp sent: ");
		tprintf(fmt,arg1,arg2,arg3,arg4);
	}
#endif
	sprintf(cb->tbuf,fmt,arg1,arg2,arg3,arg4);
	usputs(cb->s,cb->tbuf);
}
#endif

/* Wait for, read and display response from server. Return the result code. */
static int
getresp(cb,mincode)
struct smtpcli_s *cb;
int mincode;	/* Keep reading until at least this code comes back */
{
	int rval;
	char line[RLINELEN];

	usflush(cb->s);
	for(;;){
		/* Get line */
		if(recvline(cb->s,line,RLINELEN) == -1){
			rval = -1;
			break;
		}
		rip(line);		/* Remove cr/lf */
		rval = atoi(line);
#ifdef	SMTPTRACE
		if(Smtptrace)
			tprintf("smtp recv: %s\n",line);/* Display to user */
#endif
		if(rval >= 500) {	/* Save permanent error replies */
			char *tmp = mallocw(RLINELEN+24);

			if(cb->errlog == NULLRCPT) {
				sprintf(tmp,"While talking to %s:",
					cb->destname);
				logerr(cb,tmp);
			}
			if(cb->tbuf[0] != '\0') { /* Save offending command */
				rip(cb->tbuf);
				sprintf(tmp,">>> %s",cb->tbuf);
				logerr(cb,tmp);
				cb->tbuf[0] = '\0';
			}
			sprintf(tmp,"<<< %s",line);
			logerr(cb,tmp);		/* save the error reply */
			free(tmp);
		}
		/* Messages with dashes are continued */
		if(line[3] != '-' && rval >= mincode)
			break;
	}
	return rval;
}


/* Mail routing function. For now just use the hosts file */
int32
mailroute(dest)
char *dest;
{
	int32 destaddr;

	/* look up address or use the gateway */
	destaddr = resolve_mx(dest);
	if (destaddr == 0 && (destaddr = resolve(dest)) == 0)
		if (Gateway != 0)
			destaddr = Gateway; /* Use the gateway  */
	return destaddr;

}


