/* SMTP Server state machine - see RFC 821
 *  enhanced 4/88 Dave Trulli nn2z
 *
 *	Jan 92	Bill Simpson -- fix leading '.' problem.
 *		Do better error recovery on receive.
 *		Smarter stack, heap, buffer management.
 */
#include <stdio.h>
#include <time.h>
#ifdef UNIX
#include <sys/types.h>
#endif
#include <ctype.h>

#include "global.h"
#include "proc.h"
#include "socket.h"
#include "commands.h"
#include "mailutil.h"
#include "smtp.h"

/* Per-session control block used by smtp server */
struct smtpsv {
	int s;			/* the socket for this connection */
	char *system;		/* Name of remote system */
	char *from;		/* sender address */
	struct recipient_s *to; /* Linked list of recipients */
	FILE *data;		/* Temporary input file pointer */
};
#define NULLSMTPSV	(struct smtpsv *)0

static struct smtpsv *mail_create __ARGS((void));
static void mail_clean __ARGS((struct smtpsv *mp));
static void smtpserv __ARGS((int s,void *unused,void *p));

/* Command table */
static char *commands[] = {
	"helo",
#define	HELO_CMD	0
	"noop",
#define	NOOP_CMD	1
	"mail from:",
#define	MAIL_CMD	2
	"quit",
#define	QUIT_CMD	3
	"rcpt to:",
#define	RCPT_CMD	4
	"help",
#define	HELP_CMD	5
	"data",
#define	DATA_CMD	6
	"rset",
#define	RSET_CMD	7
	"expn",
#define EXPN_CMD	8
	NULLCHAR
};

/* Reply messages */
static char Help[] =	"214-Commands:\n"
			"214-HELO NOOP MAIL QUIT RCPT HELP DATA RSET EXPN\n"
			"214 End\n";
static char Banner[] =	"220 %s SMTP ready\n";
static char Closing[] = "221 Closing\n";
static char Ok[] =	"250 Ok\n";
static char Reset[] =	"250 Reset state\n";
static char Sent[] =	"250 Sent\n";
static char Ourname[] = "250 %s, Share and Enjoy!\n";
static char Enter[] =	"354 Enter mail, end with .\n";
static char Lowmem[] =	"421 System overloaded, try again later\n";
static char Ioerr[] =	"452 Temp file write error\n";
static char Badcmd[] =	"500 Command unrecognized\n";
static char Syntax[] =	"501 Syntax error\n";
static char Needrcpt[] = "503 Need RCPT (recipient)\n";
static char Unknown[] = "550 <%s> address unknown\n";
static char Noalias[] = "550 No alias for <%s>\n";

static int Ssmtp = -1; /* prototype socket for service */

/* Start up SMTP receiver service */
int
smtp1(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	struct sockaddr_in lsocket;
	int s;

	if(Ssmtp != -1){
		return 0;
	}
	psignal(Curproc,0);	/* Don't keep the parser waiting */
	chname(Curproc,"SMTP listener");

	lsocket.sin_family = AF_INET;
	lsocket.sin_addr.s_addr = INADDR_ANY;
	if(argc < 2)
		lsocket.sin_port = IPPORT_SMTP;
	else
		lsocket.sin_port = atoi(argv[1]);

	Ssmtp = socket(AF_INET,SOCK_STREAM,0);
	bind(Ssmtp,(char *)&lsocket,sizeof(lsocket));
	listen(Ssmtp,1);
	for(;;){
		if((s = accept(Ssmtp,NULLCHAR,(int *)NULL)) == -1)
			break;	/* Service is shutting down */

		if(availmem() - 3072 < Memthresh){
			usprintf(s,Lowmem);
			shutdown(s,1);
		} else {
			/* Spawn a server */
			newproc("SMTP server",3072,smtpserv,s,NULL,NULL,0);
		}
	}
	return 0;
}

/* Shutdown SMTP service (existing connections are allowed to finish) */
int
smtp0(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	close_s(Ssmtp);
	Ssmtp = -1;
	return 0;
}

static void
smtpserv(s,unused,p)
int s;
void *unused;
void *p;
{
	struct smtpsv *mp;
	char **cmdp,buf[RLINELEN],*arg,*cp,*cmd,*newaddr;
	struct recipient_s *ap,*list;
	int cnt;
	char address_type;

	sockmode(s,SOCK_ASCII);
	sockowner(s,Curproc);		/* We own it now */
	log(s,"open SMTP");

	if((mp = mail_create()) == NULLSMTPSV){
		printf(Nospace);
		log(s,"close SMTP - no space");
		close_s(s);
		return;
	}
	mp->s = s;

	(void) usprintf(s,Banner,Hostname);

loop:	if ((cnt = recvline(s,buf,sizeof(buf))) == -1) {
		/* He closed on us */
		goto quit;
	}
	if(cnt < 4){
		/* Can't be a legal command */
		usprintf(mp->s,Badcmd);
		goto loop;
	}	
	rip(buf);
	cmd = buf;

	/* Translate entire buffer to lower case */
	for(cp = cmd;*cp != '\0';cp++)
		*cp = tolower(*cp);

	/* Find command in table; if not present, return syntax error */
	for(cmdp = commands;*cmdp != NULLCHAR;cmdp++)
		if(strncmp(*cmdp,cmd,strlen(*cmdp)) == 0)
			break;
	if(*cmdp == NULLCHAR){
		(void) usprintf(mp->s,Badcmd);
		goto loop;
	}
	arg = &cmd[strlen(*cmdp)];
	/* Skip spaces after command */
	while(*arg == ' ')
		arg++;
	/* Execute specific command */
	switch(cmdp-commands) {
	case HELO_CMD:
		free(mp->system);
		mp->system = strdup(arg);
		(void) usprintf(mp->s,Ourname,Hostname);
		break;
	case NOOP_CMD:
		(void) usprintf(mp->s,Ok);
		break;
	case MAIL_CMD:
		if((cp = getname(arg)) == NULLCHAR){
			(void) usprintf(mp->s,Syntax);
			break;
		}
		free(mp->from);
		mp->from = strdup(cp);
		(void) usprintf(mp->s,Ok);
		break;
	case QUIT_CMD:
		(void) usprintf(mp->s,Closing);
		goto quit;
	case RCPT_CMD:	/* Specify recipient */
		if((cp = getname(arg)) == NULLCHAR){
			(void) usprintf(mp->s,Syntax);
			break;
		}

		/* rewrite address if possible */
		if((newaddr = rewrite_address(cp)) != NULLCHAR) {
			strcpy(buf,newaddr);
			cp = buf;
			free(newaddr);
		}

		/* check if address is ok */
		if ((address_type = validate_address(cp)) == BADADDR) {
			(void) usprintf(mp->s,Unknown,cp);
			break;
		}
		/* if a local address check for an alias */
		if (address_type == LOCAL)
			expandalias(&mp->to, cp);
		else
			/* a remote address is added to the list */
			add_rcpt(&mp->to, cp, address_type);

		(void) usprintf(mp->s,Ok);
		break;
	case HELP_CMD:
		(void) usprintf(mp->s,Help);
		break;
	case DATA_CMD:
		if(mp->to == NULLRCPT)
			(void) usprintf(mp->s,Needrcpt);
		else if ((mp->data = tmpfile()) == NULLFILE)
			(void) usprintf(mp->s,Ioerr);
		else {
			char tbuf[TLINELEN];
			time_t t;

			/* Add timestamp; ptime adds newline */
			time(&t);
			fprintf(mp->data,"Received: ");
			if(mp->system != NULLCHAR)
				fprintf(mp->data,"from %s ",mp->system);
			fprintf(mp->data,"by %s with SMTP\n\tid AA%ld ; %s",
					Hostname, get_msgid(), ptime(&t));
			if(ferror(mp->data)){
				(void) usprintf(mp->s,Ioerr);
			} else {
				(void) usprintf(mp->s,Enter);

				if ( recvmail( mp->s, tbuf, sizeof(tbuf),
						mp->data, 0 ) != -1
				  && mailit( mp->data, tbuf, sizeof(tbuf),
						mp->from, mp->to ) == 0) {
					(void) usprintf(mp->s,Sent);
				} else {
					(void) usprintf(mp->s,Ioerr);
				}
			}

			/* do some cleanup {WAS}*/
			fclose(mp->data);
			mp->data = NULLFILE;
			del_rcpt(mp->to);
			mp->to = NULLRCPT;
		}
		break;
	case RSET_CMD:
		del_rcpt(mp->to);
		mp->to = NULLRCPT;
		(void) usprintf(mp->s,Reset);
		break;
	case EXPN_CMD:
		if (*arg == '\0') {
			(void) usprintf(mp->s,Syntax);
			break;
		}

		list = NULLRCPT;
		/* rewrite address if possible */
		if((newaddr = rewrite_address(arg)) != NULLCHAR)
			if(strcmp(newaddr,arg) == 0) {
				free(newaddr);
				newaddr = NULLCHAR;
			}
			else {
				strcpy(buf,newaddr);
				arg = buf;
			}
		list = NULLRCPT;
		expandalias(&list,arg);
		if (strcmp(list->val,arg) == 0 && list->next == NULLRCPT)
			if(newaddr == NULLCHAR) {
				(void) usprintf(mp->s,Noalias,arg);
				del_rcpt(list);
				break;
			}
		ap = list;
		while (ap->next != NULLRCPT) {
			(void) usprintf(mp->s,"250-%s\n",ap->val);
			ap = ap->next;
		}
		usprintf(mp->s,"250 %s\n",ap->val);
		del_rcpt(list);
		free(newaddr);
		break;
	}
	goto loop;

quit:
	log(mp->s,"close SMTP");
	close_s(mp->s);
	mail_clean(mp);
	smtptick(NULL);			/* start SMTP daemon immediately */
}


/* Create control block, initialize */
static struct smtpsv *
mail_create()
{
	register struct smtpsv *mp;

	mp = (struct smtpsv *)callocw(1,sizeof(struct smtpsv));
	mp->from = strdup("");	/* Default to null From address */
	return mp;
}

/* Free resources, delete control block */
static void
mail_clean(mp)
register struct smtpsv *mp;
{
	if (mp == NULLSMTPSV)
		return;
	free(mp->system);
	free(mp->from);
	if(mp->data != NULLFILE)
		fclose(mp->data);
	del_rcpt(mp->to);
	free((char *)mp);
}


