/* io.c 		All base I/O routines live here
 *
 * Copyright 1991, Michael Westerhof, Sun Microsystems, Inc.
 * This software may be freely used, distributed, or modified, providing
 * this header is not removed.
 *
 * This is the only module that knows about base level UNIX/SunOS I/O
 * This is also the key dispatching module, so it knows about a lot more
 * than just I/O stuff.
 */

#undef USE_ICMP			/* not implemented yet, sorry */

#include "config.h"

#define IPPROTO_AX25 93

#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#ifdef USE_ICMP
#include <netinet/ip_icmp.h>
#endif
#include <netdb.h>
#include <fcntl.h>
#include <memory.h>
#include <stdio.h>
#include <string.h>
#include <sys/termios.h>

int ttyfd = -1;
struct termios ntermios;

#ifdef USE_ICMP
int icmpsock = -1;
#endif

int sock = -1;
struct sockaddr_in to;
struct sockaddr_in from;
int fromlen;

time_t last_bc_time;

/*
 * Initialize the io variables
 */

void
io_init()
{

/*
 * Close the file descriptors if they are open.  The idea is that we
 * will be able to support a re-initialization if sent a SIGHUP.
 */

	if(ttyfd>=0){
		(void)close(ttyfd);
		ttyfd = -1;
	}

	if(sock>=0){
		(void)close(sock);
		sock = -1;
	}

#ifdef USE_ICMP
	if(icmpsock>=0){
		(void)close(icmpsock);
		icmpsock = -1;
	}
#endif

/*
 * The bzero is not strictly required - it simply zeros out the
 * address structure.  Since both to and from are static, they are
 * already clear.
 */
	bzero( (char *)&to, sizeof(struct sockaddr) );

	to.sin_family = AF_INET;
	to.sin_port = 0;
}

/*
 * open and initialize the IO interfaces
 */

void io_open()
{
	int baudrate;

	sock = socket(AF_INET, SOCK_RAW, IPPROTO_AX25);
	if (sock<0) {
		perror("opening raw socket");
		exit(1);
	}

#ifdef USE_ICMP
	icmpsock = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP);
	if (icmpsock<0) {
		perror("opening raw ICMP socket");
		exit(1);
	}
#endif

	ttyfd = open(ttydevice, O_RDWR|O_NDELAY, 0);
	if (ttyfd<0) {
		perror("opening tty device");
		exit(1);
	}

/*
 * Hmmmm... Nothing like using BSD "select" stuff, but AT&T termio stuff!
 */

	if(ioctl(ttyfd, TCGETS, &ntermios)<0){
		perror("fetching tty device parameters");
		exit(1);
	}

	if(ttyspeed==50)baudrate=B50;
	else if(ttyspeed==50)baudrate=B50;
	else if(ttyspeed==75)baudrate=B75;
	else if(ttyspeed==110)baudrate=B110;
	else if(ttyspeed==134)baudrate=B134;
	else if(ttyspeed==150)baudrate=B150;
	else if(ttyspeed==200)baudrate=B200;
	else if(ttyspeed==300)baudrate=B300;
	else if(ttyspeed==600)baudrate=B600;
	else if(ttyspeed==1200)baudrate=B1200;
	else if(ttyspeed==1800)baudrate=B1800;
	else if(ttyspeed==2400)baudrate=B2400;
	else if(ttyspeed==4800)baudrate=B4800;
	else if(ttyspeed==9600)baudrate=B9600;
	else if(ttyspeed==19200)baudrate=B19200;
	else if(ttyspeed==38400)baudrate=B38400;
	else baudrate = B9600;

        ntermios.c_iflag = 0;
        ntermios.c_oflag = 0;
	ntermios.c_cflag = baudrate | CS8 | CREAD | CLOCAL;
	ntermios.c_lflag = 0;
        ntermios.c_cc[VMIN] = 0;
        ntermios.c_cc[VTIME] = 0;

	if(ioctl(ttyfd, TCSETS, &ntermios)<0){
		perror("setting tty device parameters");
		exit(1);
	}

	if(digi)send_params();

	last_bc_time = 0;	/* force immediate id */
}

/*
 * Start up and run the I/O mechanisms.
 *  run in a loop, using the select call to handle input.
 */

void
io_start()
{
	int n, nb, hdr_len;
	fd_set readfds;
	char buf[MAX_FRAME];
	struct timeval wait;
	struct ip *ipptr;
	time_t now;

	for(;;){

		if((bc_interval>0)&&digi){
			now = time(NULL);
			if (last_bc_time + bc_interval < now){
				last_bc_time = now;
				LOGL4("iostart: BEACON\n");
				do_beacon();
			}
		}

		wait.tv_sec =  10;
		wait.tv_usec = 0;

		FD_ZERO(&readfds);

		FD_SET(ttyfd, &readfds);
		FD_SET(sock, &readfds);

#ifdef USE_ICMP
		FD_SET(icmpsock, &readfds);
#endif

		nb = select(FD_SETSIZE,&readfds,(fd_set *)0,(fd_set *)0,&wait);

		if(nb < 0){
			perror("select");
			exit(1);
		}

		if(nb == 0){
			/* just so we go back to the top of the loop! */
			continue;
		}

		if(FD_ISSET(ttyfd, &readfds)){
			n = read(ttyfd, buf, MAX_FRAME);
			if(n<0){
				perror("read from tty");
				exit(1);
			}
			LOGL4("ttydata l=%d\n",n);
			if(n>0)assemble_kiss(buf,n);
/*
 * If we are in "beacon after" mode, reset the "last_bc_time" each time
 * we hear something on the channel.
 */
			if(!bc_every)last_bc_time = time(NULL);
		}

		if(FD_ISSET(sock, &readfds)){
			fromlen = sizeof from;
			n = recvfrom(sock, buf, MAX_FRAME, 0,
				(struct sockaddr *)&from, &fromlen);
			if(n<0){
				perror("raw socket");
			}
			ipptr = (struct ip *)buf;
			hdr_len = 4 * ipptr->ip_hl;
			LOGL4("ipdata from=%s l=%d, hl=%d\n",(char *)inet_ntoa(from.sin_addr),n,hdr_len);
			if(n>20)from_ip(buf + hdr_len, n - hdr_len);
		}

#ifdef USE_ICMP
		if(FD_ISSET(icmpsock, &readfds)){
			fromlen = sizeof from;
			n = recvfrom(icmpsock, buf, MAX_FRAME, 0,
				(struct sockaddr *)&from, &fromlen);
			if(n<0){
				perror("raw ICMP socket");
			}
			ipptr = (struct ip *)buf;
			hdr_len = 4 * ipptr->ip_hl;
			LOGL4("icmpdata from=%s l=%d, hl=%d\n",(char *)inet_ntoa(from.sin_addr),n,hdr_len);
		}
#endif

	} /* for forever */
}

/* Send an IP frame */

void
send_ip(buf, l, targetip)
unsigned char *buf;
int l;
unsigned char *targetip;
{
	if(l<=0)return;
	(void)memcpy((char *)&to.sin_addr, targetip, 4);
	LOGL4("sendipdata to=%s l=%d\n",(char *)inet_ntoa(to.sin_addr),l);
	if(sendto(sock, buf, l, 0, (struct sockaddr *)&to, sizeof to) < 0){
		perror("writing socket");
	}
}

/* Send a kiss frame */

void
send_tty(buf, l)
unsigned char *buf;
int l;
{
	int n;

	if(l<=0)return;
	LOGL4("sendttydata l=%d\n",l);
	n = write(ttyfd, buf, l);
	if(n<0){
		perror("writing to tty device");
	}
}
