/* routing.c    Routing table manipulation routines
 *
 * Copyright 1991, Michael Westerhof, Sun Microsystems, Inc.
 * This software may be freely used, distributed, or modified, providing
 * this header is not removed.
 *
 */

#include <stdio.h>
#include "config.h"
#include <sys/types.h>
#include <netinet/in.h>
#include <memory.h>

/* The routing table structure is not visible outside this module. */

#define TABLE_SIZE 128

struct route_table_entry {
	unsigned char callsign[7];
	unsigned char auto_added;
	unsigned char ip_addr[4];
} route_tbl[TABLE_SIZE];
int route_tbl_top;

unsigned char default_ip[4];
int have_default_route;


/* Initialize the routing table */
void
route_init()
{
	have_default_route = 0;
	route_tbl_top = 0;
}

/* Add a new route entry */
void
route_add(ip, call, autoadd, default_rt)
unsigned char *ip;
unsigned char *call;
int autoadd;
int default_rt;
{
	int i;

	if(ip==NULL) return;

	if(default_rt){
		(void)memcpy(default_ip, ip, 4);
		have_default_route = 1;
		LOGL4("added default route: \t%s\n",
			(char *)inet_ntoa(*(struct in_addr *)default_ip));
		return;
	}

	if(call==NULL) return;
	if(route_tbl_top >= TABLE_SIZE){
		(void)fprintf(stderr,"Routing table is full; entry ignored.\n");
	}

	for(i=0;i<6;i++)route_tbl[route_tbl_top].callsign[i] = call[i] & 0xfe;
	route_tbl[route_tbl_top].callsign[6] = (call[6] & 0x1e) | 0x60;
	route_tbl[route_tbl_top].auto_added = (autoadd!=0);
	(void)memcpy(route_tbl[route_tbl_top].ip_addr, ip, 4);
	LOGL4("added route: %s\t%s\n",
			call_to_a(route_tbl[route_tbl_top].callsign),
			(char *)inet_ntoa(*(struct in_addr *)route_tbl[route_tbl_top].ip_addr));
	route_tbl_top++;
	return;
}

/* Return a callsign given an IP address */
unsigned char *
ip_to_call(ip)
unsigned char *ip;
{
	int i;

	if(ip==NULL)return NULL;
	LOGL4("lookup ip addr %s ",(char *)inet_ntoa(*(struct in_addr *)ip));
	for(i=0;i<route_tbl_top;i++){
		if(memcmp(ip,route_tbl[i].ip_addr,4)==0){
			LOGL4("found call %s\n",call_to_a(route_tbl[i].callsign));
			return route_tbl[i].callsign;
		}
	}
	LOGL4("failed.\n");
	return NULL;
}

/* Return an IP address given a callsign */
unsigned char *
call_to_ip(call)
unsigned char *call;
{
	int i;
	unsigned char mycall[7];

	if(call==NULL)return NULL;

	for(i=0;i<6;i++)mycall[i] = call[i] & 0xfe;
	mycall[6] = (call[6] & 0x1e) | 0x60;

	LOGL4("lookup call %s ",call_to_a(mycall));
	for(i=0;i<route_tbl_top;i++){
		if(addrmatch(mycall,route_tbl[i].callsign)){
			LOGL4("found ip addr %s\n",
				(char *)inet_ntoa(*(struct in_addr *)route_tbl[i].ip_addr));
			return route_tbl[i].ip_addr;
		}
	}

	if(have_default_route){
		LOGL4("failed, using default ip addr %s\n",
				(char *)inet_ntoa(*(struct in_addr *)default_ip));
			return default_ip;
	}

	LOGL4("failed.\n");
	return NULL;
}

/* print out the list of routes */
void
dump_routes()
{
	int i;

	LOGL1("\n%d active routes, %d maximum\n",route_tbl_top,TABLE_SIZE);
	for(i=0;i<route_tbl_top;i++){
		LOGL1("  %s\t%s\n",
			call_to_a(route_tbl[i].callsign),
			(char *)inet_ntoa(*(struct in_addr *)route_tbl[i].ip_addr));
	}
	if(have_default_route){
		LOGL1("default route\t%s\n",(char *)inet_ntoa(*(struct in_addr *)default_ip));
	}
	(void)fflush(stdout);
}
